/*
 * Configger.java
 *
 * Created on February 1, 2005, 10:03 AM
 * Reads and writes to the Config file that controls which queryMap and datasetMap 
 * are preloaded. 
 */
package ShapeQuery;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.DataInputStream;
import java.io.InputStreamReader;
import java.io.IOException; 
import java.util.HashMap;
import java.util.Vector;
import java.net.URL;

/**
 * Reads and writes to the Config file that controls which queryMap and datasetMap 
 * are preloaded. 
 *
 * @author David Hessing
 *
 */

public class Configger {
    
    //private final String CONFIG = "config.txt";
    private BufferedReader br;
    private String splitter;
    
    private HashMap<String, String> queryMap;
    private Vector<String> queries;
    private HashMap<String, DataSet> datasetMap;
    private Vector<DataSet> datasets;
    private float tolerance;
    
    private final String COMMENT = "#";
    
    
    /** Creates a new instance of Configger for use in Applet */
    public Configger(String filename, URL codebase) throws IOException {
        //String userdir = System.getProperty("user.dir");
        br = getFile(filename, codebase);
        init();
    }
    /** Creates a new instance of Configger for use in Application */
    public Configger(String filename) throws IOException {
        //String userdir = System.getProperty("user.dir");
        br = new BufferedReader(new FileReader(filename));
        init();
    }

    private void init() throws IOException{
        tolerance = -1;
        String l = br.readLine().trim();
        if (l.length() == 1 && !l.equalsIgnoreCase(COMMENT)) {
            splitter = l.substring(0, 1);
            //make sure splitter isn't a reserved regex char
            if(!checkSplitter(splitter))
                splitter = "\\" + splitter;
        }
        else {
            throw new IOException("The first line of the config.txt file must consist of the splitting character.");
        }
        parseConfig();
    }
    
    //For use in Applet context.
    private BufferedReader getFile(String filename, URL codebase) throws IOException{
        URL sourceurl = null;
        DataInputStream url_input = null;
        try
        {
            sourceurl = new URL(codebase, filename);
            url_input = new DataInputStream(sourceurl.openStream());
        }
        catch (java.security.AccessControlException e)
        {
            throw new IOException(e.getMessage());
        }
        catch (java.net.MalformedURLException e)
        {
            throw new IOException(e.getMessage());
        }
        return new BufferedReader(new InputStreamReader(url_input));

        
    }
    
    /**
     * Checks to ensure that the character used for splitting the config.txt 
     * file isn't a reserved regex character.
     *
     * @param splitter A string of length 1 consisting of the current splitting character.
     * @return True if the splitting character is OK, false if it is a reserved character.
     */
    private boolean checkSplitter(String splitter) 
    { 
        String regs = "{}[]().$^&*+-?|,.\\";
        for (int i=0; i<regs.length(); i++){
            if (regs.substring(i, i+1).equalsIgnoreCase(splitter))
                return false;
        }
        return true;
    }
    

    /**
     * Parses the config file, reading the queryMap and the datasetMap.
     *
     */
    private void parseConfig() throws IOException
    {
        //(re)initialize the HashMaps
        queryMap = new HashMap<String, String>();
        datasetMap = new HashMap<String, DataSet>();
        queries = new Vector<String>();
        datasets = new Vector<DataSet>();
        
        //read the file until we get to QUERIES
        String l = br.readLine();
        while (l != null && !l.startsWith("QUERIES"))
            l = br.readLine();
        
        //read in the QUERIES
        l = br.readLine();
        while (l != null && !l.startsWith("DATASETS"))
        {
            //only process non-comment, non-blank lines
            if (l.trim().length() != 0 && !l.startsWith(COMMENT)) {
                String[] f = l.split(splitter);
                queryMap.put(f[0].trim(), f[1].trim());
                queries.add(f[1].trim());
            }
            l = br.readLine();
        }
        
        //read in the DATASETS
        l = br.readLine();
        while (l != null && !l.startsWith("TOLERANCE"))
        {
            //only process non-comment, non-blank lines
            if (l.trim().length() != 0 && !l.startsWith(COMMENT)) {
                String[] f = l.split(splitter);
                DataSet d = new DataSet(f[1].trim(), f[2].trim(), f[0].trim());
                datasetMap.put(d.visibleName, d);
                datasets.add(d);
            }
            l = br.readLine();
        }
        
        //read in the TOLERANCE
        l = br.readLine();
        while (l != null)
        {
            //only process non-comment, non-blank lines
            if (l.trim().length() != 0 && !l.startsWith(COMMENT)) {
                String f = l.trim();
                try {
                    tolerance  = Float.parseFloat(f);
                } catch (NumberFormatException nfe) {}
                //only considering one tolerance setting for now (last one)
            }
            l = br.readLine();
        }

    }
    
    public HashMap<String, String> getQueryMap() {
        return queryMap;
    }
    public HashMap<String, DataSet> getDataSetMap() {
        return datasetMap;
    }
    
    public Vector<String> getQueries() {
        return queries;
    }
    public Vector<DataSet> getDataSets() {
        return datasets;
    }

    public float getTolerance() {
        return tolerance;
    }
    
}
