package ShapeQuery;
// Constraint.java: Representation of a user-entered component of a
//                   shape query.  A constraint is independent of
//                   point sets or of query plans (contrast with
//                   'rules'); a constraint simply indicates that two
//                   pairs of points must have a particular
//                   relationship with one another.


import java.util.*;  // for vector
import java.io.*;    // for reading files
import java.net.*;

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;


class Constraint
{
    // Constraints may constrain either a distance or an angle, though
    // not both.  (If you wanted to constrain both a distance AND an
    // angle, you'd express that as two constraints.)
    public static final int RULE_TYPE_DISTANCE = 0;
    public static final int RULE_TYPE_ANGLE    = 1;


    // keep track of what constraint # this is, for fast
    // ordering/comparison
    static int nextId;
    int id;


    // ------------------------------------------------------------
    // information that user specified

    // a constraint consists of 9 pieces of information:
    // - the names of start and end points for one edge ["required edge"]
    // - the names of start and end points for a second edge ["target edge"]
    // - whether to look at the distances of the two edges or the
    //   angles ["rule type"]
    // - the A and B in the equation:
    //      y = Ax + B
    //   where 'x' is the distance or angle (as defined by "rule
    //   type") of the first edge, 'y' is the same of the second edge,
    //   and A and B are constants.  [A = "factor", B = "difference"]
    // - the tolerance
    // - a bit specifying whether the tolerance is relative (say, a
    //   percent of the distance) or absolute (say, in degrees)

    // Note that it's sometimes useful to "reverse" a constraint,
    // i.e., swap the required and target edges (since this can be
    // done with simple math, and you'd miss valid possibilities
    // without it).  The reverse() function does this.

    // the edge we need to have already to apply this constraint
    String myRequiredEdgeStartPointName;
    String myRequiredEdgeEndPointName;

    // the edge we're comparing to
    String myTargetEdgeStartPointName;
    String myTargetEdgeEndPointName;

    int myRuleType;  // distance or angle
    
    float myFactor;
    float myDifference;
    float myTolerance;
    boolean myToleranceIsAbsolute;

    //added by DJH
    Vector<String> myRequiredEdgeStartPointLabels;
    Vector<String> myRequiredEdgeEndPointLabels;
    Vector<String> myTargetEdgeStartPointLabels;
    Vector<String> myTargetEdgeEndPointLabels;
    boolean reversed;

    
    // create a new constraint with the specified parameters
    Constraint(String _requiredEdgeStartPointName,
	       String _requiredEdgeEndPointName,

	       String _targetEdgeStartPointName,
	       String _targetEdgeEndPointName,

	       int _ruleType,  // distance or angle
    
	       float _factor,
	       float _difference,
	       float _tolerance,
	       boolean _toleranceIsAbsolute)
    {
	// keep track of this constraint's id number
	id = nextId;
	nextId++;

	myRequiredEdgeStartPointName = _requiredEdgeStartPointName;
	myRequiredEdgeEndPointName = _requiredEdgeEndPointName;
	myTargetEdgeStartPointName = _targetEdgeStartPointName;
	myTargetEdgeEndPointName = _targetEdgeEndPointName;

	myRuleType = _ruleType;

	myFactor = _factor;
	myDifference = _difference;
	myTolerance = _tolerance;
	myToleranceIsAbsolute = _toleranceIsAbsolute;
        
        reversed = false;
    }
    
    public void setRequiredEdgeStartPointLabels(Vector<String> labels) {myRequiredEdgeStartPointLabels = labels;}
    public void setRequiredEdgeEndPointLabels(Vector<String> labels) {myRequiredEdgeEndPointLabels = labels;}
    public void setTargetEdgeStartPointLabels(Vector<String> labels) {myTargetEdgeStartPointLabels = labels;}
    public void setTargetEdgeEndPointLabels(Vector<String> labels) {myTargetEdgeEndPointLabels = labels;}
    
    /**
     * Gets a set of all labels on all points in this constraint with duplicates removed.
     * Used to perform query cost estimation calculation.
     *
     * @return A set of all labels on all points in this constraint with duplicates removed.
     */
    public HashSet<String> getAllLabels() {
        HashSet<String> allLabels = new HashSet<String>();
        if (myRequiredEdgeStartPointLabels != null) {
            allLabels.addAll(myRequiredEdgeStartPointLabels);
        }
        if (myRequiredEdgeEndPointLabels != null) {
            allLabels.addAll(myRequiredEdgeEndPointLabels);
        }
        if (myTargetEdgeStartPointLabels != null) {
            allLabels.addAll(myTargetEdgeStartPointLabels);
        }
        if (myTargetEdgeEndPointLabels != null) {
            allLabels.addAll(myTargetEdgeEndPointLabels);
        }
        return allLabels;
    }

    /**
     * Create a copy of this constraint. Useful, since we sometimes want a constraint to be referenced by a rule,
     * but the constraint may later be reversed when considering other query plans. Note the copy is not deep - label 
     * vectors are not recreated.
     *
     * @return A new constraint identical to this one at the time of the copy.
     */
    public Constraint copy() {
        Constraint newc = new Constraint(myRequiredEdgeStartPointName, myRequiredEdgeEndPointName, 
                                        myTargetEdgeStartPointName, myTargetEdgeEndPointName, myRuleType, myFactor, 
                                        myDifference, myTolerance, myToleranceIsAbsolute);
        newc.reversed = reversed;
        newc.setRequiredEdgeStartPointLabels(myRequiredEdgeStartPointLabels);
        newc.setRequiredEdgeEndPointLabels(myRequiredEdgeEndPointLabels);
        newc.setTargetEdgeStartPointLabels(myTargetEdgeStartPointLabels);
        newc.setTargetEdgeEndPointLabels(myTargetEdgeEndPointLabels);
        return newc;
    }
    
    // reverse this constraint's required and target edges (and adjust
    // measurmenets/tolerances appropriately)
    void reverse()
    {
        reversed = !reversed;
	// ------------------------------------------------
	// simply swap required and target edge points and their labels
	String tmpEdgeStart = myRequiredEdgeStartPointName;
	String tmpEdgeEnd = myRequiredEdgeEndPointName;

	myRequiredEdgeStartPointName = myTargetEdgeStartPointName;
	myRequiredEdgeEndPointName = myTargetEdgeEndPointName;
	
	myTargetEdgeStartPointName = tmpEdgeStart;
	myTargetEdgeEndPointName = tmpEdgeEnd;	

        Vector<String> tmpEdgeStartLabels = myRequiredEdgeStartPointLabels;
        Vector<String> tmpEdgeEndLabels = myRequiredEdgeEndPointLabels;
        
        myRequiredEdgeStartPointLabels = myTargetEdgeStartPointLabels;
        myRequiredEdgeEndPointLabels = myTargetEdgeEndPointLabels;
        
        myTargetEdgeStartPointLabels = tmpEdgeStartLabels;
        myTargetEdgeEndPointLabels = tmpEdgeEndLabels;
	// ------------------------------------------------	
	// (rule type doesn't change)

	// ------------------------------------------------	
	// factor must be inverted

	if (myFactor!=0)
	    myFactor = 1/myFactor;

	// what should we do if factor is zero?

	// ------------------------------------------------
	// for an angle, if reversing, need to invert measure.
        // otherwise, any absolute difference doesn't change.
        if (myRuleType == RULE_TYPE_ANGLE) {
            myDifference = -myDifference;
        }

	// ------------------------------------------------
	// if tolerance is relative, it must be adjusted by 'factor'.
	// (otherwise we don't change it)

	// !!! figure out if this is right
	if (myToleranceIsAbsolute)
	{
//	    myTolerance = myTolerance * myFactor;
	}
    }


    // print out this constraint
    void print()
    {
	OutputWriter.print("    ["+myRequiredEdgeStartPointName + "\t"
			 +myRequiredEdgeEndPointName + "\t"
			 +myTargetEdgeStartPointName + "\t"
			 +myTargetEdgeEndPointName + "\t");
	if (myRuleType==RULE_TYPE_DISTANCE)
	{
	    OutputWriter.print("D\t");
	}
	else if (myRuleType==RULE_TYPE_ANGLE)
	{
	    OutputWriter.print("A\t");
	}
	else
	{
	    OutputWriter.print("?\t");
	}

	OutputWriter.print(myFactor+"\t"
			 +myDifference+"\t"
			 +myTolerance+"\t");
	if (myToleranceIsAbsolute)
	    OutputWriter.print("Y");
	else
	    OutputWriter.print("N");

	OutputWriter.println("]");
    }
}
