/*
 * CssTree.java
 *
 * Created on October 4, 2005, 2:58 AM
 *
 * The CssTree is a Cache Sensitive Search Tree which is built upon a already sorted Index
 * and performs better than binary search (over the Index).
 *
 */

package ShapeQuery;

/**
 * This class represents the CSS Tree. It is the array of CSS Keys.
 * @author Pranay Jain
 */

import java.util.*;  // for vector
import java.io.*;    // for reading files
import java.lang.Math.*;

public class CssTree {

    private CssKey keys[];          // Array of CssKey. Each CssKey consists of 2 Keys, Key1 and Key2
                                    // and is sorted on (Key1, Key2)
    
    private PointPairIndex index;   // The sorted index over which the CssTree is being built.

    // If L2 Cache Line Size = 64 KB
    private int entryPerLeaf = 8;   // Each node has '8' PointPairIndex Objects of 8 bytes each.
    private int keyPNode = 8;       // Each node has '8' CssKey Objects of 8 bytes each.

    // If L2 Cache Line Size = 128 KB
    //private int entryPerLeaf = 16;   // Each node has '16' PointPairIndex Objects of 8 bytes each.
    //private int keyPNode = 16;       // Each node has '16' CssKey Objects of 8 bytes each.

    
   /* Properties of the CssTree after its generated.
    * internal: is the index of the first non-internal nodes (starting from 0).
    * thres:    is the index of the first Leaf node where the key should be found in the first half of
    *           the Index array.
    */
    private int internal, thres;
    
    /* Creates a new instance of CssTree */
    public CssTree(PointPairIndex newPointPairIndex) {
        
        index = newPointPairIndex;
        int indexOrder[] = index.getIndexOrder();

	Calendar startTime;
	Calendar finishTime;

	OutputWriter.println("      Building node array...");
	startTime = Calendar.getInstance();

	/* Allocate enough room for the nodes */
        int size = index.size();
        
        /*
         *  B           : no. of Leaf Nodes (size/entryPerLeaf).
         *  nodeSize    : Same as the keyPNode
         *  k           : Floor ( Log to the base nodeSize+1 of Leaf Nodes)
         *  mBits       : Required for faster multiplication. mBits = Log to the base 2 of keyPNode
         */
        int B, nonLeaf, mask, mBits, nodeSize, branch;
 
        /* Temp variables */
        int i, j, node, child, loc, k, expk, x;
        
        //mBits = 3 when keyPNode = 8;
        mBits = (int)Math.ceil(Math.log((double)keyPNode)/Math.log((double)2));        
        nodeSize = 1<<mBits;
        mask = nodeSize - 1;
        branch = nodeSize + 1;
        
        B = (int)Math.ceil((double)size/(double)entryPerLeaf);
        k = (int)Math.ceil(Math.log((double)B)/Math.log((double)branch));
        expk = (int)Math.pow(branch,k);
        x = (int)Math.floor((double)(expk-B)/(double)nodeSize);

        thres = (int)Math.floor( (expk-1)>>mBits );
        nonLeaf = thres - x;
        internal = nonLeaf;
        
        keys = new CssKey[nonLeaf*keyPNode];
	//OutputWriter.println("      SIZE  = " + size + "  B = " + B + "  k = " + k + "  expk = " + expk + "  x = " + x + " nonLeaf = " + nonLeaf + " thres = " + thres);

        for (i=internal-1; i >= 0; i--) {
            for (j = i*keyPNode+keyPNode-1; j >= i*keyPNode; j--) {
                child = branch*i + (j&mask) + 1;
                
                while (child < nonLeaf)
                    child = branch*child + branch;
                
                loc = child - thres;          // loc is the node index                
                loc = loc * entryPerLeaf;     // loc is the entry index
                //OutputWriter.println("child = " + (child) + "  loc = " + (loc)); 
    
                int pos;
                if (loc < 0) {
                    loc += size;
                    pos = loc + entryPerLeaf - 1;
                    if(pos >= size)
                        pos = size-1;
                    //OutputWriter.println("1 = " + pos); 
                }
                else {
                    if (loc < size - x*entryPerLeaf) {
                        pos = loc + entryPerLeaf - 1;
                        if(pos >= size)
                            pos = size-1;
                        //OutputWriter.println("2 = " + pos); 
                    }
                    else {
                       pos =  size - x*entryPerLeaf - 1;
                       //OutputWriter.println("3 = " + pos); 
                    }
                }
                
                /* Get the Keys from PointPairIndex's (pos+1)th Node
                 * The Keys essentially would be the arguments pointed by value of indexOrder[0] and indexOrder[1]
                */
                
                PointPairNode tempNode = newPointPairIndex.getPointPairNode(pos);
                CssKey key = new CssKey((float)tempNode.getArgument(0),(float)tempNode.getArgument(1));
                keys[j] = key;
            }
        }
         
          
	finishTime = Calendar.getInstance();
	OutputWriter.println("      ===> keys array build time: "
			   +(finishTime.getTimeInMillis()-
			     startTime.getTimeInMillis())
			   +" milliseconds");
    }
    
    /*
     * Prints out the details of this index for testing.
     */
    public void print(){
        String s = "";
        for(int i=0; i < keys.length; i++) {
            s += keys[i].getDescription();
        }
        OutputWriter.println(s);
    }

    
    // Get elements whose point ids match 'id' and whose measures are
    // between 'lowMeasure' and 'highMeasure'.  !!! this assumes that
    // every index has an id as the first parameter and an angle or
    // distance as the second
 
    PointPair[] find(int matchArgument,
		     float lowMeasure, float highMeasure)
    {   
        Vector<PointPair> matchingPointPairs = new Vector<PointPair>();

	// To correctly accomodate angle wrapping, we need to deal
	// differently with angles.  For starters, we move angles to
	// range -180 < n <= 180.  (This may make lowMeasure be
	// greater than highMeasure, in which case we need to compare
	// angles with x axis.)  See ShapeQuery.java for more on this.
	if (index.getIndexOrder()[1]==IndexType.ANGLE)
	{
	    // first perform adjustment to get angles in right range

	    // if lower bound = -180, that means, "find anything >
	    // -180".  this might cause us to miss a valid match if
	    // the angle is exactly 180.0f, since a 180-degree angle
	    // is stored only at the end of the list, not at the
	    // beginning !!!
	    if (lowMeasure!=-180.0f)
		lowMeasure = ShapeQuery.angleAdjust(lowMeasure);
 	    highMeasure = ShapeQuery.angleAdjust(highMeasure);
	
	    // idea for future:
	    //
	    // we can cut out a bit of work if there's an angle
	    // wrapping situation, because in order for there to be
	    // any matches, either the first or last nodes (or more
	    // likely, both) of this point id must match this search
	    //
	    // for now:
	    //
	    // just do two lookups, one for lowmeasure < n <= 180, and
	    // the other for -180 < n < highmeasure
	    if (lowMeasure > highMeasure)
	    {
		// perform this search with two lookups

		PointPair[] lowLookup;
		PointPair[] highLookup;

		lowLookup = find(matchArgument,
				 lowMeasure, 180.0f);

		highLookup = find(matchArgument,
				  -180.0f, highMeasure);

		PointPair[] fullResult = new PointPair[lowLookup.length+
						       highLookup.length];

		for (int i = 0; i < lowLookup.length; i++)
		{
		    fullResult[i] = lowLookup[i];
		}

		for (int i = 0; i < highLookup.length; i++)
		{
		    fullResult[i+lowLookup.length]=highLookup[i];
		}


		return fullResult;

	    }
	}

	// use CssTree search to find ANY node that involves point 'id'
	// and that's between 'lowMeasure' and 'highMeasure'
        
        float[][] range = new float[2][2]; //2 Keys and each key would have a lowMeasure and highMesaure
        range[0][0] = range[0][1] = matchArgument;
        range[1][0] = lowMeasure;
        range[1][1] = highMeasure;

        int matchingPointPair = cssTreeSearch64(range);

        return index.collectMatches(matchingPointPair, range);

    }

    
    
    /* In the first dimension,
     * the array must be of the length = 2. In the second dimension,
     * the array must be of length 2. Each range[i][0] represents a lower bound
     * on the range of acceptable values for sort key of Keyi. Likewise
     * range[i][1] is an upper bound on the range of acceptable values for sort 
     * key of Keyi.
     *
     * For example, if the index is sorted on <firstID, Distance>, and if range[0][0] = 3,
     * range[0][1] = 3, range[1][0] = .5, range[1][1] = .75, then we are searching 
     * for all nodes whose first point's id is 3 and whose distances lie in the range
     * .5 to .75 (inclusive). 
     *
     * Note that this method does not address the problem of angle wrapping. This 
     * must be handled externally.
     *
     * @param range 2D float array as described above.
     * @return array of matching nodes.
     */
    PointPair[] find(float[][] range)
    {

        int matchingPointPair = cssTreeSearch64(range);
        
        return index.collectMatches(matchingPointPair, range);
    }
   
 
    private int cssTreeSearch64(float[][] range) {
        
    int p, loc=0, node;
    //mBits = 3 when keyPNode = 8;
    int  mBits = (int)Math.ceil(Math.log((double)keyPNode)/Math.log((double)2));   

    node=0;
    while ( node < internal) {
        p = node<<mBits;
        
        if (keys[p+4].compareTo(range)>=0) {
            if (keys[p+1].compareTo(range) >= 0) {
                if (keys[p+0].compareTo(range) >= 0)
                    loc=1;
                else
                    loc=2;
                }
            else {  // keys[p+1]<key
                if (keys[p+2].compareTo(range) >= 0)
                    loc=3;
                else
                    if (keys[p+3].compareTo(range) >= 0)
                        loc=4;
                    else
                        loc=5;
            }
        }
        else { // keys[p+4]<key
            if (keys[p+6].compareTo(range) >= 0) {
                if (keys[p+5].compareTo(range) >= 0)
                    loc=6;
                else
                    loc=7;
            }    
            else {
                if (keys[p+7].compareTo(range) >= 0)
                    loc=8;
                else
                    loc=9;                            
            }
        }

        node = node*(keyPNode+1) + loc;
    } 

    //OutputWriter.print("\nnode = " + (node) + "  loc = " + (loc));

    loc = node-thres;
    loc = (loc*keyPNode);     // loc is the entry index
    //OutputWriter.print("\nl = " + (loc));

    //System.out.println(loc + ", " + index.size());
    if (loc<0)
        loc+=index.size();

    //now at the  leaves
    int pLeaf = loc;
    //OutputWriter.print("\npLeaf = " + (pLeaf) + " " + keys.length);
    loc = index.leafSearch64(pLeaf, range);
    //OutputWriter.print("\npLeaf + loc = " + (pLeaf + loc) + "\n");  
  
    if (loc<keyPNode)
        return (pLeaf + loc);
    else
        return -1;

    }
    

    private int cssTreeSearch128(float[][] range) {
        
    int p, loc=0, node;
    //mBits = 4 when keyPNode = 16;
    int  mBits = (int)Math.ceil(Math.log((double)keyPNode)/Math.log((double)2));   

    node=0;
    while ( node < internal) {
        p = node<<mBits;
        
        if (keys[p+8].compareTo(range)>=0) {
            if (keys[p+3].compareTo(range)>=0) {
		if (keys[p+1].compareTo(range)>=0) {
	  		if (keys[p+0].compareTo(range)>=0)
                            loc=1;
	  		else
                            loc=2;
		}
		else {  //keys[p+1]<key
                    if (keys[p+2].compareTo(range)>=0)
                        loc=3;
                    else
	    		loc=4;
		}
            }
            else { // keys[p+3]<key
                if (keys[p+5].compareTo(range)>=0) {
                    if (keys[p+4].compareTo(range)>=0)
                        loc=5;
                    else
                        loc=6;
                    }
                else {
                    if (keys[p+6].compareTo(range)>=0)
                        loc=7;
                    else {
                        if (keys[p+7].compareTo(range)>=0)
                            loc=8;
                        else
                            loc=9;
                    }
                }
            }
        }
        else { // keys[p+8]<key {
            if (keys[p+12].compareTo(range)>=0) {
                if (keys[p+10].compareTo(range)>=0) {
                    if (keys[p+9].compareTo(range)>=0)
                        loc=10;
                    else
                        loc=11;
                }
                else {  // keys[p+1]<key
                    if (keys[p+11].compareTo(range)>=0)
                        loc=12;
                    else
                        loc=13;
                }
            }
            else { // keys[p+3]<key
                if (keys[p+14].compareTo(range)>=0) {
                    if (keys[p+13].compareTo(range)>=0)
                        loc=14;
                    else
                        loc=15;
                }
                else {
                    if (keys[p+15].compareTo(range)>=0)
                        loc=16;
                    else
                        loc=17;
                }
            }
        }
        node = node*(keyPNode+1) + loc;
    }

    //OutputWriter.print("\nnode = " + (node) + "  loc = " + (loc));

    loc = node-thres;
    loc = (loc*keyPNode);     // loc is the entry index
    //OutputWriter.print("\nl = " + (loc));

    if (loc<0)
        loc+=index.size();

    //now at the  leaves
    int pLeaf = loc;
    //OutputWriter.print("\npLeaf = " + (pLeaf) + " " + keys.length);
    loc = index.leafSearch128(pLeaf, range);
    //OutputWriter.print("\npLeaf + loc = " + (pLeaf + loc) + "\n");  
  
    if (loc<keyPNode)
        return (pLeaf + loc);
    else
        return -1;

    }
    
    
    
}