/*
 * EdgeIndexLookupRule.java
 *
 * Created on June 12, 2005, 12:43 PM
 */

package ShapeQuery;

import java.util.Calendar;
import java.util.Iterator;
import java.util.Vector;

/**
 * For two edges, if only one edge is known (already exist in 
 * the intermediate solutions), this class performs an index lookup to 
 * find matching second edges. This class uses <Angle, Distance> and 
 * <Distance, Angle> indices to perform lookups.
 *
 * This is mostly Vespe's code from PointIndexLookupRule, modified to lookup
 * edges.
 *
 * @author David Hessing
 */
public class EdgeIndexLookupRule extends ShapeQueryRule {
    
   


    public boolean consReversed; //flag indicating if this rule was built from a reversed constraint. For debugging purposes.

    // information extracted from the constraint on what this rule
    // means
    String myRelativeEdgeStartPointName;
    String myRelativeEdgeEndPointName;
    int myRuleType;
    float myFactor;
    float myDifference;
    float myTolerance;
    boolean myToleranceIsAbsolute;
    
    Constraint con;


    /** Creates a new instance of EdgeIndexLookupRule */
    public EdgeIndexLookupRule(Constraint constraint)
    {
        
        myRelativeEdgeStartPointName = constraint.myRequiredEdgeStartPointName;
	myRelativeEdgeEndPointName = constraint.myRequiredEdgeEndPointName;
	myRuleType = constraint.myRuleType;
	myFactor = constraint.myFactor;
	myDifference = constraint.myDifference; //(constraint.reversed) ? -constraint.myDifference : constraint.myDifference;
	myTolerance = constraint.myTolerance;
	myToleranceIsAbsolute = constraint.myToleranceIsAbsolute;
        consReversed = constraint.reversed;
        
        con = constraint.copy();
    }


    void print()
    {
	OutputWriter.println("> index lookup rule for edge: (" + con.myTargetEdgeStartPointName 
			   + "," + con.myTargetEdgeEndPointName + ")                " + getDescription());
    }

    String getDescription() {
        String s = "EdgeIndexLookupRule - ";
        switch(myRuleType)
	{
	    case Constraint.RULE_TYPE_DISTANCE:
                s += "Distance Lookup: "; break;
            case Constraint.RULE_TYPE_ANGLE:
                s += "Angle Lookup: "; break;
            default:
                s += "Unknown Ruletype! "; break;
        }
        s += "RelEdge:(" + con.myRequiredEdgeStartPointName + "," + con.myRequiredEdgeEndPointName + ") ";
        s += "TarEdge:(" + con.myTargetEdgeStartPointName + "," + con.myTargetEdgeEndPointName + ") ";
        s += "Factor=" + con.myFactor + " Diff=" + con.myDifference + " Tol=" + con.myTolerance + " Abs=" + con.myToleranceIsAbsolute + " ";
        s += "Reversed=" + consReversed;
        return s;
        
    }

    float estimateCost(PointUniverse universe, float rows)
    {
	// For each row in a set, we need to do an index lookup, which
	// is on the expensive side.  This perhaps overblows the cost
	// of an index lookup -- it'd probably be better modeled as
	// the log of the average rowcount instead.

        //corrected to incorporate the work done while searching through the index - Pranay
        //float cost = rows*universe.averageRowCount();
        //float cost = rows*(float)Math.log(universe.averageRowCount());
        
        float cost = (float)rows*(float)(Math.log((double)universe.averageRowCount())/Math.log((double)2));
        
        return cost;
    }

    
    //FIX THIS TO TAKE INTO ACCOUNT NEW INDEX TYPES! -DJH 
    //FOR NOW ASSUME THIS TYPE OF INDEX SEARCH IS MORE EXPENSIVE THEN 
    //THE <ID, ANGLE> OR <ID, DISTANCE> SEARCH BY SOME CONSTANT FACTOR.
    // the number of rows returned is calculated for angles based on
    // arithmetic (assuming angles are evenly distributed 0-360) or,
    // for distances, by performing a lookup on the median pointpair
    // in each pointset and averaging those results
    float estimateReturnedRowMultiplier(PointUniverse universe)
    {
	float rowCount = 0f;

	switch(myRuleType)
	{
	    case Constraint.RULE_TYPE_DISTANCE:
	    {
		// find rowCount for each median edge of each pointset;
		// return average of those

		if (universe.pointSets.size()!=0)
		{
		    Iterator iter = universe.pointSets.iterator();
		    while (iter.hasNext())
		    {
			PointSet set = (PointSet) iter.next();
		    
			rowCount += (set.getIndex(IndexType.DISTANCE_ANGLE)).estimateReturnedRows(
			    myFactor,
			    myDifference,
			    myTolerance,
			    myToleranceIsAbsolute
			    );
		    }
		    rowCount = rowCount / universe.pointSets.size();
		}

		break;
            }
	    case Constraint.RULE_TYPE_ANGLE:
	    {
		// find rowCount for each median edge of each pointset;
		// return average of those

		if (universe.pointSets.size()!=0)
		{
		    // look only at first pointset, since they'll all
		    // have the same result
		    PointSet set = (PointSet) universe.pointSets.firstElement();
		    
		    rowCount = (set.getIndex(IndexType.ANGLE_DISTANCE)).estimateReturnedRows(
			myFactor,
			myDifference,
			myTolerance,
			myToleranceIsAbsolute
			);                    
		}
		break;
	    }
	    default:
	    {
		OutputWriter.println("don't know what index to use");
		break;
	    }
	}
        
        
        //now, reduce the rowcount by the label reduction factor - DJH
        rowCount *= computeLabelReduction(con.getAllLabels(), universe);
        
	return rowCount;
    }

    

    // NOTE!!! In the special case where two disconnected edges have both a required
    // distance relationship and a required angle relationship, and AT LEAST one
    // of the relationships is an equality (not a range), (e.g. two 
    // parallel lines of the same length), then ONE lookup in a <ANGLE, DISTANCE>
    // or <DISTANCE, ANGLE> index will suffice, even though this will 
    // be expressed in the query as two seperate constraints. Thus, it may 
    // be worthwhile as part of the query plan to try to combine contraints
    // as an initial step. The current method below doesn't support this, 
    // and, in fact, probably a modified constraint object will be required. -DJH
    
    
    // use 'solutions' to build new QuerySolutionSet, which contains
    // all valid extensions of members of 'solutions'
    QuerySolutionSet apply(QuerySolutionSet<QuerySolution> solutions,
			   PointSet pointSet)

    {
	OutputWriter.printDebugDetail("  applying indexlookup rule for edge: (" + 
                con.myTargetEdgeStartPointName + "," + con.myTargetEdgeEndPointName + ")");
	
	// !!! is this where the program sucks up a ton of memory???

        
        //Decide which index we're going to use
        IndexType type = null;
        switch(myRuleType)
        {
            case Constraint.RULE_TYPE_DISTANCE:
            {
                type = IndexType.DISTANCE_ANGLE;
                break;
            }
            case Constraint.RULE_TYPE_ANGLE:
            {
                type = IndexType.ANGLE_DISTANCE;
                break;
            }
            default:
            {
                OutputWriter.println("don't know what index to use!");
                return null;
            }
        }       
        
        //initialize the range array
        float[][] range = new float[pointSet.getIndex(type).getIndexOrder().length][2];
        
        
	// we'll copy from initial solutions to final solutions all
	// solutions that match (into which we add an additional two points)
	Iterator iter = solutions.iterator(); // initial solutions
	QuerySolutionSet<QuerySolution> finalSolutions = new QuerySolutionSet<QuerySolution>(solutions.getEdgesToDraw());  // final solns
	
	int count = 0;
	while (iter.hasNext())
	{
	    // look at next solution
	    QuerySolution currentSolution = (QuerySolution)iter.next();
	    
            
	    // -------------------------------------------------
	    // figure out what angle/distance we're looking for

	    // the measure we're looking for
	    float targetMeasure = 0.0f;
            PointPair relativeEdge = null;
	    if (myRelativeEdgeStartPointName!="")
	    {
		// this is a relative constraint; construct the relative edge
		relativeEdge = new PointPair(
		    currentSolution.getPoint(myRelativeEdgeStartPointName),
		    currentSolution.getPoint(myRelativeEdgeEndPointName));

		switch(myRuleType)
		{
		    case Constraint.RULE_TYPE_DISTANCE:
		    {
			targetMeasure = relativeEdge.distance;
			break;
		    }
		    case Constraint.RULE_TYPE_ANGLE:
		    {
			targetMeasure = relativeEdge.angle;
			break;
		    }
		}

		// calculate the target based on the multiplier and difference
		targetMeasure = myFactor * targetMeasure + myDifference;

	    }
	    else
	    {
		// an absolute constraint; 'difference' is our target
		// measure
		targetMeasure = myDifference;
	    }

	    // -------------------------------------------------
	    // apply tolerance to find max and min
	    float lowMeasure, highMeasure;
	    if (myToleranceIsAbsolute)
	    {
		lowMeasure = targetMeasure - myTolerance;
		highMeasure = targetMeasure + myTolerance;
	    }
	    else
	    {
		lowMeasure = targetMeasure*(1-myTolerance/100.0f);
		highMeasure = targetMeasure*(1+myTolerance/100.0f);
	    }


            
            
            //set the parameters for the index search
            
            //Note we may be using only half of the index's cabability. See the 
            //comments immediately preceeding this method for more explanation. -DJH
            range[0][0] = lowMeasure;
            range[0][1] = highMeasure;
            range[1][0] = Float.NEGATIVE_INFINITY;
            range[1][1] = Float.POSITIVE_INFINITY;
            
            
            
            // -------------------------------------------------
	    // perform index lookup

	    // a list of the pointpairs that match this index search
	    PointPair results[];

            Calendar startTime;
            Calendar finishTime;
            startTime = Calendar.getInstance();

            //results = (pointSet.getIndex(type)).find(range);
            
            // Call the CssTree Search instead of the Index Binary Search - Pranay
            results = (pointSet.getCssTree(type)).find(range);

            finishTime = Calendar.getInstance();
            long cssTreeBuildTime = finishTime.getTimeInMillis() - startTime.getTimeInMillis();
            //OutputWriter.println("\n===> EdgeIndexLookUp time: " + cssTreeBuildTime + " milliseconds for " + results.length + " results.");

            
            //output debugging stuff to the log 
            //USEFUL WHEN TRYING TO UNDERSTAND THE INDEX LOOKUPS - DJH
            /*
            Point pnt1 = currentSolution.getPoint(myRelativeEdgeStartPointName);
            Point pnt2 = currentSolution.getPoint(myRelativeEdgeEndPointName);
            OutputWriter.print(" Edge in CurrSol: [" + myRelativeEdgeStartPointName + "," + pnt1.getExternalID() + "]" + 
                "->[" + myRelativeEdgeEndPointName + "," + pnt2.getExternalID() + "](" + relativeEdge.angle + ")");
            for(int i = 0; i < results.length; i++) {
                OutputWriter.print(" Found: [" + results[i].firstPoint.getExternalID() + "]->[" + results[i].secondPoint.getExternalID() + "](" + results[i].angle + ")");
            }
            OutputWriter.print("\n");
            */

	    // -------------------------------------------------
	    // examine results of index search
            
            //LABEL STUFF - DJH
            //first, figure out which set of labels must be matched
            //myNewPointName is either the constraint's targetEndPoint or the targetStartPoint
            Vector<String> sPointLabels = con.myTargetEdgeStartPointLabels;
            Vector<String> ePointLabels = con.myTargetEdgeEndPointLabels;
           

            
	    // for each returned match that fits current solution...
	    for (int j = 0; j < results.length; j++)
	    {
                
		// figure out which point in returned edge is new
		Point sPoint = results[j].firstPoint;
		Point ePoint = results[j].secondPoint;
                
		// make a new solution, just like the current one, but
		// with the new points added
		QuerySolution newSolution = new QuerySolution(currentSolution);

                if (sPoint.labelMatches(sPointLabels) && ePoint.labelMatches(ePointLabels) && 
                    newSolution.addPoint(con.myTargetEdgeStartPointName, sPoint) && newSolution.addPoint(con.myTargetEdgeEndPointName, ePoint)) {
                        finalSolutions.add(newSolution);
                }
	    }
	    count++;
	    if (count%1000==0)
		OutputWriter.printDebugDetail(".");
	}
	return finalSolutions;
    }
}


