package ShapeQuery;
// PointIndexLookupRule.java: perform index searches to add points to
//                            a solution set

/*
 * For two edges, if three out of the four points are known (already exist in 
 * the intermediate solution) this class performs an index lookup to find the
 * fourth point. This class uses <Point_ID, Angle> and <Point_ID, Distance> 
 * indices to perform lookups.
 *
 *
 */


import java.util.*;  // for vector
import java.io.*;    // for reading files
import java.net.*;

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;


class PointIndexLookupRule extends ShapeQueryRule
{
    public boolean consReversed; //flag indicating if this rule was built from a reversed constraint. For debugging purposes.
    // information extracted from the constraint on what this rule
    // means
    String myNewPointName;
    String myRelativePointName;
    String myRelativeEdgeStartPointName;
    String myRelativeEdgeEndPointName;
    int myRuleType;
    float myFactor;
    float myDifference;
    float myTolerance;
    boolean myToleranceIsAbsolute;
    
    Constraint con;


    PointIndexLookupRule(String _newPointName,
			 String _relativePointName,
			 Constraint constraint)
    {
	myNewPointName = _newPointName;
	myRelativePointName = _relativePointName;
        
        myRelativeEdgeStartPointName = constraint.myRequiredEdgeStartPointName;
	myRelativeEdgeEndPointName = constraint.myRequiredEdgeEndPointName;
	myRuleType = constraint.myRuleType;
	myFactor = constraint.myFactor;
	myDifference = (constraint.reversed) ? -constraint.myDifference : constraint.myDifference;
	myTolerance = constraint.myTolerance;
	myToleranceIsAbsolute = constraint.myToleranceIsAbsolute;
        consReversed = constraint.reversed;
        
        con = constraint.copy();
    }


    void print()
    {
	OutputWriter.println("> PointIndex lookup rule for point: "
			   +myNewPointName + "                " + getDescription());
    }

    String getDescription() {
        String s = "PointIndexLookupRule - ";
        switch(myRuleType)
	{
	    case Constraint.RULE_TYPE_DISTANCE:
                s += "Distance Lookup: "; break;
            case Constraint.RULE_TYPE_ANGLE:
                s += "Angle Lookup: "; break;
            default:
                s += "Unknown Ruletype! "; break;
        }
        s += "RelEdge:(" + myRelativeEdgeStartPointName + "," + myRelativeEdgeEndPointName + ") ";
        s += "RelPoint:" + myRelativePointName + " ";
        s += "NewPoint:" + myNewPointName + " ";
        s += "Factor=" + myFactor + " Diff=" + myDifference + " Tol=" + myTolerance + " Abs=" + myToleranceIsAbsolute + " ";
        s += "Reversed=" + consReversed;
        return s;
        
    }

    float estimateCost(PointUniverse universe, float rows)
    {
	// For each row in a set, we need to do an index lookup, which
	// is on the expensive side.  This perhaps overblows the cost
	// of an index lookup -- it'd probably be better modeled as
	// the log of the average rowcount instead.

        //corrected to incorporate the work done while searching through the index - Pranay
        //float cost = rows*universe.averageRowCount();
        //float cost = rows*(float)Math.log(universe.averageRowCount());
        
        float cost = (float)rows*(float)(Math.log((double)universe.averageRowCount())/Math.log((double)2));
        
        return cost;
    }

    // the number of rows returned is calculated for angles based on
    // arithmetic (assuming angles are evenly distributed 0-360) or,
    // for distances, by performing a lookup on the median pointpair
    // in each pointset and averaging those results
    float estimateReturnedRowMultiplier(PointUniverse universe)
    {
	float rowCount = 0f;

	switch(myRuleType)
	{
	    case Constraint.RULE_TYPE_DISTANCE:
	    {
		// find rowCount for each median edge of each pointset;
		// return average of those

		if (universe.pointSets.size()!=0)
		{
		    Iterator iter = universe.pointSets.iterator();
		    while (iter.hasNext())
		    {
			PointSet set = (PointSet) iter.next();
		    
			rowCount += (set.getIndex(IndexType.ID_DISTANCE)).estimateReturnedRows(
			    myFactor,
			    myDifference,
			    myTolerance,
			    myToleranceIsAbsolute
			    );
		    }
		    rowCount = rowCount / universe.pointSets.size();
		}

		break;
	    }
	    case Constraint.RULE_TYPE_ANGLE:
	    {
		// find rowCount for each median edge of each pointset;
		// return average of those

		if (universe.pointSets.size()!=0)
		{
		    // look only at first pointset, since they'll all
		    // have the same result
		    PointSet set = (PointSet) universe.pointSets.firstElement();
		    
		    rowCount = (set.getIndex(IndexType.ID_ANGLE)).estimateReturnedRows(
			myFactor,
			myDifference,
			myTolerance,
			myToleranceIsAbsolute
			);
		}
		break;
	    }
	    default:
	    {
		OutputWriter.println("don't know what index to use");
		break;
	    }
	}
        
        
        //now, reduce the rowcount by the label reduction factor - DJH
        rowCount *= computeLabelReduction(con.getAllLabels(), universe);
        
	return rowCount;
    }

    

    // use 'solutions' to build new QuerySolutionSet, which contains
    // all valid extensions of members of 'solutions'
    QuerySolutionSet apply(QuerySolutionSet<QuerySolution> solutions,
			   PointSet pointSet)

    {
	OutputWriter.printDebugDetail("  applying indexlookup rule for point: "+myNewPointName);
	
	// !!! is this where the program sucks up a ton of memory???

	// we'll copy from initial solutions to final solutions all
	// solutions that match (into which we add an additional point)
	Iterator iter = solutions.iterator(); // initial solutions
	QuerySolutionSet<QuerySolution> finalSolutions = new QuerySolutionSet<QuerySolution>(solutions.getEdgesToDraw());  // final solns
	
	
	int count = 0;
	while (iter.hasNext())
	{
	    // look at next solution
	    QuerySolution currentSolution = (QuerySolution)iter.next();
	    
	    int relativePointId;  // the point id we're using in the
				  // index search
	    relativePointId = currentSolution.getPoint(myRelativePointName).getId();

	    // -------------------------------------------------
	    // figure out what angle/distance we're looking for

	    // the measure we're looking for
	    float targetMeasure = 0.0f;
            PointPair relativeEdge = null;
	    if (myRelativeEdgeStartPointName!="")
	    {
		// this is a relative constraint; construct the relative edge
		relativeEdge = new PointPair(
		    currentSolution.getPoint(myRelativeEdgeStartPointName),
		    currentSolution.getPoint(myRelativeEdgeEndPointName));

		switch(myRuleType)
		{
		    case Constraint.RULE_TYPE_DISTANCE:
		    {
			targetMeasure = relativeEdge.distance;
			break;
		    }
		    case Constraint.RULE_TYPE_ANGLE:
		    {
			targetMeasure = relativeEdge.angle;
			break;
		    }
		}

		// calculate the target based on the multiplier and difference
		targetMeasure = myFactor * targetMeasure + myDifference;

	    }
	    else
	    {
		// an absolute constraint; 'difference' is our target
		// measure
		targetMeasure = myDifference;
	    }

	    // -------------------------------------------------
	    // apply tolerance to find max and min
	    float lowMeasure, highMeasure;
	    if (myToleranceIsAbsolute)
	    {
		lowMeasure = targetMeasure - myTolerance;
		highMeasure = targetMeasure + myTolerance;
	    }
	    else
	    {
		lowMeasure = targetMeasure*(1-myTolerance/100.0f);
		highMeasure = targetMeasure*(1+myTolerance/100.0f);
	    }

	    // -------------------------------------------------
	    // perform index lookup

	    // a list of the pointpairs that match this index search
	    PointPair results[];

            IndexType type = null;
	    switch(myRuleType)
	    {
	        case Constraint.RULE_TYPE_DISTANCE:
		{
                    //Commented as for Debugging Only - Pranay
                    //OutputWriter.print(" D:" + relativePointId + ":" + (int)lowMeasure + ":" + (int)highMeasure + " ");
		    type = IndexType.ID_DISTANCE;
		    break;
		}
	        case Constraint.RULE_TYPE_ANGLE:
		{
                    //Commented as for Debugging Only - Pranay
                    //OutputWriter.print(" A:" + relativePointId + ":" + (int)ShapeQuery.angleAdjust(lowMeasure) + ":" + (int)ShapeQuery.angleAdjust(highMeasure) + " ");
		    type = IndexType.ID_ANGLE;
		    break;
		}
	        default:
		{
		    OutputWriter.println("don't know what index to use");
		    results = null;
		    break;
		}
	    }
            
            Calendar startTime;
            Calendar finishTime;
            startTime = Calendar.getInstance();

            //results = (pointSet.getIndex(type)).find(relativePointId, lowMeasure, highMeasure);
            
            //Call the CssTree Search instead of the Index Binary Search - Pranay
            results = (pointSet.getCssTree(type)).find(relativePointId, lowMeasure, highMeasure);

            finishTime = Calendar.getInstance();
            long cssTreeBuildTime = finishTime.getTimeInMillis() - startTime.getTimeInMillis();
            //OutputWriter.println("\n===> PointIndexLookUp time: " + cssTreeBuildTime + " milliseconds for " + results.length + " results.");

            
            //output debugging stuff to the log 
            //USEFUL WHEN TRYING TO UNDERSTAND THE INDEX LOOKUPS - DJH
            /*
            Point pnt1 = currentSolution.getPoint(myRelativeEdgeStartPointName);
            Point pnt2 = currentSolution.getPoint(myRelativeEdgeEndPointName);    
            OutputWriter.print(" Edge in CurrSol: [" + myRelativeEdgeStartPointName + "," + pnt1.getExternalID() + "]" + 
                "->[" + myRelativeEdgeEndPointName + "," + pnt2.getExternalID() + "](" + relativeEdge.angle + ")   Searched in: " + myRelativePointName);
            for(int i = 0; i < results.length; i++) {
                OutputWriter.print(" Found: [" + results[i].firstPoint.getExternalID() + "]->[" + results[i].secondPoint.getExternalID() + "](" + results[i].angle + ")");
            }
            OutputWriter.print("\n");
            */
            
            //*/

	    // -------------------------------------------------
	    // examine results of index search
            
            //LABEL STUFF - DJH
            //first, figure out which set of labels must be matched
            //myNewPointName is either the constraint's targetEndPoint or the targetStartPoint
            Vector<String> newPointLabels = null;
            if (myNewPointName.equals(con.myTargetEdgeEndPointName)) {
                newPointLabels = con.myTargetEdgeEndPointLabels;
            }
            else if (myNewPointName.equals(con.myTargetEdgeStartPointName)) {
                newPointLabels = con.myTargetEdgeStartPointLabels;
            }
            else {
                OutputWriter.printerror("PointIndexLookupRule - For some inexplicable reason, myNewPointName is not one of the target edge endpoints!");
            }
           

            
	    // for each returned match that fits current solution...
	    for (int j = 0; j < results.length; j++)
	    {
		// figure out which point in returned edge is new
		Point newPoint = results[j].firstPoint;
		if (results[j].getFirstId() == relativePointId)
		    newPoint = results[j].secondPoint;
                
		// make a new solution, just like the current one, but
		// with newPoint added
		QuerySolution newSolution = new QuerySolution(currentSolution);

		// only accept this solution if the labels on the new point match required labels 
                // AND if it doesn't duplicate any points that we already have
		if (newPoint.labelMatches(newPointLabels) && newSolution.addPoint(myNewPointName, newPoint))
		{
		    // add this new solution to the set we'll return
		    finalSolutions.add(newSolution);
		}
	    }
	    count++;
	    if (count%1000==0)
		OutputWriter.printDebugDetail(".");
	}
	return finalSolutions;
    }
}


