package ShapeQuery;
// PointPairIndex.java: Represents an index structure composed of
//                       pointpair nodes.
//
// The aim is to be general-purpose by not constraining which fields
// are indexed, or in what order they're indexed.
//
// However, to be completely general would have added a lot of
// complexity and the general cases wouldn't have been used, so there
// are some constraints on what will work.
//
// Currently only fully supports the following type of two-dimensional
// index:
//   a. an integer
//   b. a float
//
// A search must specify an integer and a lower and upper bound for a
// float.  Any node whose integer matches exactly the query integer
// exactly, and whose float is within that range, will be considered a
// match.

import java.util.*;  // for vector
import java.io.*;    // for reading files

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;







class PointPairIndex
{
    private PointPairNode nodes[];
    private int indexOrder[];

    // create this index based on indexOrder using these nodes.
    // will sort 'nodes' according to indexOrder.
    PointPairIndex(Vector pointPairs, int newIndexOrder[])
    {
	indexOrder = newIndexOrder;

	Calendar startTime;
	Calendar finishTime;

	OutputWriter.print("      Building node array...");
	startTime = Calendar.getInstance();

	// --------------------------------------------------------
	// 1. make 'pointPairs' into nodes and insert into 'nodes'

	// allocate enough room for the nodes
	nodes = new PointPairNode[pointPairs.size()];

	int nextNodePosition = 0;  // where we'll insert the next node
				   // from the list

	// iterate through pairs to form nodes and insert them
	Iterator pairIterator = pointPairs.iterator();
	while (pairIterator.hasNext())
	{
	    // make a node out of this pair
	    PointPairNode node = new PointPairNode(
		(PointPair)(pairIterator.next()),
		indexOrder);

	    // insert this node
	    nodes[nextNodePosition] = node;

	    // get ready to insert the next node
	    nextNodePosition++;
	}

       	OutputWriter.println("length of the index is " + nodes.length);
	finishTime = Calendar.getInstance();
	OutputWriter.println("      ===> node array build time: "
			   +(finishTime.getTimeInMillis()-
			     startTime.getTimeInMillis())
			   +" milliseconds");


	// --------------------------------------------------------
	// 2. sort our nodes

	OutputWriter.println("      Sorting nodes...");
	startTime = Calendar.getInstance();
	Arrays.sort(nodes);
	finishTime = Calendar.getInstance();
	OutputWriter.println("      ===> node index sort time: "
			   +(finishTime.getTimeInMillis()-
			     startTime.getTimeInMillis())
			   +" milliseconds");

// 	// --------------------------------------------------------

    }

    /**
     * Tells you the size of the Index, as in the no. of PointPairs indexed - Pranay
     */
    public int size(){
        return nodes.length;
    }

    /**
     * Returns the PointPairNode Object asked for - Pranay
     */
    public PointPairNode getPointPairNode(int i){
        return  nodes[i];
    }

    
    
    /**
     * Prints out the details of this index for testing.
     */
    public void print(){
        String s = "";
        for(int i=0; i < nodes.length; i++) {
            s += nodes[i].getDescription();
        }
        OutputWriter.println(s);
    }

    // get elements whose point ids match 'id' and whose measures are
    // between 'lowMeasure' and 'highMeasure'.  !!! this assumes that
    // every index has an id as the first parameter and an angle or
    // distance as the second
    PointPair[] find(int matchArgument,
		     float lowMeasure, float highMeasure)
    {
	Vector<PointPair> matchingPointPairs = new Vector<PointPair>();

	// To correctly accomodate angle wrapping, we need to deal
	// differently with angles.  For starters, we move angles to
	// range -180 < n <= 180.  (This may make lowMeasure be
	// greater than highMeasure, in which case we need to compare
	// angles with x axis.)  See ShapeQuery.java for more on this.
	if (indexOrder[1]==IndexType.ANGLE)
	{
	    // first perform adjustment to get angles in right range

	    // if lower bound = -180, that means, "find anything >
	    // -180".  this might cause us to miss a valid match if
	    // the angle is exactly 180.0f, since a 180-degree angle
	    // is stored only at the end of the list, not at the
	    // beginning !!!
	    if (lowMeasure!=-180.0f)
		lowMeasure = ShapeQuery.angleAdjust(lowMeasure);
 	    highMeasure = ShapeQuery.angleAdjust(highMeasure);

// 	    if (lowMeasure > 160.0f || highMeasure < -160.0f)
// 		OutputWriter.println("angle index search, pointid="
// 				   +matchArgument+", low="
// 				   +lowMeasure+", high="
// 				   +highMeasure);
	
	    // idea for future:
	    //
	    // we can cut out a bit of work if there's an angle
	    // wrapping situation, because in order for there to be
	    // any matches, either the first or last nodes (or more
	    // likely, both) of this point id must match this search
	    //
	    // for now:
	    //
	    // just do two lookups, one for lowmeasure < n <= 180, and
	    // the other for -180 < n < highmeasure
	    if (lowMeasure > highMeasure)
	    {
		// perform this search with two lookups

//  		OutputWriter.println("wrapped angle index search, pointid="
//  				   +matchArgument+", low="
//  				   +lowMeasure+", high="
//  				   +highMeasure);

		PointPair[] lowLookup;
		PointPair[] highLookup;

		lowLookup = find(matchArgument,
				 lowMeasure, 180.0f);

		highLookup = find(matchArgument,
				  -180.0f, highMeasure);

		PointPair[] fullResult = new PointPair[lowLookup.length+
						       highLookup.length];

		for (int i = 0; i < lowLookup.length; i++)
		{
		    fullResult[i] = lowLookup[i];
		}

		for (int i = 0; i < highLookup.length; i++)
		{
		    fullResult[i+lowLookup.length]=highLookup[i];
		}

		/*
		OutputWriter.print("low list: ");
		for (int i = 0; i < lowLookup.length; i++)
		{
		    lowLookup[i].print();
		}
		OutputWriter.println();

		OutputWriter.print("high list: ");
		for (int i = 0; i < highLookup.length; i++)
		{
		    highLookup[i].print();
		}
		OutputWriter.println();

		OutputWriter.print("complete list: ");
		for (int i = 0; i < fullResult.length; i++)
		{
		    fullResult[i].print();
		}
		OutputWriter.println();
		*/

		return fullResult;


		// ideas for better solution:
		// 
		// do binary search for a node in this point set with
		// measure > -180; keep looking until we find a node in
		// the previous point set so we know we have first node

		// test if this node matches; if so, take all nodes to
		// right until we find one that doesn't match

		// next, do binary search for a node in this point set
		// with measure <=180; keep looking until we find a
		// node in the next point set so we know we have last
		// node

		// test if this node matches; if so, take all nodes to
		// left until we find one that doesn't match

		// return the resulting list of pointpairs
	    }
	}

	// use binary search to find ANY node that involves point 'id'
	// and that's between 'lowMeasure' and 'highMeasure'
        
        float[][] range = new float[indexOrder.length][2];
        range[0][0] = range[0][1] = matchArgument;
        range[1][0] = lowMeasure;
        range[1][1] = highMeasure;

	// Index of the first and last nodes (in the sequence) we've
	// checked.  Initially, we haven't checked any, so we say -1
	// (the node before the first) and nodes.length (the node
	// after the last, index-wise).
	int firstChecked = -1;
	int lastChecked = nodes.length;
	int currentNodeToCheck;

	int matchingPointPair = -1; // a node that matches the query

// 	OutputWriter.println("    ***binary search for matching node...");
// 	OutputWriter.println("    ***list size:"+lastChecked);

	while (firstChecked+1<lastChecked)
	{
	    // now we'll examine the node midway between the first and
	    // the last nodes in the sequence that we've checked
	    currentNodeToCheck = (firstChecked + lastChecked)/2;

            if (nodes[currentNodeToCheck].compareTo(range) > 0)
	    {
		// need to move to left in node array, because we're
		// either on a point id that's too high, or we're on
		// the correct point id but the measure is too high.

		lastChecked = currentNodeToCheck;
	    }
            else if (nodes[currentNodeToCheck].compareTo(range) < 0)
	    {
		// need to move to right in node array, because we're
		// either on a point id that's too low, or we're on
		// the correct point id but the measure is too low.

		firstChecked = currentNodeToCheck;
	    }
	    else
	    {
		matchingPointPair = currentNodeToCheck;
		break;
	    }
	}


    return collectMatches(matchingPointPair, range);

    }


    
    
 

    /** 
     * This is a working general find method, like the one Vespe began but never 
     * finished. This find uses binary search to find all nodes that satisfy 
     * the given search criteria. The search criteria are given in the 2D 
     * 'range' array. 
     *
     * In the first dimension,
     * the array must be the same length as indexOrder. In the second dimension,
     * the array must be of length 2. Each range[i][0] represents a lower bound
     * on the range of acceptable values for sort key of indexOrder[i]. Likewise
     * range[i][1] is an upper bound on the range of acceptable values for sort 
     * key of indexOrder[i].
     *
     * For example, if the index is sorted on <firstID, Distance>, and if range[0][0] = 3,
     * range[0][1] = 3, range[1][0] = .5, range[1][1] = .75, then we are searching 
     * for all nodes whose first point's id is 3 and whose distances lie in the range
     * .5 to .75 (inclusive). 
     *
     * Note that this method does not address the problem of angle wrapping. This 
     * must be handled externally.
     *
     * @param range 2D float array as described above.
     * @return array of matching nodes.
     */
    PointPair[] find(float[][] range)
    {
	Vector<PointPair> matchingPointPairs = new Vector<PointPair>();

        //Initialize binary search variables
	int firstChecked = -1;
	int lastChecked = nodes.length;
	int currentNodeToCheck;

	int matchingPointPair = -1; // a node that matches the query


	while (firstChecked+1<lastChecked)
	{
	    // check the midpoint
	    currentNodeToCheck = (firstChecked + lastChecked)/2;

            if (nodes[currentNodeToCheck].compareTo(range) > 0)
	    {
                //move left, we're too far to the right in the array
		lastChecked = currentNodeToCheck;
	    }
            else if (nodes[currentNodeToCheck].compareTo(range) < 0)
	    {
                //move right, we're too far to the left in the array
		firstChecked = currentNodeToCheck;
	    }
	    else
	    {
                //we are on a matching node
		matchingPointPair = currentNodeToCheck;
		break;
	    }
	}

    return collectMatches(matchingPointPair, range);
    }
   
    
    
    /**
     * Given an index to start from in the node array, and criteria for 
     * determining if the node is a match, collects all nodes to the left and to 
     * the right that satisfy the criteria. This is called after some search 
     * (binary or otherwise) has found an intial match in the sorted index array.
     *
     * @param match The index of the node to start from.
     * @param range Two dimensional array of low and high values that determine
     * if a node 'matches.'
     * @return An array of nodes that match. 
     *
     */
    public PointPair[] collectMatches(int match, float[][] range) {
       	Vector<PointPair> matchingPointPairs = new Vector<PointPair>();

        if (match >= 0 && match < nodes.length) {
            // take all matching nodes >= the node we found
            int offset = 0;
            while (match + offset < nodes.length && nodes[match+offset].compareTo(range) == 0)
            {
                matchingPointPairs.addElement(nodes[match+offset].getPointPair());
                offset++;
            }

            // take all matching nodes < the node we found
            offset = -1;
            while (match + offset >= 0 && nodes[match+offset].compareTo(range) == 0)
            {
                matchingPointPairs.addElement(nodes[match+offset].getPointPair());
                offset--;
            }
        }

	PointPair[] matchingPointPairArray =
	    new PointPair[matchingPointPairs.size()];
	
	matchingPointPairArray = (PointPair[])(matchingPointPairs.toArray(matchingPointPairArray));

	return matchingPointPairArray;
    } 
    
    
    
    public int[] getIndexOrder() {
        return indexOrder;
    }
    

    /*
     * This function finds the right node with the leaf node of a CssTree.
     */
    public int leafSearch64(int p, float[][] range) {
        
    int leaf=8;
    try {
        if (nodes[p+3].compareTo(range)>=0) {
            if (nodes[p+1].compareTo(range)>=0) {
                if (nodes[p+0].compareTo(range)>=0)
                    leaf=0;
                else
                    leaf=1;
            } else {
                if (nodes[p+2].compareTo(range)>=0)
                    leaf=2;
                else
                    leaf=3;
            }
        } else {
            if (nodes[p+5].compareTo(range)>=0) {
                if (nodes[p+4].compareTo(range)>=0)
                    leaf=4;
                else
                    leaf=5;
            } else {
                if (nodes[p+6].compareTo(range)>=0)
                    leaf=6;
                else if (nodes[p+7].compareTo(range)>=0)
                    leaf=7;
                else
                    leaf=8;
            }
        }
    }
    catch(ArrayIndexOutOfBoundsException e) {
        return 8;
    }
    
    return leaf;
    }


    /*
     * This function finds the right node with the leaf node of a CssTree.
     */
    public int leafSearch128(int p, float[][] range) {
        
    int leaf=16;
 
    try {
        if (nodes[p+7].compareTo(range)>=0) {
            if (nodes[p+3].compareTo(range)>=0) {
                if (nodes[p+1].compareTo(range)>=0) {
                    if (nodes[p+0].compareTo(range)>=0)
                        leaf=0;
                    else
                        leaf=1;
                } else {  //nodes[p+1]<key
                    if (nodes[p+2].compareTo(range)>=0)
                        leaf=2;
                    else
                        leaf=3;
                }
            } else { // nodes[p+3]<key
                if (nodes[p+5].compareTo(range)>=0) {
                    if (nodes[p+4].compareTo(range)>=0)
                        leaf=4;
                    else
                        leaf=5;
                } else {
                    if (nodes[p+6].compareTo(range)>=0)
                        leaf=6;
                    else
                        leaf=7;
                }
            }
        } else { // nodes[p+7]<key
            if (nodes[p+11].compareTo(range)>=0) {
                if (nodes[p+9].compareTo(range)>=0) {
                    if (nodes[p+8].compareTo(range)>=0)
                        leaf=8;
                    else
                        leaf=9;
                } else {  // nodes[p+9]<key
                    if (nodes[p+10].compareTo(range)>=0)
                        leaf=10;
                    else
                        leaf=11;
                }
            } else { // nodes[p+11]<key
                if (nodes[p+13].compareTo(range)>=0) {
                    if (nodes[p+12].compareTo(range)>=0)
                        leaf=12;
                    else
                        leaf=13;
                } else {
                    if (nodes[p+14].compareTo(range)>=0)
                        leaf=14;
                    else {
                        if (nodes[p+15].compareTo(range)>=0)
                            leaf=15;
                        else
                            leaf=16;
                    }
                }
            }
        }
    }
    catch(ArrayIndexOutOfBoundsException e) {
        return 16;
    }
 
    return leaf;
    }

    
    //THIS MUST BE MODIFIED TO ACCOMODATE THE TWO NEW INDEX TYPES! -DJH
    // MODIFIED TO INCORPORATE ALL INDICES CORRECTLY --- Pranay
    // estimate how many rows we'll get back -- used for cost analysis
    int estimateReturnedRows(float factor, float difference,
			     float tolerance, boolean toleranceIsAbsolute)
    {
	int numRows = 0;

	// based on what kind of index we're dealing with, figure out
	// how we'll estimate how many rows we get back

	//OutputWriter.print(" * estimating index search cost...  ");

      	if (indexOrder[0]==IndexType.FIRSTID && indexOrder[1]==IndexType.ANGLE)
	{
	    // assume this is an absolute measure of tolerance
	    numRows = (int) Math.ceil((2*tolerance/360.0f) * Math.sqrt(nodes.length));

	    //OutputWriter.print(", "+numRows+" expected based on angle arithmetic of ID_ANGLE INDEX.");
	}
	else if (indexOrder[0]==IndexType.FIRSTID && indexOrder[1]==IndexType.DISTANCE)
	{
	    // perform distance estimate by trying search assuming
	    // median edge, and see what proportion of results we get
	    // back for that 

            //Pick any Random Node instead of medianPairNode. May want to do it across 3 such instances.
            for(int i=0;i<3;i++) {
                
                int index = (int)(Math.random()*nodes.length);
                PointPairNode medianPairNode = nodes[index];
                float targetMeasure = medianPairNode.getPointPair().distance * factor;

                float lowMeasure, highMeasure;

                if (toleranceIsAbsolute)
                {
                    lowMeasure = targetMeasure - tolerance;
                    highMeasure = targetMeasure + tolerance;
                }
                else
                {
                    lowMeasure = targetMeasure * (1- tolerance/100.0f);
                    highMeasure = targetMeasure * (1+ tolerance/100.0f);
                }

                PointPair[] medianResult = find((int)medianPairNode.getArgument(0), lowMeasure, highMeasure);

                numRows += medianResult.length;

                //OutputWriter.print(", "+numRows+" expected based on median pointpair examination ID_DISTANCE INDEX.");
            }
            numRows=numRows/3;
	}
        else if (indexOrder[0]==IndexType.ANGLE && indexOrder[1]==IndexType.DISTANCE)
	{
	    // assume this is an absolute measure of tolerance
	    numRows = (int) Math.ceil((2*tolerance/360.0f) * nodes.length);
            
	    //OutputWriter.print(", "+numRows+" expected based on angle arithmetic ANGLE_DISTANCE INDEX.");
	}
	else if (indexOrder[0]==IndexType.DISTANCE && indexOrder[1]==IndexType.ANGLE)
	{
	    // perform distance estimate by trying search assuming
	    // median edge, and see what proportion of results we get
	    // back for that. CHANGED TO RANDOM EDGES INSTEAD OF MEDIAN

	    //PointPairNode medianPairNode = nodes[nodes.length/2];
            //Pick any 3 (or k) Random Nodes instead of medianPairNode
            for(int i=0;i<3;i++) {
                
                int index = (int)(Math.random()*nodes.length);
                PointPairNode randomPairNode = nodes[index];
                float targetMeasure = randomPairNode.getPointPair().distance * factor;

                float lowMeasure, highMeasure;

                if (toleranceIsAbsolute)
                {
                    lowMeasure = targetMeasure - tolerance;
                    highMeasure = targetMeasure + tolerance;
                }
                else
                {
                    lowMeasure = targetMeasure * (1- tolerance/100.0f);
                    highMeasure = targetMeasure * (1+ tolerance/100.0f);
                }

                
            
                //Statement below was changed into the new "find()" function to search via the 2 new indices - Pranay
                //PointPair[] medianResult = find((int)medianPairNode.getArgument(0),lowMeasure, highMeasure);
                float[][] range = new float[2][2];
                range[0][0]=lowMeasure;             range[0][1]=highMeasure;
                range[1][0]=-180.0f;                   range[1][1]=180.0f;        //Any Angle
                PointPair[] randomResult = find(range);

                numRows += randomResult.length;

                //OutputWriter.print(", "+numRows+" expected based on random pointpair examination ID_DISTANCE INDEX.");
            }
            numRows=numRows/3;
	}

	return numRows;
    }

    // perform a band join by joining the current index with itself;
    // use rules as described in 'constraint'
    QuerySolutionSet bandJoin(Constraint constraint, HashMap edgesToDraw)
    {
	// ----------------------------------------------------------------
	// verbal example of what's going on

	// have three pointers -- one to current candidate first edge
	// (CCFE), one to current candidate second edge (CCSE), and
	// one to minimum second edge (MSE).  here's a picture:

	// query is: "edge 2 shares a point with edge 1; the length of
	// edge 2 is twice the length of edge 1, plus or minus 3 units"

	// here is what the index looks like
	//     first-point   distance  second-point      label   
	//     1             10	       2              	   A     
	//     1             11	       3              	   B     
	//     1             17	       4              	   C     
	//     1             21	       5              	   D     
	//     1             24	       6              	   E     
	//     2             10	       1                   F     
	//     ...

	// step:   CCFE    CCSE    MSE    action:
	// 1       A       A       A      no good; too low to match constraint  (10*2-3 > 10)
	//                                 => advance CCSE, MSE
	// 2       A       B       B      no good; too low to match constraint  (10*2-3 > 11)
	//                                 => advance CCSE, MSE
	// 3       A       C       C      good; matches constraint              (10*2-3 <= 17 <= 10*2+3)
	//                                 => record this match, advance CCSE
	// 4       A       D       C      good; matches constraint              (10*2-3 <= 21 <= 10*2+3)
	//                                 => record this match, advance CCSE
	// 5       A       E       C      no good; too high to match constraint (10*2-3 <= 24 <= 10*2+3)
	//                                 => set CCSE = MSE; increment CCFE
	// 6       B       C       C      no good; too low to match constraint  (11*2-3 > 17)
	//                                 => advance CCSE, MSE
	// 7       B       D       D      good; matches constraint              (11*2-3 <= 21 <= 11*2+3)
	//                                 => record this match, advance CCSE
	// 8       B       E       D      good; matches constraint              (11*2-3 <= 24 <= 11*2+3)
	//                                 => record this match, advance CCSE
	// 9       B       F       D      no good; wrong first-point
	//                                 => set CCSE = MSE; increment CCFE
	// 10      C       D       D      no good; too low to match constraint  (17*2-3 > 21)
	//                                 => advance CCSE, MSE
	// 11      C       E       E      no good; too low to match constraint  (17*2-3 > 24)
	//                                 => advance CCSE, MSE
	// 12      C       F       F      no good; wrong first-point and all possible
	//                                matches so far were too low
	//                                 => advance CCFE, CCSE, MSE to next first-point
	// 13      F       F       F      ...

	// end verbal example
	// ----------------------------------------------------------------

	QuerySolutionSet<QuerySolution> results = new QuerySolutionSet<QuerySolution>(edgesToDraw);

	// variables to use while iterating
	int currentCandidateFirstEdgePointer = 0;
	int currentCandidateSecondEdgePointer = 0;
	int minimumSecondEdgePointer = 0;

	boolean doPrint = false;

	int count=0;

	// iterate until we've looked at every node
	while (currentCandidateFirstEdgePointer < nodes.length)
	{
	    count++;
// 	    if (count>10000)
// 		System.exit(0);

	    if (doPrint)
		System.out.println("ccfe: "+currentCandidateFirstEdgePointer
				   +", ccse: "+currentCandidateSecondEdgePointer
				   +", mse: "+minimumSecondEdgePointer);

	    // make sure ccse doesn't run off end of array
	    if (currentCandidateSecondEdgePointer==nodes.length)
	    {
		currentCandidateSecondEdgePointer = minimumSecondEdgePointer;
		currentCandidateFirstEdgePointer++;
		continue;
	    }
			       
	    PointPairNode cCFE = nodes[currentCandidateFirstEdgePointer];
	    PointPairNode cCSE = nodes[currentCandidateSecondEdgePointer];
	    PointPairNode mSE = nodes[minimumSecondEdgePointer];

	    // first points don't match, as in step 9 above
	    if (cCFE.getArgument(0) != cCSE.getArgument(0))
	    {
		if (doPrint)
		    System.out.println("=>first arg mismatch");
		if (currentCandidateSecondEdgePointer >
		    currentCandidateFirstEdgePointer)
		{
		    // ccse has run past ccfe's first point
		    currentCandidateSecondEdgePointer = minimumSecondEdgePointer;
		    currentCandidateFirstEdgePointer++;
		}
		else
		{
		    // ccfe has run past ccse -- as when ccfe changes
		    // from one first point to another first point

		    currentCandidateSecondEdgePointer = currentCandidateFirstEdgePointer;
		    minimumSecondEdgePointer = currentCandidateFirstEdgePointer;
		}
	    }
	    else // ccfe and ccse are for same first-point
	    {
		PointPair firstPair = cCFE.getPointPair();
		PointPair secondPair = cCSE.getPointPair();
		int match = matchesConstraint(firstPair,
					      secondPair,
					      constraint);
		if (match==-1)
		{
		    if (doPrint)
			System.out.println("=>too low");

		    // too low, e.g. steps 1 and 2 above
		    currentCandidateSecondEdgePointer++;
		    minimumSecondEdgePointer = currentCandidateSecondEdgePointer;
		}
		else if (match==1)
		{
		    if (doPrint)
			System.out.println("=>too high");
		    // too high, e.g. step 5 above
		    currentCandidateSecondEdgePointer = minimumSecondEdgePointer;
		    currentCandidateFirstEdgePointer++;
		}
		else if (match==0)
		{
		    // these edges match
		    if (doPrint)
			System.out.println("=>match");

		    if (currentCandidateFirstEdgePointer!=currentCandidateSecondEdgePointer)
		    {
			// these aren't the same edge.
			// (if they are, we ignore the match)

			QuerySolution solution = new QuerySolution();

			// !!! should figure out which edge's points
			// correspond to which point labels
			
			solution.addPoint(constraint.myRequiredEdgeEndPointName,
					  firstPair.firstPoint);
			solution.addPoint(constraint.myRequiredEdgeStartPointName,
					  firstPair.secondPoint);

			Point lastPoint = secondPair.firstPoint;
			if (secondPair.firstPoint==firstPair.firstPoint ||
			    secondPair.firstPoint==firstPair.secondPoint)
			{
			    lastPoint = secondPair.secondPoint;
			}

			solution.addPoint(constraint.myTargetEdgeEndPointName,
					  lastPoint);

			if (doPrint)
			    System.out.println(constraint.myRequiredEdgeStartPointName
					   +"=("+firstPair.firstPoint.getX()
					   +","+firstPair.firstPoint.getY()+"); "

					   +constraint.myRequiredEdgeEndPointName
					   +"=("+firstPair.secondPoint.getX()
					   +","+firstPair.secondPoint.getY()+"); "

					   +constraint.myTargetEdgeEndPointName
					   +"=("+secondPair.secondPoint.getX()
					   +","+secondPair.secondPoint.getY()+"); "

					   +"d(p1-p2)="+firstPair.distance
					   +"; d(p1-p3)="+
					   
					   (new PointPair(firstPair.firstPoint,
							  secondPair.secondPoint)).distance);

			results.add(solution);
		    }

		    currentCandidateSecondEdgePointer++;
		}
		else
		{
		    // invalid result
		    OutputWriter.println("invalid match result, aborting band join");
		    return null;
		}
	    }
	}

	return results;
    }

    // return 0 if edges match constraint, -1 if match is too low, +1
    // if too high.  this doesn't really work for angles, since "too
    // high" and "too low" can be the same thing (due to wrapping)
    static public int matchesConstraint(PointPair relative,
					PointPair target, 
					Constraint constraint)
    {
	// -------------------------------------------------
	// figure out what angle/distance we're looking for

	// the measure we're looking for
	float targetMeasure = 0.0f;

	switch(constraint.myRuleType)
	{
	    case Constraint.RULE_TYPE_DISTANCE:
	    {
		targetMeasure = relative.distance;
		break;
	    }
	    case Constraint.RULE_TYPE_ANGLE:
	    {
		targetMeasure = relative.angle;
		break;
	    }
	}

	// calculate the target based on the multiplier and difference
	targetMeasure = constraint.myFactor * targetMeasure + constraint.myDifference;


	// -------------------------------------------------
	// apply tolerance to find max and min
	float lowMeasure, highMeasure;
	if (constraint.myToleranceIsAbsolute)
	{
	    lowMeasure = targetMeasure - constraint.myTolerance;
	    highMeasure = targetMeasure + constraint.myTolerance;
	}
	else
	{
	    lowMeasure = targetMeasure*(1-constraint.myTolerance/100.0f);
	    highMeasure = targetMeasure*(1+constraint.myTolerance/100.0f);
	}

	// -------------------------------------------------
	// compare with target edge

	switch(constraint.myRuleType)
	{
	    case Constraint.RULE_TYPE_DISTANCE:
	    {
		// check if target edge doesn't meet requirement
		if (target.distance > highMeasure)
		    return 1;
		if (target.distance < lowMeasure)
		    return -1;

		// target edge must meet requirement
		return 0;
	    }

	    // to implement for angle, we need to figure out what it
	    // means to be "less than" a particular angle.
	    //
	    // e.g., is 175 "less than" -175?  sometimes it is...
/*
	    case Constraint.RULE_TYPE_ANGLE:
	    {
		// adjust measures to be in range -180 < n <= 180
		highMeasure = ShapeQuery.angleAdjust(highMeasure);
		lowMeasure = ShapeQuery.angleAdjust(lowMeasure);

		// check order of angles; if highMeasure <
		// lowMeasure, that's an angle wrapping situation
		// that we need to resolve
		if (highMeasure > lowMeasure)
		{
		    if (target.angle > highMeasure)
		    {
			return 
			target.angle < lowMeasure)
		    {
			// target edge doesn't meet requirement 
			return false;
		    }
		    // target edge must meet requirement 
		    return true;
		}
		else // not simple angle situation
		{
		    // we have an angle wrapping situation -- e.g.:
		    //
		    //     low \  |
		    //   measure\ |
		    //           \|
		    // ------------------------
		    //           /|
		    //   high   / |
		    // measure /  |
		    //
		    // (lowmeasure = 135, highmeasure = -135
		    // (adjusted from 225)

		    // test for angle around x axis
		    if ((target.angle < highMeasure &&
			 target.angle > -180.0f) ||
			(target.angle > lowMeasure &&
			 target.angle <= 180.0f))
		    {
			// matches
			return true;
		    }
		    else
		    {
			// target edge doesn't meet requirement 
			return false;
		    }
		}
	    }
*/
	} // end of switch

	OutputWriter.println("ERROR: constraint match test failed");
	return -2;
     }
}
