package ShapeQuery;
// PointPairNode.java: represents a node (containing a single point
//                     pair) in an index data structure

import java.util.*;  // for vector
import java.io.*;    // for reading files

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;


class PointPairNode implements Comparable
{
    // The types of indicies we may build
    //
    // Note that I wrote this to be general, but at a some points in
    // development I assumed a certain structure since any other
    // structure wouldn't work.  Specifically, I frequently assumed
    // that indexOrder[] is of the form {first-id, distance} or
    // {first-id, angle}.



    // Every point pair node contains a point pair and a sort order.
    // 
    // The pointpair needs to know its own sort order so that when
    // it's in an array and Arrays.sort() calls compareTo(), this node
    // knows its place.
    private PointPair pointPair;
    private int indexOrder[];


    // make a new node
    PointPairNode(PointPair newPointPair, int newIndexOrder[])
    {
	pointPair = newPointPair;
	indexOrder = newIndexOrder; // need to know our index order so
				    // we can sort with other nodes
    }


    
    
    // to determine ordering in array, so we can sort
    public int compareTo(Object o)
    {
	PointPairNode node = (PointPairNode)o;

	// iterate through each layer of index.  any time we see a
	// difference, stop looking and return the value according to
	// that difference.  if we don't see any differences, we'll
	// return 0 in the end.
	for (int i = 0; i < indexOrder.length; i++)
	{
            if (getArgument(i) < node.getArgument(i)) 
                return -1;
            if (getArgument(i) > node.getArgument(i)) 
                return 1;            
	}

	// no difference!
	return 0;
    }

    
    
    
    
    /**
     * Compares this node to values in a 2D array. In the first dimension,
     * the array must be the same length as indexOrder. In the second dimension,
     * the array must be of length 2. The comparison iterates through each 
     * layer i of the index, and returns a negative value if the value of this 
     * node is less than range[i][0], and a positive value if the value of this 
     * node is greater than range[i][1]. Returns 0 if this node's values fall 
     * between the values of range[i][0] and range[i][1] for all i.
     *
     * To do equality tests simply set range[i][0] = range[i][1] = value.
     *
     * This method allows for easy searches through the index.
     *
     * @param range A 2D array as described above.
     * @return -1 if this node is less than range, 1 if this node is greater than 
     * range. 0 otherwise.
     *
     */
    public int compareTo(float[][] range) throws IllegalArgumentException
    {
        //check that range has the right properties. dimension 1 must be as long as indexOrder
        //dimension 2 must be 2.
        if (range.length != indexOrder.length || range[0].length != 2) {
            OutputWriter.printerror("2D array 'range' is not correctly sized.");
            throw new IllegalArgumentException("2D array 'range' is not correctly sized.");
        }

	// iterate through each layer of index.  any time we see a
	// difference, stop looking and return the value according to
	// that difference.  if we don't see any differences, we'll
	// return 0 in the end.
	for (int i = 0; i < indexOrder.length; i++)
	{
            if (getArgument(i) < range[i][0]) 
                return -1;
            if (getArgument(i) > range[i][1]) 
                return 1;
	}

	// no difference!
	return 0;
    }
    
    
    
    
    
    
    float getArgument(int i)
    {
        if (i > indexOrder.length) {
            OutputWriter.printerror("No argument exists at " + i + ". Index is only sorted on " + indexOrder.length + " arguments.\n");
            return -1f;
        }
	switch (indexOrder[i])
	{
	    case IndexType.FIRSTID:
	    {
		return pointPair.getFirstId();
	    }	    
	    case IndexType.SECONDID:
	    {
		return pointPair.getSecondId();
	    }	    
	    case IndexType.DISTANCE:
	    {
		return pointPair.distance;
	    }	    
	    case IndexType.ANGLE:
	    {
		return pointPair.angle;
	    }	    
	    default:
	    {
		OutputWriter.printerror("Invalid index.\n");
		return -1f;
	    }
	}
    }
    


    // return the pointpair in this node so you can look at its
    // distance/angle etc. 
    PointPair getPointPair()
    {
	return pointPair;
    }
    
    /**
     * Returns the indexOrder - Pranay
     */    
     public int[] getIndexOrder() {
         return indexOrder;
    }

    
    
    /**
     * Returns a string containing the details of this node for testing.
     */
    public String getDescription() {
        String s ="[";
        s += (pointPair.firstPoint.hasExternalID()) ? pointPair.firstPoint.getExternalID() : pointPair.firstPoint.getX() + "," + pointPair.firstPoint.getY(); 
        s += "->";
        s += (pointPair.secondPoint.hasExternalID()) ? pointPair.secondPoint.getExternalID() : pointPair.secondPoint.getX() + "," + pointPair.secondPoint.getY(); 
        s += "]KEY(";
        s += getArgument(0);
        for (int i = 1; i < indexOrder.length; i++) {
            s += "," + getArgument(i);
        }
        s += ") ";
        return s;
    }
}
