package ShapeQuery;
// PointSet.java: A pointset is a set of points in a plane.  PointSets
//                are the most fundamental unit that can be queried
//                for shapes.


import java.util.*;  // for vector
import java.io.*;    // for reading files

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.applet.*;


class PointSet
{
    
    // pending error message
    private String errorMessage;

    // names associated with this set
    public String visibleName;

    // the points and edges in this set
    Vector<Point> points;
    Vector<PointPair> pointPairs;

    // the indicies in this set
    // DJH - I changed indices to be a single array indexed on the enum IndexTypes.
    private PointPairIndex indices[] = new PointPairIndex[IndexType.size()];

    //This is the CSS Tree Declaration - Pranay
    private CssTree cssTree[] = new CssTree[IndexType.size()];
    
    // solutions to the most recent query on this pointset
    private QuerySolutionSet<QuerySolution> solutionSet;

    // what solution we're up to in this solution set, if we're
    // currently iterating over solutions
    Iterator<QuerySolution> solutionIterator;
    
    //flag indicating whether this pointset was cached - DJH
    private boolean cached;
    private long indexTime;  //time for node creation and indexing
    private long cssTreeBuildTime; //time for building CSS Tree

    // applet-specific variables
    boolean runStandalone;  // are we running as an applet?
    URL codeBase;           // where to look for files

    // create a new set of points to be queried
    PointSet(boolean _runStandalone,
	     URL _codeBase,
	     String _visibleName)
    {
	OutputWriter.println("* making new pointset");
	points = new Vector<Point>();
	pointPairs = new Vector<PointPair>();
	runStandalone = _runStandalone;
	codeBase = _codeBase;
	visibleName = _visibleName;

	solutionIterator = null;
	solutionSet = null;
        
        cached = false;
        indexTime = 0;
        cssTreeBuildTime=0;
    }

    // return pending error
    String getError()
    {
	return errorMessage;
    }

    /*
     * Get the name associated with this PointSet.
     * @return The name of the PointSet.
     */
/*    public String getName() {
        return visibleName;
    }
 */
    
    
    /**
     * Call to indicate that this PointSet was pulled from a cache.
     */
    public void flagCached() {
        cached = true;
    }
    
    /**
     * Get the time in ms that it took to create nodes and to index them.
     */
    public long getIndexTime() {
        return indexTime;
    }

    /**
     * Get the time in ms that it took to create CSS Tree on the sorted index.
     */
    public long getCssTreeBuildTime() {
        return cssTreeBuildTime;
    }

    
    /**
     * Check if this PointSet was ever pulled from a cache. Dependent on the fact that if 
     * and when it was pulled out of the cache, the programmer called flagCached() method.
     */
    public boolean fromCache() {
        return cached;
    }
    
    
    /**
     * Gets the index sorted of the specified type residing in this PointSet. Type
     * refers to which data the index is sorted on.
     *
     * @param type A constant IndexType that specifies which index is desired.
     * @returns The index of the specifed type.
     */
    public PointPairIndex getIndex(IndexType type) {
        return indices[type.ord];
    }

    /**
     * Gets the CssTree of the index sorted of the specified type residing in this PointSet. Type
     * refers to which data the index is sorted on.
     *
     * @param type A constant IndexType that specifies which index is desired.
     * @returns The CssTree of the specifed type.
     */
    public CssTree getCssTree(IndexType type) {
        return cssTree[type.ord];
    }

    
    
    boolean loadData(Vector pts) {
        
        Point p = null;
        //put into local vector (originally point was COPIED into local vector. there seemed to be no reason for this DJH)
        for (int i=0; i<pts.size(); i++) {
            p = (Point)pts.get(i);
            points.addElement(p);
        }
        createPointPairs();
        createIndices();
        createCssTree();
        return true;
    }
    
    
    // load data from this file into this pointset.  return true if
    // successful.
    boolean loadData(String filename)
    {
	Calendar startTime;
	Calendar finishTime;

	// generic reader (regardless of standalone or applet)
	BufferedReader input = null;

	// if applet, need to do extra work
	DataInputStream url_input = null;
	URL sourceurl = null;

	if (!runStandalone)
	{
	    // as applet, open url
	    try
	    {
		sourceurl = new URL(codeBase, filename);
		url_input = new DataInputStream(sourceurl.openStream());
	    }
	    catch (java.security.AccessControlException e)
	    {
		errorMessage = "unable to open file due to security problem: "
		    +e.getMessage();
		return false;
	    }
	    catch (java.net.MalformedURLException e)
	    {
		errorMessage = "url problem: "+e.getMessage();
		return false;
	    }
	    catch (java.io.IOException e)
	    {
		errorMessage = "file problem: "+e.getMessage();
		return false;
	    }
	}
	 
	OutputWriter.print("  Reading points file ");
	startTime = Calendar.getInstance();
	// ---------------------------------------
	// open the file

	try
	{
	    if (runStandalone)
	    {
		OutputWriter.println("directly from filesystem...");
		input = new BufferedReader(new FileReader(filename));
	    }
	    else
	    {
		OutputWriter.println("from url...["+codeBase+filename+"]");
		input = new BufferedReader(new InputStreamReader(url_input));
	    }
	}
	catch (java.io.FileNotFoundException e)
	{
	    errorMessage = "file '"+filename+"'not found";
	    OutputWriter.println(errorMessage);
	    return false;
	}

	// ---------------------------------------
	// read in the file; use to populate vector of points
	try
	{
	    OutputWriter.println("  reading line-by-line");
	    while (input!=null && input.ready())
	    {
//		OutputWriter.print("  getting a line...  ");
		String line = input.readLine();
//		OutputWriter.println("  got line: ["+line+"]");

		// ignore comments
		if (line.length()==0 || line.charAt(0)=='#')
		    continue;

		// tokenize this string
		StringTokenizer tokenizer = new StringTokenizer(line,
								",");
		String x, y;
		try
		{
		    x = tokenizer.nextToken();
		    y = tokenizer.nextToken();
		}
		catch (NoSuchElementException e)
		{
		    OutputWriter.println("parse error, ignoring line: ["
			+line+"]");
		    continue;
		}


		Point point = new Point((new Float(x)).floatValue(),(new Float(y)).floatValue());

                points.addElement(point);
	    }
	    OutputWriter.println("  file read complete");
	    input.close();
	    OutputWriter.println("  file closed");
	}
	catch (java.io.IOException e)
	{
	    errorMessage = "io error: "+e.getMessage();
	    OutputWriter.println(errorMessage);
	    return false;
	}
	catch (Exception e)
	{
	    errorMessage = "general error: "+e.getMessage();
	    OutputWriter.println(errorMessage);
	    return false;
	}

	finishTime = Calendar.getInstance();

	// my ie dies on this line:
	long timeDifference = finishTime.getTimeInMillis()-
	    startTime.getTimeInMillis();

	OutputWriter.println("===> file read time: "
			     +timeDifference
			     +" milliseconds");


        createPointPairs();
        createIndices();
        createCssTree();
	return true;
    }


    /**
     * Iterates through the points and creates PointPairs. These represent
     * edges, and are used in index creation.
     *
     */
    private void createPointPairs() {
        // iterate over every pair of points and create a PointPair
	// for each
        OutputWriter.println("  Computing edges...");
	Iterator outer = points.iterator();

	while (outer.hasNext())
	{
	    Point outerPoint = (Point)(outer.next());
	    Iterator inner = points.iterator();

	    while (inner.hasNext())
	    {
		Point innerPoint = (Point)(inner.next());
		
 	        // don't consider lines from a point to itself
		if (Point.isEqual(outerPoint,innerPoint))
		{
		    continue;
		}

		PointPair p = new PointPair(outerPoint, innerPoint);
		pointPairs.addElement(p);
	    }

	}
    }

    
    /**
     * Create the the indices after a load. 
     *
     */
    private void createIndices() {
	Calendar startTime;
	Calendar finishTime;
	startTime = Calendar.getInstance();

	OutputWriter.println("  Creating indicies...");
        
        //create all the indexes defined by IndexType.java
        for (IndexType type = IndexType.first(); type != null; type = type.next()) {
            OutputWriter.println("    Creating index: " + type.toString());
            indices[type.ord] = new PointPairIndex(pointPairs, type.order());
            //indices[type.ord].print();
        }
        
	finishTime = Calendar.getInstance();
        indexTime = finishTime.getTimeInMillis() - startTime.getTimeInMillis();
	OutputWriter.println("\n===> total index creation time: " + indexTime + " milliseconds");
    }
    

     /**
     * Create the the CSS Tree over the indices. - Pranay
     *
     */
    private void createCssTree() {
	Calendar startTime;
	Calendar finishTime;
	startTime = Calendar.getInstance();

	OutputWriter.println("  Creating CSS Tree...");
        
        //create all the indexes defined by IndexType.java
        for (IndexType type = IndexType.first(); type != null; type = type.next()) {
            OutputWriter.println("    Creating CSS Tree for Index: " + type.toString());
            cssTree[type.ord] = new CssTree(indices[type.ord]);
            //cssTree[type.ord].print();
        }
        
	finishTime = Calendar.getInstance();
        cssTreeBuildTime = finishTime.getTimeInMillis() - startTime.getTimeInMillis();
	OutputWriter.println("\n===> total CSS Tree creation time: " + cssTreeBuildTime + " milliseconds");
    }

     
    
    /**
     * Gets a string that contains SVG output that draws the points in this pointset.
     * @param mult Multiplicative factor by which to translate the x,y coordinates.
     * @param trans Additive factor by which to translate the x,y coordinates.
     * @return A string containing valid SVG output.
     */
    String getSVG(float mult, float trans) {
        String s = "<g>\n";
        Point p = null;
        for (int i = 0; i < points.size(); i++) {
            p = (Point)points.get(i);
            s += "\t<circle cx=\"" + (p.getX() * mult + trans) + "\" cy=\"" +  (p.getY() * mult + trans) + "\"";
            s += " r=\"3\" fill=\"red\"/>\n";
        }
        s += "</g>";
        return s;
    }
    
    
    /**
     * Gets the number of points in this pointset.
     * @return The number of points in this pointset.
     */
    public int getPointCount()
    {
        return points.size();
    }
    
    
    
    
    String getVisibleName()
    {
	return visibleName;
    }

    public void setSolutionSet(QuerySolutionSet<QuerySolution> _set)
    {
	solutionSet = _set;
	solutionIterator = null;
    }

    public QuerySolutionSet<QuerySolution> getSolutionSet()
    {
	return solutionSet;
    }

    // get the next solution to display in this set
    QuerySolution getNextSolution()
    {
	// check that we have some solutions to return
	if (solutionSet == null)
	{
	    return null;
	}

	// if the iterator isn't initialized or it's at the end,
	// re-initialize it
	if (solutionIterator==null || !solutionIterator.hasNext())
	{
	    solutionIterator = solutionSet.iterator();
	}

	return (QuerySolution) solutionIterator.next();
    }
}

