/*
 * PointSetContentHandler.java
 * 
 * Created on March 9, 2005, 10:47 AM
 * This class provides the event-driven methods for parsing XML pointset data.
 */
package ShapeQuery;


import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.ContentHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import java.util.Vector;
import java.util.HashMap;

/**
 * This class provides the event-driven methods for parsing XML pointset data.
 * @author David Hessing
 */

public class PointSetContentHandler extends DefaultHandler implements ContentHandler{
    
    String inElement;
    String currSetName;
    String pointID;
    float x;
    float y;
    HashMap<String, Vector<Point>> pointsets;
    boolean refVec; //indicates if the points are part of a reference vector.
    int refCount; //there must be exactly two points in a reference vector; only one such vector is allowed.
    String rawXML; //rebuild the raw XML of each pointset to use as a key when caching.
    HashMap<String, String> rawXMLs; //During parsing, the pointsets are keyed on their names. So, we key the rawXML on the names, and only later use the rawXML as keys when caching.
    Vector<String> labels;
    
    HashMap<String, Integer> labelcount; //counts the number of times each label appears accross all pointsets. Used for cost estimation when building queryplans
    
    /** Creates a new instance of PointSetContentHandler */
    public PointSetContentHandler() {
        inElement = ""; 
        currSetName = "";
        pointsets = new HashMap<String, Vector<Point>>();
        pointID = "";
        refVec = false;
        refCount = 0;
        rawXML = "";
        rawXMLs = new HashMap<String, String>();
        labels = new Vector<String>();
        labelcount = new HashMap<String, Integer>();        
    }
    

    /**
     *
     *
     * @param uri the Namespace URI, or the empty string if the
     *        element has no Namespace URI or if Namespace
     *        processing is not being performed
     * @param localName the local name (without prefix), or the
     *        empty string if Namespace processing is not being
     *        performed
     * @param qName the qualified name (with prefix), or the
     *        empty string if qualified names are not available
     * @param atts the attributes attached to the element.  If
     *        there are no attributes, it shall be an empty
     *        Attributes object.  The value of this object after
     *        startElement returns is undefined
     * @throws org.xml.sax.SAXException any SAX exception, possibly
     *            wrapping another exception
     */
    public void startElement (String uri, String localName,
			      String qName, Attributes atts) throws SAXException {
        
        String setORid = atts.getValue(0);
        inElement = qName.trim();
        rawXML += "<" + qName.trim();
        if (inElement.equalsIgnoreCase("pointset") && setORid != null) {
            currSetName = setORid;
            rawXML = "<" + qName.trim() + " name=\"" + currSetName + "\"";
            //create a vector in the hashmap with the name as the key
            pointsets.put(currSetName, new Vector<Point>());
            //reset refCount
            refCount = 0;
            
            //Delete the pointSet if it already exists in the database for statistical analysis
            //System.out.println(QueryBean.STATISTICAL_ANALAYSIS);
            if(QueryBean.STATISTICAL_ANALAYSIS) {
                QueryBean.db.deletePointSet(currSetName);
            }

            
        }
        else if (inElement.equalsIgnoreCase("point") && setORid != null) {
            pointID = setORid;
            rawXML += " id=\"" + pointID + "\"";
        }
        else if (inElement.equalsIgnoreCase("referenceVector")) {
            refVec = true;
        }
        rawXML += ">";
        //testing
        //System.out.println("STARTELEMENT - uri:" + uri + " localName:" + localName + " qName:" +qName);
        //if (setORid != null)
        //    System.out.println("Att: " + setORid);
    }
    
    /**
     *
     * @param uri the Namespace URI, or the empty string if the
     *        element has no Namespace URI or if Namespace
     *        processing is not being performed
     * @param localName the local name (without prefix), or the
     *        empty string if Namespace processing is not being
     *        performed
     * @param qName the qualified XML name (with prefix), or the
     *        empty string if qualified names are not available
     * @throws org.xml.sax.SAXException any SAX exception, possibly
     *            wrapping another exception
     */
    public void endElement (String uri, String localName,
			    String qName)
	throws SAXException {
        
        rawXML += "</" + qName.trim() + ">";
        
        if (qName.trim().equalsIgnoreCase("pointset")) {
            rawXMLs.put(currSetName, rawXML);
            currSetName = "";
            rawXML = new String();
        }
        else if (qName.trim().equalsIgnoreCase("point")) {
            //Point p = (pointID.length() > 0) ? new Point(x, y, pointID) : new Point(x, y);
            Point p = new Point(x, y, pointID, labels, refVec);
            //add the point to the vector in the hashset
            //System.out.println("Adding " + p.getDescription() + " to " + currSetName);
           
            //Add the point to the database for statistical analysis
            if(QueryBean.STATISTICAL_ANALAYSIS) {
                QueryBean.db.loadPointSet(currSetName, pointID, labels, x, y);
            }
            
            Vector<Point> v = pointsets.get(currSetName);
            v.add(p);
            
            //reset the pointID (since it's not required that all points have them)
            pointID = "";
            
            //reset the label vector
            labels = new Vector<String>();
            
            //increment the refCount if the point was a reference point in a reference vector
            if (refVec) {refCount++;}
        }
        else if (qName.trim().equalsIgnoreCase("referenceVector")) {
            //turn off the refVec flag
            refVec = false;
            
            //throw error if there were not exactly two reference points in the ref vector
            if (refCount != 2) {
                throw new SAXException("There must be exactly two points in the optional reference vector.");
            }
        }
        
        //testing
        //System.out.println("ENDELEMENT - uri:" + uri + " localName:" + localName + " qName:" +qName);
    }


    /**
     *
     * @param ch the characters from the XML document
     * @param start the start position in the array
     * @param length the number of characters to read from the array
     * @throws org.xml.sax.SAXException any SAX exception, possibly
     *            wrapping another exception
     */
    public void characters (char ch[], int start, int length)
	throws SAXException {
        
        String s = new String(ch, start, length);
        rawXML += s.trim();
        if (s.trim().length() == 0) 
            return;
        try {
            if (inElement.equalsIgnoreCase("x")) {
                x = Float.parseFloat(s);
            }
            else if (inElement.equalsIgnoreCase("y")) {
                y = Float.parseFloat(s);
            }
            else if (inElement.equalsIgnoreCase("label")) {
                labels.add(s);
                
                //keep track of how many times this label has appeared in ALL datasets
                Integer obji = new Integer(1);
                if (labelcount.containsKey(s)) {
                    obji = (Integer)labelcount.get(s);
                    obji = new Integer(obji.intValue() + 1);
                }
                labelcount.put(s, obji);
            }
        } catch (NumberFormatException nfe) {
            OutputWriter.printerror("XML data is not a number: " + nfe.getMessage());
        }
        
        
        //testing
        //System.out.println("characters:" + s.trim());
    }

    
    /**
     * Get the HashMap that holds the vectors of loaded pointsets. Call this method after parsing an XML input of datasets.
     * @return The hashmap of vectors of pointsets.
     */
    public HashMap<String, Vector<Point>> getPointSets() {
        return pointsets;
    }
    
    /**
     * Get the HashMap that holds the count of how many times each label appeared accross all pointsets.
     * This information is useful for query cost estimation.
     * Call this method after parsing an XML input of datasets. 
     * @return The hashmap of counts of each encountered label.
     */
    public HashMap<String, Integer> getLabelCount() {
        return labelcount;
    }
    
    
    /**
     * Get the HashMap that holds the raw XML for each pointset. 
     * The raw XML is used to key the pointset for caching, to safeguard against 
     * the possibility that a user will modify a pointset, but not change its name.
     * 
     * @return The hashmap of raw XML for each pointset, keyed on the pointsets' names.
     */
    public HashMap<String, String> getRawXMLs() {
        return rawXMLs;
    }
    
}
