package ShapeQuery;
// PointUniverse.java: A point universe is a set of point sets that
//                     the user executes a query over.


import java.util.*;  // for vector
import java.io.*;    // for reading files
import java.net.*;

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;

class PointUniverse
{
    Vector<PointSet> pointSets;
    boolean runStandalone;    // are we running as an applet?
    URL codeBase;            // where to look for files
    
    /** A hashmap of label counts across all pointsets in the universe.
     * Calculated during parsing of the XML files! */
    private HashMap labelcount; 

    /** Count of all points in all pointsets in this universe. */
    private int pointcount;
    
    // create a new universe of point sets over which to do queries
    PointUniverse(boolean newRunStandalone, URL newCodeBase)
    {
	OutputWriter.println("* making new point universe");
	pointSets = new Vector<PointSet>();
	runStandalone = newRunStandalone;
	codeBase = newCodeBase;
        pointcount = 0;
    }
    //default constructor for running standalone
    PointUniverse()
    {
	OutputWriter.println("* making new point universe");
	pointSets = new Vector<PointSet>();
	runStandalone = true;
        pointcount = 0;
    }

    

    // add points specified by this file to our point universe.
    // return true if successful.
    PointSet addPointSet(String pointFileName, String visibleName)
    {
	PointSet set = new PointSet(runStandalone, codeBase, visibleName);
	if (set.loadData(pointFileName))
	{
            pointcount += set.getPointCount();
	    
            pointSets.addElement(set);
	    return set;
	}

	OutputWriter.println("File load failed: "+set.getError());

	return null;
    }

    // add points specified contained in the vector to our point universe.
    // return true if successful.
    PointSet addPointSet(Vector points, String visibleName)
    {
	PointSet set = new PointSet(runStandalone, codeBase, visibleName);
	if (set.loadData(points))
	{
            pointcount += set.getPointCount();
	    pointSets.addElement(set);
	    return set;
	}

	OutputWriter.println("Vector load failed: "+set.getError());

	return null;
    }
    
    void addPointSet(PointSet set) {
        pointcount += set.getPointCount();
        pointSets.addElement(set);
    }

    // return the average number of pointpairs in all my component
    // point sets, for cost estimation
    float averageRowCount()
    {
	float totalCount = 0f;
	
	// if we don't have any point sets, don't try to divide by
	// zero -- just return that there will be nothing to search
	if (pointSets.size()==0)
	{
	    return 0;
	}

	Iterator iter = pointSets.iterator();
	
	while (iter.hasNext())
	{
	    totalCount+= ((PointSet) (iter.next())).pointPairs.size();
	}

	return totalCount/pointSets.size();
    }
    
    
    /**
     * Gets the number of datasets that are currently in this Universe.
     * 
     * @return the number of datasets in this universe.
     */
    public int getPointSetCount () {
        return pointSets.size();
    }
    
    /**
     * Gets the total time in ms spent on node creation and indexing for all pointsets in this universe.
     * PoinSets that were previously cached to not contribute, since there indices had been created previously.
     *
     * @return Time in ms it took to create nodes and to index them for all points in all pointsets that had not been cached.
     */
    public long getTotalIndexTime() {
        long t = 0;
        for (int i = 0; i < pointSets.size(); i++) {
            if (!(pointSets.get(i).fromCache())) {
                t += pointSets.get(i).getIndexTime();
            }
        }
        return t;
    }
    
    /**
     * Gets the total time in ms spent on building the CSS Tree for all indexes in this universe.
     *
     * @return Time in ms it took to build the CSS Tree for all indexes in all pointsets that had not been cached.
     */

        public long getTotalCssTreeBuildTime() {
        long t = 0;
        for (int i = 0; i < pointSets.size(); i++) {
            if (!(pointSets.get(i).fromCache())) {
                t += pointSets.get(i).getCssTreeBuildTime();
            }
        }
        return t;
    }

    
    /**
     * Set the HashMap that contains labels as keys and Integer counts of how 
     * many times the label appeared in all pointsets in this universe. Note that
     * these counts are determined during parsing of the XML datasets file. Thus it
     * is important to add pointsets and set the labelcount in a consistent way. 
     *
     * @param countedlabels HashMap of all labels in all pointsets in this universe 
     * and their counts.
     */
    public void setLabelCount(HashMap countedlabels) {
        labelcount = countedlabels;
    }

    /**
     * Computes a value x, 0 < x <= 1, for use in cost estimation when building
     * query plans. The assumption is that the presence of a label in the query
     * will reduce the query cost by a factor proportional to the number of
     * times the label appears in the pointsets.
     * 
     * @param label The label for which to compute a reductive factor.
     * @return The number of times this label appears divided by the total number 
     * of points.
     */
    public float getLabelReduction(String label) {
        Integer obji = (Integer)labelcount.get(label);
        if (obji == null) {
            return 1; //the label wasn't found
        }
        return obji.floatValue() / (float)pointcount;
    }
    
    
    /**
     * Gets the PointSet of the DataSet at the indicated index.
     * 
     * @param index the index in the DataSet vector.
     * @return the PointSet at the specified index.
     */
    public PointSet getPointSetAt(int index) {
        if (index >= pointSets.size()) {
            OutputWriter.println("PointUniverse: No Dataset at that location.");
            return null;
        }
        PointSet p = (PointSet)pointSets.get(index);
        return p;
        
    }
    
    
}
