/*
 * QueryContentHandler.java
 *
 * Created on March 9, 2005, 10:47 AM
 * This class provides the event-driven methods for parsing XML query data.
 */
package ShapeQuery;


import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.ContentHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import java.util.Vector;
import java.util.HashMap;

/**
 * This class provides the event-driven methods for parsing XML query data.
 * @author David Hessing
 */

public class QueryContentHandler extends DefaultHandler implements ContentHandler{

    String inElement;
    String edge;
    int edg1;
    int edg2;
    String[] line;
    Vector<Constraint> queryconstraints;
    String queryname;
    String removeDupes;
    private final int FIELDS = 9;
    Vector<String>[] pointlabels;
    HashMap<String, Vector<String>> prevlabels; //a hashtable of point handles and labels that were specified for that point
    //this handles the fact that queries may reference the same point handle more than once, but not
    //give the same labels for that point. The parser tracks all labels given anywhere in the query
    //and adds those labels to all constraints containing that point

    /** Creates a new instance of QueryContentHandler */
    public QueryContentHandler() {
        inElement = "";
        edge = "";
        edg1 = 0;
        edg2 = 2;
        queryname = "";
        line = new String[FIELDS];
        queryconstraints = new Vector<Constraint>();
        pointlabels = new Vector[4];
        prevlabels = new HashMap<String, Vector<String>>();
    } 


    /**
     *
     *
     * @param uri the Namespace URI, or the empty string if the
     *        element has no Namespace URI or if Namespace
     *        processing is not being performed
     * @param localName the local name (without prefix), or the
     *        empty string if Namespace processing is not being
     *        performed
     * @param qName the qualified name (with prefix), or the
     *        empty string if qualified names are not available
     * @param atts the attributes attached to the element.  If
     *        there are no attributes, it shall be an empty
     *        Attributes object.  The value of this object after
     *        startElement returns is undefined
     * @throws org.xml.sax.SAXException any SAX exception, possibly
     *            wrapping another exception
     */
    public void startElement (String uri, String localName,
			      String qName, Attributes atts) throws SAXException {

        inElement = qName.trim();

        if (inElement.equalsIgnoreCase("query")) {
            removeDupes = atts.getValue("removeduplicates");
            String name = atts.getValue("name");
            if (name == null) {
                throw new SAXException("The required 'name' attribute is missing for this query!");
            }
            queryname = name;
        }
        //if we start a new constraint reset all objects parsed out of the constraint
        if (inElement.equalsIgnoreCase("constraint")) {
            resetLine();
            resetPointLabels();
        }

        if (inElement.equalsIgnoreCase("edge1")) {
            edge = "edg1";
        }
        else if (inElement.equalsIgnoreCase("edge2")) {
            edge = "edg2";
        }


        //testing
        //System.out.println("STARTELEMENT - uri:" + uri + " localName:" + localName + " qName:" +qName);
        //if (setName != null)
        //    System.out.println("Att: " + setName);
    }

    /**
     *
     * @param uri the Namespace URI, or the empty string if the
     *        element has no Namespace URI or if Namespace
     *        processing is not being performed
     * @param localName the local name (without prefix), or the
     *        empty string if Namespace processing is not being
     *        performed
     * @param qName the qualified XML name (with prefix), or the
     *        empty string if qualified names are not available
     * @throws org.xml.sax.SAXException any SAX exception, possibly
     *            wrapping another exception
     */
    public void endElement (String uri, String localName,
			    String qName)
	throws SAXException {

        //put the line in the vector if we finished a constraint
        if (qName.trim().equalsIgnoreCase("constraint")) {
            queryconstraints.add(buildConstraint());
            edg1 = 0; edg2 = 2;
        }
        else if (qName.trim().equalsIgnoreCase("edge1") || qName.trim().equalsIgnoreCase("edge2")) {
            edge = "";
        }
        else if (qName.trim().equalsIgnoreCase("point")) {
            if (edge.equalsIgnoreCase("edg1"))
                edg1++;
            else if (edge.equalsIgnoreCase("edg2"))
                edg2++;
        }
        //testing
        //System.out.println("ENDELEMENT - uri:" + uri + " localName:" + localName + " qName:" +qName);
    }


    /**
     *
     * @param ch the characters from the XML document
     * @param start the start position in the array
     * @param length the number of characters to read from the array
     * @throws org.xml.sax.SAXException any SAX exception, possibly
     *            wrapping another exception
     */
    public void characters (char ch[], int start, int length)
	throws SAXException {

        String s = new String(ch, start, length);
        s = s.trim();
        if (s.length() == 0)
            return;

        if (inElement.equalsIgnoreCase("handle")) {
            if (edge.equalsIgnoreCase("edg1"))
                line[edg1] = s;
            else if (edge.equalsIgnoreCase("edg2"))
                line[edg2] = s;
        }
        else if (inElement.equalsIgnoreCase("label")) {
            if (edge.equalsIgnoreCase("edg1"))
                pointlabels[edg1].add(s);
            else if (edge.equalsIgnoreCase("edg2"))
                pointlabels[edg2].add(s);
        }
        else if (inElement.equalsIgnoreCase("ruleType")) {
            line[4] = s;
        }
        else if (inElement.equalsIgnoreCase("factor")) {
            line[5] = s;
        }
        else if (inElement.equalsIgnoreCase("difference")) {
            line[6] = s;
        }
        else if (inElement.equalsIgnoreCase("tolerance")) {
            line[7] = s;
        }
        else if (inElement.equalsIgnoreCase("absolute")) {
            line[8] = s;
        }


        //testing
        //System.out.println("characters:" + s.trim());
    }


    private void resetLine() {
        line = new String[FIELDS];
        for (int i = 0; i < line.length; i++)
            line[i] = "";
    }

    private void resetPointLabels() {
        for (int i = 0; i < pointlabels.length; i++)
            pointlabels[i] = new Vector<String>();
    }

    /**
     * Parse the line array and build a constraint. This method is similar to the parseLine 
     * method in ShapeQuery, but it was added because we are extending queries to support 
     * point labeling and reference vectors. In effect it replaces that method, since that method 
     * will not support these more complex queries. Note that labels are NOT added to the constraints
     * here but rather immediately before constraints are returned. This is because point names 
     * can be reused in queries, and we can't be sure the labels are regiven.
     *
     * @return a constraint object.
     */
    private Constraint buildConstraint() throws SAXException{
        
        int myRuleType = parseRuleType(line[4]);

        float myFactor = (new Float(line[5])).floatValue();
        float myDifference = (new Float(line[6])).floatValue();
        float myTolerance = (new Float(line[7])).floatValue();

        boolean myToleranceIsAbsolute = parseTolerance(line[8]);

        Constraint con = new Constraint(
            line[0], line[1],
            line[2], line[3],
            parseRuleType(line[4]),
            myFactor,
            myDifference,
            myTolerance,
            parseTolerance(line[8]));
        
        
        //keep track of any labels seen
        for (int i = 0; i < pointlabels.length; i++) {
            hashLabels(line[i], pointlabels[i]);
        }


        return con;
        
    }
    
    /**
     * Used to track all labels seen for all pointnames.
     *
     */
    private void hashLabels(String pointhandle, Vector<String> labels) {
        if (!prevlabels.containsKey(pointhandle)) {
            prevlabels.put(pointhandle, labels);
            return;
        }
        //else add unseen labels to the existing vector of labels
        Vector<String> v = (Vector<String>)prevlabels.get(pointhandle);
        String s = "";
        for (int i = 0; i < labels.size(); i++) {
            s = (String)labels.get(i);
            if (!v.contains(s)) {
                v.add(s);
            }
        }
    }
    
    private boolean parseTolerance(String toleranceIsAbsolute) throws SAXException{
        if (toleranceIsAbsolute.equalsIgnoreCase("Y") || toleranceIsAbsolute.equalsIgnoreCase("T"))
            return true;
        if (toleranceIsAbsolute.equalsIgnoreCase("N") || toleranceIsAbsolute.equalsIgnoreCase("F"))
            return false;
        throw new SAXException("The flag indicating if tolerance is absolute must be \"Y/N\" or \"T/F\""); 
    }
    
    private int parseRuleType(String ruleType) throws SAXException {
        if (ruleType.equalsIgnoreCase("D")) {
            return Constraint.RULE_TYPE_DISTANCE;
        }
        else if (ruleType.equalsIgnoreCase("A")) {
            return Constraint.RULE_TYPE_ANGLE;
        }
        throw new SAXException("Ruletype must be \"A\" or \"D\""); 
    }
    
    private String buildLine() {
        String s = "";
        for (int i = 0; i < line.length; i++) {
            s += line[i];
            if (i < (line.length - 1))
                s += "\t";
        }
        //System.out.println(s);
        return s;
    }

    /**
     * Gets a vector of all constraints constructed when parsing the query. Adds all labels 
     * to the appropriate points in all constraints prior to returning the constraints.
     *
     * @return A vector of all constraints constructed when parsing the query.
     */
    public Vector<Constraint> getConstraints() {
        Constraint con = null;
        for (int i = 0; i < queryconstraints.size(); i++) {
            con = (Constraint)queryconstraints.get(i);
            con.setRequiredEdgeStartPointLabels(((Vector<String>)prevlabels.get(con.myRequiredEdgeStartPointName)));
            con.setRequiredEdgeEndPointLabels(((Vector<String>)prevlabels.get(con.myRequiredEdgeEndPointName)));
            con.setTargetEdgeStartPointLabels(((Vector<String>)prevlabels.get(con.myTargetEdgeStartPointName)));
            con.setTargetEdgeEndPointLabels(((Vector<String>)prevlabels.get(con.myTargetEdgeEndPointName)));
        }
        
        return queryconstraints;
    }

    public String getQueryname() {
        return queryname;
    }
    
    public boolean removeDuplicates() {
        if (removeDupes != null && removeDupes.equalsIgnoreCase("false")) 
            return false;
        return true;
    }

}
