/*
 * QueryTestLoader.java
 *
 * Created on February 23, 2005, 11:05 AM
 * Provides functionality for easily loading queries and pointsets from either
 * config files or XML files.
 */
package ShapeQuery;

import java.io.IOException;
import java.io.Reader;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.xml.sax.XMLReader;
import org.xml.sax.InputSource;
import org.xml.sax.SAXParseException;
import java.util.HashMap;
import java.util.Vector;
import java.util.Iterator;
import java.net.URL;

/**
 * Provides functionality for easily loading queries and pointsets from either
 * config files or XML files.
 * @author David Hessing
 */

public class QueryLoader {
    
    private String configFiles[];
    //array of array of filenames. Each array in 2nd dim is the query filenames from the config file.
    private String queryfiles[][];
    private float tolerances[];
    private DataSet datasets[][];
    //private PointUniverse universes[];

    
    /** Creates a new instance of QueryTestLoader */
    public QueryLoader(String configFilenames[]) {
        configFiles = configFilenames;
        //universes = new PointUniverse[numconfigFiles];
        load();
    }
    
    
    /**
     * Iterates over configger objects, and gets the query filenames, the tolerances and the datasets
     * specified in each. Loads the datasets into a PointUniverse object. All these are in parallel arrays.
     */
    private void load() {
        DataSet ds[] = null;
        Configger configgers[] = readConfigFiles();
        
        if (configgers != null) {
            queryfiles = new String[configgers.length][];
            tolerances = new float[configgers.length];
            datasets = new DataSet[configgers.length][];

            for (int i=0; i < configgers.length; i++) 
            { 
                queryfiles[i] = (String[]) configgers[i].getQueries().toArray(new String[0]);
                tolerances[i] = configgers[i].getTolerance();
                datasets[i] = (DataSet[]) configgers[i].getDataSets().toArray(new DataSet[0]);
                //universes[i] = loadDataSets(ds);
            }
        }
    }
    

    /**
     * Reads the config files and creates configger objects for each.
     *
     * @return Array of configger objects, parallel to the array of config filenames passed to the constructor.
     */
    private Configger[] readConfigFiles()
    {
        Configger con[] = new Configger[configFiles.length];
        try{
            for(int i=0; i < configFiles.length; i++) {
                con[i] = new Configger(configFiles[i]);
            }
        }
        catch (IOException ioe)
        {
            OutputWriter.println("Reading config file failed");
            return null;
        }
        return con;
                
    }
    
    

    /**
     * Loads all of the datasets in an array into a new point universe.
     * 
     * @param ds Array of datasets to load.
     * @return A PointUniverse with the datasets loaded.
     */
    public PointUniverse loadDataSets(DataSet ds[]) {
        return loadDataSets(ds, 0, ds.length);
    }
    
    
    /**
     * Loads some or all of the datasets in an array into a new point universe.
     * 
     * @param ds Array of datasets to load.
     * @param startIndex index of the first dataset to load.
     * @param numToLoad the number of datasets to load, starting from startIndex.
     * @return A PointUniverse with the datasets loaded.
     */
    public PointUniverse loadDataSets(DataSet ds[], int startIndex, int numToLoad) 
    {
        if (startIndex < 0) {startIndex = 0;}
        if (startIndex > ds.length) {startIndex = ds.length;}

        int endIndex = startIndex + numToLoad;
        if (endIndex < 0) {endIndex = 0;}
        if (endIndex > ds.length) {endIndex = ds.length;}

        PointUniverse u = new PointUniverse();
        for (int i = startIndex; i < endIndex; i++) 
        {
            if (u.addPointSet(ds[i].fileName, ds[i].visibleName) == null)
            {
                OutputWriter.print("Error loading point set: '"+ds[i].visibleName+"'.\n");
                return null;
            }
        }
        return u;
        
    }
    

    /**
     * Loads all of the datasets specified in XML into a new point universe.
     * 
     * @param reader A buffered reader underneath which are pointsets in proper XML format.
     * @return A PointUniverse with the pointsets loaded.
     */
    public static PointUniverse loadXMLDataSets(Reader reader) {
        InputSource source = new InputSource(reader);
        return loadXMLDataSets(source);
    }
    
    /**
     * Loads all of the datasets specified in XML into a new point universe.
     * 
     * @param location A URL pointing to an XML file of pointsets in proper format. The URL must be fully qualified!
     * @return A PointUniverse with the pointsets loaded.
     */
    public static PointUniverse loadXMLDataSets(URL location) {
        InputSource source = new InputSource(location.toString());
        return loadXMLDataSets(source);
    }
    
    /**
     * Loads all of the datasets specified in XML into a new point universe.
     * 
     * @param source An InputSource object underneath which are pointsets in proper XML format.
     * @return A PointUniverse with the pointsets loaded.
     */
    private static PointUniverse loadXMLDataSets(InputSource source) 
    {
        String s = "";
        PointUniverse u = new PointUniverse();

        XMLReader parse = getXMLReader();
        PointSetContentHandler handler = new PointSetContentHandler();
        parse.setContentHandler(handler);
        try {
            parse.parse(source); 
        }
        catch (IOException ioe) {
            OutputWriter.printerror("Error creating input source: " + ioe.getMessage() + "\n");
            return null;
        }
        catch (org.xml.sax.SAXException se) {
            OutputWriter.printerror("Error parsing: " + se.getMessage() +"\n");
            //se.printStackTrace(System.out);
            return null;
        }
        //get the hashmap and put the vectors inside it into the point universe
        HashMap pointsets = handler.getPointSets();
        Iterator iter = pointsets.keySet().iterator();
        while (iter.hasNext()) {
            s = (String)iter.next();
            u.addPointSet((Vector)pointsets.get(s), s);
        }
        //Finally, add the count of labels accross all pointsets. 
        u.setLabelCount(handler.getLabelCount());
        return u;
        
    }


    /**
     * Loads all of the datasets specified in XML into a new point universe.
     * 
     * @param reader A buffered reader underneath which are pointsets in proper XML format.
     * @param preloaded A hashmap containing datasets that were preloaded. Datasets referenced in the XML that already
     * exist in preloaded will not be reloaded and their indices will not be recreated.
     * @return A PointUniverse with the pointsets loaded.
     */
    public static PointUniverse loadXMLDataSets(Reader reader, HashMap<String, PointSet> preloaded) {
        InputSource source = new InputSource(reader);
        return loadXMLDataSets(source, preloaded);
    }
     
    /** 
     * Loads all of the datasets specified in XML into a new point universe.
     * 
     * @param location A URL pointing to an XML file of pointsets in proper format. The URL must be fully qualified!
     * @param preloaded A hashmap containing datasets that were preloaded. Datasets referenced in the XML that already
     * exist in preloaded will not be reloaded and their indices will not be recreated.
     * @return A PointUniverse with the pointsets loaded.
     */
    public static PointUniverse loadXMLDataSets(URL location, HashMap<String, PointSet> preloaded) {
        InputSource source = new InputSource(location.toString());
        return loadXMLDataSets(source, preloaded);
    }
    
    /**
     * Loads all of the datasets specified in XML into a new point universe.
     * 
     * @param source An InputSource object underneath which are pointsets in proper XML format.
     * @param preloaded A hashmap containing datasets that were preloaded. Datasets referenced in the XML that already
     * exist in preloaded will not be reloaded and their indices will not be recreated.
     * @return A PointUniverse with the pointsets loaded.
     */
    private static PointUniverse loadXMLDataSets(InputSource source, HashMap<String, PointSet> preloaded) 
    {
        String s = "";
        String raw = "";
        PointUniverse u = new PointUniverse();

        XMLReader parse = getXMLReader();
        PointSetContentHandler handler = new PointSetContentHandler();
        parse.setContentHandler(handler);
        try {
            parse.parse(source);
        }
        catch (IOException ioe) {
            OutputWriter.printerror("Error creating input source: " + ioe.getMessage() + "\n");
            return null;
        }
        catch (org.xml.sax.SAXException se) {
            OutputWriter.printerror("Error parsing: " + se.getMessage() +"\n");
            //se.printStackTrace(System.out);
            return null;
        }
        //get the hashmap and put the vectors inside it into the point universe
        HashMap pointsets = handler.getPointSets();
        HashMap rawXMLs = handler.getRawXMLs();
        Iterator iter = pointsets.keySet().iterator();
        while (iter.hasNext()) {
            s = (String)iter.next();
            //using the pointset name, get the raw XML for the pointset in order to properly cache
            raw = (String)rawXMLs.get(s);
            //check if this dataset is in preloaded
            if (preloaded.containsKey(raw)) {
                //put that dataset into the universe with a cleared solutionset
                PointSet p = ((PointSet)preloaded.get(raw));
                p.setSolutionSet(null);
                p.flagCached(); //flag that the pointset was pulled from a cache.
                u.addPointSet(p);
            }
            //otherwise create the new pointset, but add it to the hashmap in case it's referenced again
            else {
                PointSet p = u.addPointSet((Vector)pointsets.get(s), s);
                preloaded.put(raw, p);
            }
        }
        //Finally, add the count of labels accross all pointsets. Note that caching does 
        //not present a problem here since we are relying on the assumption that
        //identically named pointsets are identical (for a particular session).
        //Thus even though the labels were counted from XML, and a pointset may
        //be loaded from the cache, there will be no discrepancies.
        u.setLabelCount(handler.getLabelCount());
        return u;
        
    }    
    
    
    
    
    
    
    /**
     * Loads the query specified in XML into a new ShapeQuery.
     * 
     * @param reader A buffered reader underneath which is the query in proper XML format.
     * @return A ShapeQuery with the query loaded.
     */
    public static ShapeQuery loadXMLQuery(Reader reader) 
    {
      InputSource source = new InputSource(reader);
      return loadXMLQuery(source);
    }

    /**
     * Loads the query specified in XML into a new ShapeQuery.
     * 
     * @param location A URL pointing to an XML query file in the proper format. The URL must be fully qualified!
     * @return A ShapeQuery with the query loaded.
     */
    public static ShapeQuery loadXMLQuery(URL location) 
    {
      InputSource source = new InputSource(location.toString());
      return loadXMLQuery(source);
    }

    /**
     * Loads the query specified in XML into a new ShapeQuery.
     * 
     * @param source An InputSource object underneath which is the query in proper XML format.
     * @return A ShapeQuery with the query loaded.
     */
    private static ShapeQuery loadXMLQuery(InputSource source) 
    {
        String s = "";

        XMLReader parse = getXMLReader();
        QueryContentHandler handler = new QueryContentHandler();
        parse.setContentHandler(handler);
        try {
            parse.parse(source);
        }
        catch (IOException ioe) {
            OutputWriter.printerror("Error creating input source: " + ioe.getMessage() + "\n");
            return null;
        }
        catch (org.xml.sax.SAXException se) {
            OutputWriter.printerror("Error parsing: " + se.getMessage() +"\n");
            //se.printStackTrace(System.out);
            return null;
        }
	ShapeQuery query = new ShapeQuery(handler.removeDuplicates());
        Vector queryconstraints = handler.getConstraints();
        query.addConstraints(handler.getQueryname(), queryconstraints);
        //query.readQueryVector(handler.getQueryname(), lines, -1);

        return query;
        
    }
    
    /**
     * Attempts to create an XMLReader with which to parse pointsets and queries passed as XML.
     * @return An instance of an XMLReader, or null if creation was not successful.
     */
    private static XMLReader getXMLReader() {
        SAXParserFactory factory = null;
        SAXParser parser = null;
        XMLReader reader = null;
        try {
            factory = SAXParserFactory.newInstance();
            parser = factory.newSAXParser();
            reader = parser.getXMLReader();
        }
        catch (javax.xml.parsers.FactoryConfigurationError fce) {
            OutputWriter.printerror("Error creating parser factory: " + fce.getMessage());
        }
        catch (javax.xml.parsers.ParserConfigurationException pce) {
            OutputWriter.printerror("Configuration error while creating parser: " + pce.getMessage());
        }
        catch (org.xml.sax.SAXException se) {
            OutputWriter.printerror("SAX error while creating parser or reader: " + se.getMessage());
        }
        
        return reader;
    }
    
    /**
     * Loads the indicated query file into a new ShapeQuery object.
     *
     * @param filename The file containing the query.
     * @param tolerance A tolerance value to use for the query.
     * @return A ShapeQuery object with the query loaded.
     */
    public ShapeQuery loadQuery(String filename, float tolerance)
    {
        
	ShapeQuery query = new ShapeQuery();

        if (!query.readQueryFile(filename, tolerance))
        {
            OutputWriter.printerror("Unable to read query file: '" + filename + "'.\n");
            return null;
        }
        return query;
    }
    
    
    
    
    
    
    /**
     * Get the datasets that were specified by the i'th config file in the array
     * of config file names that was originally passed to the constructor.
     * Returns NULL if ANY of the config files specified failed to load.
     *
     * @param i the index in relative to the config filename array passed to the constructor.
     * @return array of datasets that were specified by the i'th config file.
     */
    public DataSet[] getDataSets(int i)
    {
        if (datasets == null) {return null;}
        return datasets[i];
    }
    
    /**
     * Get the filenames of the queries that were specified by the i'th config file in the array
     * of config file names that was originally passed to the constructor.
     * Returns NULL if ANY of the config files specified failed to load.
     *
     * @param i the index in relative to the config filename array passed to the constructor.
     * @return array of query filenames that were specified by the i'th config file.
     */
    public String[] getQueryFileNames(int i)
    {
        if (queryfiles == null) {return null;}
        return queryfiles[i];
    }

    /**
     * Get the tolerance that were specified by the i'th config file in the array
     * of config file names that was originally passed to the constructor.
     * Returns -1 if ANY of the config files specified failed to load.
     *
     * @param i the index in relative to the config filename array passed to the constructor.
     * @return the tolerance that was specified by the i'th config file. If no tolerance was specified, returns -1.     */
    public float getTolerance(int i)
    {
        if (tolerances == null) {return -1;}
        return tolerances[i];
    }
}
