package ShapeQuery;
// QuerySolution.java: Class that represents a single solution.
//
// Contains a set of tuples of the form:
//
//     point-identifier   =>   x,y coordinates of corresponding
//                             point in this solution
//
// One tuple for each point identifier in the part of the query that
// this is a solution to.
//
// When we're building a query solution, we build the solutions to
// intermediate problems, so a QuerySolution may contain just a
// partial answer to the overall query (i.e. some but not all points).

import java.util.*;  // for vector
import java.io.*;    // for reading files
import java.net.*;

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;


class QuerySolution
{
    private HashMap<String, Point> pointNames;  // maps point name, e.g. "p1", to
				 // coordinates, e.g. "(123, 456)"
    private HashSet<Point> pointList;   // list of points in this query
				 // solution, e.g. { (123,456),
				 // (10,20) }

    QuerySolution()
    {
	pointNames = new HashMap<String, Point>();
	pointList = new HashSet<Point>();
    }

    QuerySolution(QuerySolution q)
    {
	pointNames = (HashMap<String, Point>)(q.getPointNames().clone());
	pointList = (HashSet<Point>)(q.getPointList().clone());
    }

    // if point is already in set, return false
    boolean addPoint(String name, Point point)
    {
	if (pointNames.containsValue(point))
	    return false;

	Point oldPoint = pointNames.put(name, point);
        if (oldPoint != null)
            pointList.remove(oldPoint);
	pointList.add(point);
	return true;
    }

    Point getPoint(String name)
    {
	return (Point)(pointNames.get(name));
    }

    // function is to enable creation of a new solutionset that's just
    // like this one
    HashMap<String, Point> getPointNames()
    {
	return pointNames;
    }

    // To perform duplicate solution elimination, we'll store this
    // solution in a hash, with key = the list of points in the set.
    // This function returns that key.
    // 
    // this function is also used in one of the constructors.
    //
    // probably this should be replaced by extracting the points from
    // pointNames.
    HashSet<Point> getPointList()
    {
	return pointList;
    }
    
    
//    Collection<Point> getPointsInSolution()
//    {
//        return pointNames.values();
//    }

    // print this solution
    void print()
    {
	OutputWriter.println("--------");
	Iterator iter = pointNames.keySet().iterator();
	while (iter.hasNext())
	{
	    String name = (String)iter.next();
	    OutputWriter.println(name+"="+pointNames.get(name));
	}
    }
    
    
    String getSVG(HashMap edgesToDraw) {
        return getSVG(edgesToDraw, 1, 0, -1, true);
    }
    
    String getSVG(HashMap edgesToDraw, float mult, float trans, int id, boolean visible) {
        String s = "<g";
        if (id >0) {
            s += " id=\"" + id + "\"";
        }
        if (!visible) {
            s += " visibility=\"hidden\"";
        }
        s += ">\n";
        
        Iterator edgeIter = edgesToDraw.keySet().iterator();
        Point p1 = null, p2 = null; 
        String sp1 = "", sp2 = "";
        float x1 = 0, y1 = 0, x2 = 0, y2 = 0;
        while(edgeIter.hasNext()) {
            HashSet edge = (HashSet)edgeIter.next();
            if (edge.size() > 2) { 
                //note there can be one point in the edge, an edge from a point to itself.
                //this will be the case when searching for sides of a specific exact length.
                OutputWriter.printerror("Wrong number of points in edge."); 
                return "<!-- Error building SVG string: Wrong number of points in edge. -->";
            }
            Iterator pnameIter = edge.iterator();
            sp1 = ((String)pnameIter.next());
            sp2 = (edge.size() == 2 ) ? ((String)pnameIter.next()) : sp1;
            p1 = (Point)pointNames.get(sp1);
            p2 = (Point)pointNames.get(sp2);
            if (p1 == null || p2 == null) {continue;} //this edge hasn't yet been found in this solution

            s += "\t<line ";
            x1 = p1.getX() * mult + trans;  y1 = p1.getY() * mult + trans;
            x2 = p2.getX() * mult + trans;  y2 = p2.getY() * mult + trans;

            s += "x1=\"" + (int)x1 + "\" y1=\"" + (int)y1 + "\" ";
            s += "x2=\"" + (int)x2 + "\" y2=\"" + (int)y2 + "\" ";
            s += "style=\"fill:white;stroke:blue;stroke-width:2\"/>\n";
            
            s += "\t<text x=\"" + (x1 - 10) + "\" y=\"" + (y1 - 10) + "\" fill=\"yellow\">" + sp1 + "</text>\n";
            s += "\t<text x=\"" + (x2 - 10) + "\" y=\"" + (y2 - 10) + "\" fill=\"green\">" + sp2 + "</text>\n";
        }
        //s += "\t<text x=\"" + (x2 - 3) + "\" y=\"" + (y2 - 3) + "\">" + sp2 + "</text>\n";
        s += "</g>";
        return s;
    }
    
}
