package ShapeQuery;
// ShapeQuery class -- reads a constraint list from file, builds a
//                     query plan, executes it

// A note on displaying edges:
//
//   Concept #1: the "platonic form" of a list of edges to draw.  This
//   is, for example, "draw line p1-p2; draw line p2-p3"
//
//   Concept #2: a real-life list of edges to draw.  For example,
//   "draw a line from (200,400) to (300,400); draw a line from
//   (300,400) to (300,100)".
//
// The "form" of the shape can be determined as soon as the list of
// constraints is complete; the real edges to draw is the output of a
// search.

// see notes on file format, below


import java.util.*;  // for vector
import java.io.*;    // for reading files
import java.net.*;

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;


class ShapeQuery
{
    // how large a column of the query plan search chart can be before
    // we have to cut it down to size
    public static final int MAX_PLAN_CHART_SIZE = 1000000;

    private LinkedList<Constraint> constraints;   // constraints we need to build query rules


    private String queryname;
    private boolean removedDuplicates;
    
    private boolean removeDupes; //whatever this is set to is the default behaviour for the "execute()" call.
    // -----------------
    // what shape to draw

    // the new way -- includes the colors that will be used.  a key of
    // this map is a HashSet of points in the edge (there are always
    // two, but i use a hash set so order doesn't matter).  the value
    // of the key is the color used to draw this edge.
    HashMap<HashSet, Color> edgesToDraw;

    // -----------------

    // applet-related variables
    boolean runStandalone;    // are we running as an applet?
    URL codeBase;             // where to look for files

    String errorMessage;      // pending error message

    LinkedList<ShapeQueryPlan> possibleQueryPlans; // the plans we might use; we'll
				   // evaluate this list and choose
				   // the best from among them

    // make a new shape query
    public ShapeQuery(boolean newRunStandalone, URL newCodeBase)
    {
	runStandalone = newRunStandalone;
	codeBase = newCodeBase;
        init();
    }
    //two constructors for default running standalone. DJH
    public ShapeQuery() 
    {
        removeDupes = true;
        runStandalone = true;
        init();
    }
    public ShapeQuery(boolean removeDuplicates) 
    {
        removeDupes = removeDuplicates;
        runStandalone = true;
        init();
    }
    //stuff common to all constructors. DJH
    private void init() 
    {
        OutputWriter.println("\n\n\n________________________________\n\n\n* making new shapequery");
	constraints = new LinkedList<Constraint>();

	possibleQueryPlans = new LinkedList<ShapeQueryPlan>();
	edgesToDraw = null;
        
        queryname = "";
        removedDuplicates = false;
    }
    
    /**
     * Gets the queryname. If the query was read in from a file, the queryname is the filename.
     * Otherwise, if the query was read in from an XML stream, the queryname was specified there.
     * @return The name of the last query loaded into this object.
     */
    public String getQueryName() {
        return queryname;
    }
    
    /**
     * Indicates whether the last time a query was executed duplicates were removed.
     * Note the value makes no sense if no query was executed (it will be false by default).
     * @return True if duplicates were removed, false otherwise.
     */
    public boolean getDuplicateStatus() {
        return removedDuplicates;
    }
    
    
    private void addConstraint(Constraint constraint)
    {
	constraints.add(constraint);
    }

    
    // -------------------------------------------------------------
    // functions to read in a query (and related helper functions)

    // read query from the supplied string.
    // always returns true
    public boolean readQuery(String queryString, float tolerance)
    {
	edgesToDraw = null;  // forget any cached info about how to draw this shape
	constraints = new LinkedList<Constraint>();
	OutputWriter.println("* reading query string input...");

	StringTokenizer tokenizer = new StringTokenizer(queryString,
							"\n");

	int lineCount = 0;
	try
	{
	    while (true)
	    {
		String line=tokenizer.nextToken();
		parseLine(line, tolerance);
		lineCount++;
	    }
	}
	catch (NoSuchElementException e)
	{
	    // end of input

// 	    OutputWriter.println("parse error, ignoring line: ["
// 				 +line+"]");
	}

	OutputWriter.println("* read "+lineCount+" lines of query");

	return true;
    }

    // read query from the supplied filename.
    // return true if file read succeeds (file may still not parse).
    public boolean readQueryFile(String fileName, float tolerance)
    {
        queryname = fileName;
        
	edgesToDraw = null;  // forget any cached info about how to draw this shape
        BufferedReader input;
	constraints = new LinkedList<Constraint>();

	// if applet, need to do extra work
	DataInputStream url_input = null;
	URL sourceurl = null;

	OutputWriter.print("* reading query file ");

	if (!runStandalone)
	{
	    // as applet, open url
	    OutputWriter.println("from url...");

	    try
	    {
		sourceurl = new URL(codeBase, fileName);
		url_input = new DataInputStream(sourceurl.openStream());
	    }
            catch (java.security.AccessControlException e)
            {
                OutputWriter.println("unable to open file due to security problem:"
                    +e.getMessage());
                return false;
            }
	    catch (java.net.MalformedURLException e)
	    {
		errorMessage = "url problem: "+e.getMessage();
		return false;
	    }
	    catch (java.io.IOException e)
	    {
		errorMessage = "file problem: "+e.getMessage();
		return false;
	    }
	}

	// -----------------------
	// open file

        try                                                           
	{
	    if (runStandalone)
	    {
		OutputWriter.println("directly from filesystem...");
		input = new BufferedReader(new FileReader(fileName));
	    }
	    else
	    {
		input = new BufferedReader(new InputStreamReader(url_input));
	    }
        }                                                             
        catch (java.io.FileNotFoundException e)                       
        {                                                             
            OutputWriter.println("file not found");                     
            return false;                                             
        }                                                             

	// ---------------------------
	// read each line
        try
        {
            while (input.ready())
            {
                String line = input.readLine();
		parseLine(line, tolerance);
	    }
	    input.close();
	}
	catch (java.io.IOException e)
        {
            OutputWriter.println("io error: "+e.getMessage());
	    return false;
        }
	catch (Exception e)
        {
            OutputWriter.println("general error: "+e.getMessage()
                               +" (do you reference points, edges, or edgepairs that do not exist?  do you have spaces where you shouldn't?  file must be tabdelimited)"
		);
	    return false;
        }


	OutputWriter.println("* file read complete.");

	return true;
    }


    /** 
     * Given a vector of lines, where each line is a constraint, parses the lines
     * to build a query. Originally, XML queries were first parsed into the old format in order
     * to reuse Vespe's code. Now, the QueryContentHandler builds constraint objects directly.
     *
     * @param qname Name of this query
     * @param lines Vector of query lines in the old format specified by Vespe
     * @param tolerance an optional overriding tolerance to use for the query. Set negative if it should not be used.
     */
    public void readQueryVector(String qname, Vector lines, float tolerance) {
        queryname = qname;
        String line = "";
        for (int i=0; i < lines.size(); i++) {
            line = (String)lines.get(i);
            parseLine(line, tolerance);
        }
    }
    
    
    public void addConstraints(String qname, Vector constraints) {
        queryname = qname;
        for (int i=0; i < constraints.size(); i++) {
            addConstraint(((Constraint)constraints.get(i)));
        }
    }
    
    // parse the supplied line, and add it (as a constraint) to my
    // list of constraints.  '_tolerance' if non-negative will
    // override whatever tolerance exists on this line.
    //
    // line format:
    // - all fields must be separated by a SINGLE TAB.  spaces or
    //   multiple tags will confuse it.
    // - any line that is blank or begins with a # is ignored
    // - there are 9 fields, in this order
    //   # relative-edge |target-edge    |Dist/  |factor |differ-|toler- |absol-
    //   # start  end    |start  end     |Angle  |       |ence   |ance   |ute?

    private void parseLine(String line, float _tolerance)
    {
	// ignore comments
	if (line.length()==0 || line.charAt(0)=='#')
	    return;

	// tokenize this string
	StringTokenizer tokenizer = new StringTokenizer(line,
							"\t");

	String requiredEdgeStartPointName,requiredEdgeEndPointName;
	String targetEdgeStartPointName, targetEdgeEndPointName;
	String ruleType;
	String factor;
	String difference;
	String tolerance;
	String toleranceIsAbsolute;

	try
	{
	    requiredEdgeStartPointName=tokenizer.nextToken();
	    requiredEdgeEndPointName=tokenizer.nextToken();
	    targetEdgeStartPointName=tokenizer.nextToken();
	    targetEdgeEndPointName=tokenizer.nextToken();
	    ruleType=tokenizer.nextToken();
	    factor=tokenizer.nextToken();
	    difference=tokenizer.nextToken();
	    tolerance=tokenizer.nextToken();
	    toleranceIsAbsolute=tokenizer.nextToken();


	    int myRuleType = -1;
	    if (ruleType.equals("D"))
	    {
		myRuleType = Constraint.RULE_TYPE_DISTANCE;
	    }
	    else if (ruleType.equals("A"))
	    {
		myRuleType = Constraint.RULE_TYPE_ANGLE;
	    }
	    else
	    {
		OutputWriter.println("Error: constraint type should be either 'A' or 'D' (found '"+ruleType+"')");
		return;
	    }

	    float myFactor = (new Float(factor)).floatValue();
	    float myDifference = (new Float(difference)).floatValue();
	    float myTolerance = (new Float(tolerance)).floatValue();

	    boolean myToleranceIsAbsolute = false;

	    if (toleranceIsAbsolute.equals("Y") ||
		toleranceIsAbsolute.equals("T") ||
		toleranceIsAbsolute.equals("y") ||
		toleranceIsAbsolute.equals("t"))
		myToleranceIsAbsolute = true;
	    
	    // use overriding tolerance
	    if (_tolerance >= 0)
	    {
		// for distances, use the tolerance as fraction of the
		// length in the constraint
		if (myRuleType==Constraint.RULE_TYPE_DISTANCE)
		{
		    // convert from decimal to percent
		    myTolerance = _tolerance*100;
		    myToleranceIsAbsolute = false;
		}
		// for angles, use as an absolute number of degrees,
		// but scale by 50 (this appears to be a reasonable
		// number).
		else if (myRuleType==Constraint.RULE_TYPE_ANGLE)
		{
		    myTolerance = _tolerance*50;
		    myToleranceIsAbsolute = true;
		}
	    }
	    
	    Constraint con= new Constraint(
		requiredEdgeStartPointName,requiredEdgeEndPointName,
		targetEdgeStartPointName, targetEdgeEndPointName,
		myRuleType,
		myFactor,
		myDifference,
		myTolerance,
		myToleranceIsAbsolute);

	    addConstraint(con);
	}
	catch (NoSuchElementException e)
	{
	    OutputWriter.println("parse error, ignoring line: ["
				 +line+"]");
	}
	return;
    }

    // to figure out what edges we should draw, look through
    // constraints, and draw any edge that appears in a constraint.
    HashMap getEdgesToDraw()
    {
	// use cached copy if we have one
	if (edgesToDraw != null)
	    return edgesToDraw;

	// each item of this set is a hashset of points.
	// using a hash set here will enforce uniqueness, regardless of ordering.
	edgesToDraw = new HashMap<HashSet, Color>();

	Iterator edgeIter = constraints.iterator();

	while (edgeIter.hasNext())
	{
	    Constraint curr = (Constraint)edgeIter.next();
	    
	    // form set containing relative edge
	    HashSet<String> relativeEdge = new HashSet<String>();
	    relativeEdge.add(curr.myRequiredEdgeStartPointName);
	    relativeEdge.add(curr.myRequiredEdgeEndPointName);

	    // form set containing target edge
	    HashSet<String> targetEdge = new HashSet<String>();
	    targetEdge.add(curr.myTargetEdgeStartPointName);
	    targetEdge.add(curr.myTargetEdgeEndPointName);

	    // add these edges only if they're not present already
	    if (!edgesToDraw.keySet().contains(relativeEdge))
	    {
		// use the next color for this edge
		Color color = intToColor(edgesToDraw.size());

		edgesToDraw.put(relativeEdge, color);
	    }
	    if (!edgesToDraw.keySet().contains(targetEdge))
	    {
		// use the next color for this edge
		Color color = intToColor(edgesToDraw.size());

		edgesToDraw.put(targetEdge, color);
	    }
	}

	return edgesToDraw;
    }


    // pick a color for this edge
    private static Color intToColor(int i)
    {
	i = i%8; // cut down to a number from 0-7

	boolean printColor = false;

	switch (i)
	{
	    case 0:
	    {
		if (printColor)
		    OutputWriter.println("color=red");
		return Color.red;
	    }
	    case 1:
	    {
		if (printColor)
		    OutputWriter.println("color=green");
		return Color.green;
	    }
	    case 2:
	    {
		if (printColor)
		    OutputWriter.println("color=blue");
		return Color.blue;
	    }
	    case 3:
	    {
		if (printColor)
		    OutputWriter.println("color=yellow");
		return Color.yellow;
	    }
	    case 4:
	    {
		if (printColor)
		    OutputWriter.println("color=magenta");
		return Color.magenta;
	    }
	    case 5:
	    {
		if (printColor)
		    OutputWriter.println("color=orange");
		return Color.orange;
	    }
	    case 6:
	    {
		if (printColor)
		    OutputWriter.println("color=pink");
		return Color.pink;
	    }
	    case 7:
	    {
		if (printColor)
		    OutputWriter.println("color=cyan");
		return Color.cyan;
	    }
	}
	return Color.red;
    }

    // -------------------------------------------------------------
    // query plan building functions

    // build the list of possible query plans we might execute.
    // returns true if we got at least one valid plan.
    public void buildQueryPlanList(PointUniverse universe)
    {
	int FIRST_PLAN=0;
	int EXHAUSTIVE_PLAN_SEARCH=1;
	//	int DYNAMIC_PLAN_SEARCH=2;
	int DYNAMIC_PROGRAMMING_SEARCH=3;
	int planSearch = DYNAMIC_PROGRAMMING_SEARCH;

	// need to clear out list of possible query plans
	possibleQueryPlans.clear();

	if (planSearch == EXHAUSTIVE_PLAN_SEARCH)
	{
	    OutputWriter.println("==>looking for query orders...");

	    // find all possible query plans
	    LinkedList possiblePlanOrders =  extendPossibleQueryPlanOrders(constraints, new LinkedList<Constraint>());

	    OutputWriter.println("==>"+possiblePlanOrders.size()+" possible query orders found (not all necessarily unique/feasible)");

	    // figure out all possible plan orders
	    Iterator iter = possiblePlanOrders.iterator();

	    ShapeQueryPlan currentPlan = null;
	    while (iter.hasNext())
	    {
		currentPlan = new ShapeQueryPlan();

		if (currentPlan.addConstraintListAsNextRules(constraints))
		{
		    possibleQueryPlans.add(currentPlan);
		}
	    }
	}
	else if (planSearch == FIRST_PLAN)
	{
	    ShapeQueryPlan currentPlan = new ShapeQueryPlan();

	    if (currentPlan.addConstraintListAsNextRules(constraints))
	    {
		possibleQueryPlans.add(currentPlan);
	    }
	}
// 	else if (planSearch == DYNAMIC_PLAN_SEARCH)
// 	{
// 	    // this was the first attempt at a dynamic programming solution -- it's not so good

// 	    ShapeQueryPlan emptyPlan = new ShapeQueryPlan();

// 	    // build a dynamically-chosen best plan
// 	    possibleQueryPlans.add(addDynamicConstraint(universe,
// 							constraints,
// 							emptyPlan));
//	}
	else if (planSearch == DYNAMIC_PROGRAMMING_SEARCH)
	{
	    possibleQueryPlans.add(dynamicProgrammingSolution(universe));
	}
    }


    // -------------------------------------
    // EXHAUSTIVE PLAN SEARCH implementation -- impossibly slow, and
    // results are not better than cheaper searches in practice
    private LinkedList<LinkedList> extendPossibleQueryPlanOrders(LinkedList<Constraint> constraintsToAdd,
						     LinkedList<Constraint> currentConstraintOrder)
    // recursive function to populate list of all possible query p8lans.
    // constraintsToAdd is the constraints we haven't added yet but need to.
    // currentConstraintOrder is the constraints we've already added;
    // we'll extend this list.
    {
	// we'll return a list of lists
	LinkedList<LinkedList> result = new LinkedList<LinkedList>();

	OutputWriter.println("  finding query plans, left to add="+constraintsToAdd.size()+
			     ", current set contains="+currentConstraintOrder.size());

	if (constraintsToAdd.size()==1)
	{
	    // base case -- add this item and return
	    currentConstraintOrder.add(constraintsToAdd.iterator().next());
	    result.add(currentConstraintOrder);
	}
	else
	{
	    // need to iterate over remaining constraints to come up
	    // with all possible combinations
	    Iterator<Constraint> iter = constraintsToAdd.iterator();

	    while (iter.hasNext())
	    {
		// for each constraint to add, duplicate current
		// constraint order and add this constraint to the end

		Constraint newConstraint = (Constraint)(iter.next());

		LinkedList<Constraint> newConstraintOrder = (LinkedList<Constraint>)currentConstraintOrder.clone();
		newConstraintOrder.add(newConstraint);

		LinkedList<Constraint> newConstraintsToAdd = (LinkedList<Constraint>)constraintsToAdd.clone();
		Iterator remover = newConstraintsToAdd.iterator();
		while (remover.hasNext())
		{
		    if (remover.next()==newConstraint)
		    {
			remover.remove();
		    }
		}

		LinkedList<LinkedList> newResult = new LinkedList<LinkedList>();
		newResult = extendPossibleQueryPlanOrders(newConstraintsToAdd,
							  newConstraintOrder);

		Iterator<LinkedList> addResultIter = newResult.iterator();
		while (addResultIter.hasNext())
		{
		    result.add(addResultIter.next());
		}

	    }
	}
	return result;
    }

    // -------------------------------------
    // recursively add the cheapest constraint to the existing query
    // plan, and return the new plan
    /*
    private ShapeQueryPlan addGreedyConstraint(PointUniverse universe,
						LinkedList constraintsToAdd,
						ShapeQueryPlan existingPlan)
    {
	Iterator nextConstraintIter = constraints.iterator();

	OutputWriter.println("=========> adding constraint, remaining constraint list size=" + constraintsToAdd.size()+
	    " existing plan constraint list size="+existingPlan.getConstraints().size());

	if (constraintsToAdd.size()==1)
	{
	    // base case

	    // list of constraints we'll add to new plan
	    LinkedList currentConstraints = (LinkedList) existingPlan.getConstraints().clone();
	    currentConstraints.add(constraintsToAdd.iterator().next());

	    // make a query plan for the new set of constraints

	    ShapeQueryPlan potentialPlan = new ShapeQueryPlan();

	    if (potentialPlan.addConstraintListAsNextRules(currentConstraints))
	    {
		return potentialPlan;
	    }
	    else
	    {
		OutputWriter.println("ERROR: no legal plans found, aborting");
		return null;
	    }
	}

	// all the plans we've generated this round -- we'll choose
	// the best of these and keep it
	LinkedList candidatePlansThisRound = new LinkedList();

	// remember what new plans correspond to which constraints (so
	// we can remove the correct constraint when we recurse)
	HashMap candidatePlanCorrespondingConstraints = new HashMap();
	
	while (nextConstraintIter.hasNext())
	{
	    Constraint nextConstraint = (Constraint) nextConstraintIter.next();
	    
	    // Should add this constraint to a clone of the existing
	    // plan, and tell plan to extend itself by the new
	    // constraint.
	    // 
	    // For now we add this constraint and have plan rebuild
	    // itself from scratch.

	    // list of constraints we'll add to new plan
	    LinkedList currentConstraints = (LinkedList) existingPlan.getConstraints().clone();
	    currentConstraints.add(nextConstraint);

	    // make a query plan for the new set of constraints
	    ShapeQueryPlan potentialPlan = new ShapeQueryPlan();

	    // only add this plan if we successfully found a plan
	    if (potentialPlan.addConstraintListAsNextRules(currentConstraints))
	    {
		// for now, add it to our list of candidates for this round
		candidatePlansThisRound.add(potentialPlan);

		// remember what constraint we removed for this particular plan
		candidatePlanCorrespondingConstraints.put(potentialPlan, nextConstraint);
	    }
	}

	if (candidatePlansThisRound.size()==0)
	{
	    OutputWriter.println("ERROR: no legal plans found, aborting");
	    return null;
	}

	Iterator nextCandidateIterator = candidatePlansThisRound.iterator();

	ShapeQueryPlan currentCandidate = (ShapeQueryPlan) nextCandidateIterator.next();

	// assume first choice is best
	float bestPlanCost = currentCandidate.estimateCost(universe);
	ShapeQueryPlan bestCandidate = currentCandidate;

	// look for a plan that's better than the best one we currently have
	while (nextCandidateIterator.hasNext())
	{
	    currentCandidate = (ShapeQueryPlan) nextCandidateIterator.next();

	    float currentPlanCost = currentCandidate.estimateCost(universe);

	    // we've found a better plan
	    if (currentPlanCost < bestPlanCost)
	    {
		bestCandidate = currentCandidate;
		bestPlanCost = currentPlanCost;
	    }
	}

	// need to remove the constraint we just added from the list
	// of constraints to add
	//
	// !!! need to be able to undo this if we want to use a plan
	// that's not the "best" plan above


	LinkedList remainingConstraints = (LinkedList) constraintsToAdd.clone();
	nextConstraintIter = remainingConstraints.iterator();
	while (nextConstraintIter.hasNext())
	{
	    Constraint currentConstraint = (Constraint) nextConstraintIter.next();

	    // remove the constraint that we used for the best plan
	    if (currentConstraint==
		candidatePlanCorrespondingConstraints.get(bestCandidate))
	    {
		nextConstraintIter.remove();
		break;
	    }
	}

	OutputWriter.println("Started with "+constraintsToAdd.size()+" constraints; now have "
			     +remainingConstraints.size()+" constraints left to add.");

	if (remainingConstraints.size() == constraintsToAdd.size())
	{
	    OutputWriter.println("======> unable to remove a constraint");
	    return null;
	}

	// should check for a null here and try evaluating a different
	// plan
	return addDynamicConstraint(universe,
				    remainingConstraints,
				    bestCandidate);
    }
    */

    // ------------------------------------------------
    // build using "smart" dynamic-programming approach

    // planList is a chart of best plans for each set of constraints
    // (i.e. planList[0] is best plans for 1 constraint, planList[1]
    // is best plans for 2 constraints).  An element of the planList
    // is a HashMap which maps a sorted list of constraint id's to the
    // best plan for those ids
    private ShapeQueryPlan bestDynamicPlan(Vector<HashMap<HashSet, ShapeQueryPlan>> planList,
					   int depth,
					   PointUniverse universe,
					   boolean useGreedy)
    {
	OutputWriter.println("\n\n+++ filling in chart at depth: "+depth);

	// the chart of plans at this depth; maps a constraint-set to
	// the best plan that currently fulfills it
	HashMap<HashSet, ShapeQueryPlan> plansAtThisDepth = new HashMap<HashSet, ShapeQueryPlan>();

	// the last plan we've created, to shortcut the process of
	// returning a value if this is the last plan we're doing, and
	// also to have a starting-off point when searching for the
	// best plan.
	ShapeQueryPlan lastPlan = null;
	HashSet lastConstraintSet = null; // similar to above

	if (depth==0)
	{
	    // create a list of plans with 1 constraint
	    Iterator iter = constraints.iterator();
	    OutputWriter.println(constraints.size()+" constraints to use in first round.");
	    while (iter.hasNext())
	    {
		Constraint currentConstraint = (Constraint) iter.next();
		ShapeQueryPlan currentPlan = new ShapeQueryPlan();

		HashSet<Constraint> constraintSet = new HashSet<Constraint>();
		constraintSet.add(currentConstraint);
		
		if (currentPlan.addConstraintAsNextRule(currentConstraint))
		{
		    OutputWriter.print("add(first) [");

		    for (Iterator i = currentPlan.knownPoints.iterator();
			 i.hasNext();
			 )
		    {
			OutputWriter.print( (String)i.next()+" ");
		    }
		    OutputWriter.println("]");//", cost="+currentPlan.estimateCost(universe));
                    
                    OutputWriter.println("Current plan:");
        	    currentPlan.printQueryPlan();
                    OutputWriter.print("Plan Cost = ");
                    OutputWriter.println(" = " + currentPlan.estimateCost(universe));
                    
                    //  +++ adding single-constraint plan");
		    // new Integer(constraintIdList.hashCode())
		    plansAtThisDepth.put(constraintSet, currentPlan);
		}
		else
		{
		    OutputWriter.print("skip(ineff): ");
		    currentConstraint.print();
		}

		// in the event that this query takes just one
		// constraint, must store this plan so we'll have a
		// starting-off point below.
		lastPlan = currentPlan;
	    }
	}
	else // a previous round's chart exists
	{
	    // take previous plan list; expand on it
            HashMap plansAtPreviousDepth = (HashMap) planList.elementAt(depth-1);
	    Iterator iter = plansAtPreviousDepth.keySet().iterator();

	    while (iter.hasNext())
	    {
                OutputWriter.println("\n---Adding constraints on one of the Previous Plans---");
		HashSet previousConstraintSet = (HashSet) iter.next();

		ShapeQueryPlan subPlan = (ShapeQueryPlan) plansAtPreviousDepth.get(previousConstraintSet);
                //previousConstraintIdList);

                //OutputWriter.println("   +++ building on subplan which contains "
                //		     +subPlan.getConstraints().size()+" constraints");

		// for each constraint we need to add
		Iterator cIter = constraints.iterator();
		while (cIter.hasNext())
		{
		    Constraint constraint = (Constraint) cIter.next();

		    // if this plan already has this constraint, we
		    // don't need to add it
		    if (subPlan.containsConstraint(constraint))
		    {
//			OutputWriter.println("  +++ skipping plan because constraints repeat");
			OutputWriter.println("\nskip(repeat)");

			continue;
		    }

		    // make a new plan that's a clone of the subplan
		    // we're currently examining
		    ShapeQueryPlan newPlan = (ShapeQueryPlan) subPlan.clone();

		    if (newPlan==null)
		    {
			OutputWriter.println("ERROR: newPlan is null");
		    }

		    // try to add the next constraint to this plan;
		    // if we can't, skip to next constraint
		    if (!newPlan.addConstraintAsNextRule(constraint))
		    {
			OutputWriter.print("skip(ineff) [");
			for (Iterator i = newPlan.knownPoints.iterator();
			     i.hasNext();
			    )
			{
			    OutputWriter.print( (String)i.next()+" ");
			}
			OutputWriter.println("]");

			// +++ skipping plan because it can't be done efficiently");
			continue;
		    }

		    // we successfully added this constraint to this plan
		    // ---------------------------------------------
		    // 

		    // add it to our list of plans at this depth if
		    // it's less expensive than our current best plan
		    // for this set of constraints

		    // list of constraints includes all constraints
		    // from the plan we're building on
		    HashSet<Constraint> constraintSet = (HashSet<Constraint>) previousConstraintSet.clone();

		    // tack on our new constraint
		    constraintSet.add(constraint);
		    
		    // take the best plan for this list of constraints so far
		    ShapeQueryPlan competingPlan = (ShapeQueryPlan) plansAtThisDepth.get(constraintSet);

		    // if this is the first plan for these
		    // constraints, or if this plan is better than the
		    // existing plan for these constraints, make this
		    // plan the best one (for these constraints)
		    if (competingPlan==null || competingPlan.estimateCost(universe,false)>newPlan.estimateCost(universe,false))
		    {
			if (competingPlan==null)
			{
			    OutputWriter.print("add(first) [");
			    for (Iterator i = newPlan.knownPoints.iterator(); i.hasNext();)
			    {
				OutputWriter.print( (String)i.next()+" ");
			    }
			    OutputWriter.println("]");//", cost="+newPlan.estimateCost(universe));

                            OutputWriter.println("Current plan:");
			    newPlan.printQueryPlan();
                            OutputWriter.print("Current Plan Cost = ");
                            OutputWriter.println(" = " + newPlan.estimateCost(universe));

                            //OutputWriter.println("  +++ adding plan because it's first one that matches these constraints");
			}
			else
			{
			    OutputWriter.print("add(better) [");
			    for (Iterator i = newPlan.knownPoints.iterator();
				 i.hasNext();
				)
			    {
				OutputWriter.print( (String)i.next()+" ");
			    }
			    OutputWriter.println("]");//, new cost="+newPlan.estimateCost(universe) +", old cost="+competingPlan.estimateCost(universe));

			    OutputWriter.println("Discarded plan:");
			    competingPlan.printQueryPlan();
                            OutputWriter.print("Competing Plan Cost = ");
                            OutputWriter.println(" = " + competingPlan.estimateCost(universe));

                            OutputWriter.println("Kept plan:");
			    newPlan.printQueryPlan();
                            OutputWriter.print("Kept Plan Cost = ");
                            OutputWriter.println(" = " + newPlan.estimateCost(universe));

			    //+++ adding plan because it's better than previous best one that matches these constraints");
			}

			plansAtThisDepth.put(constraintSet, newPlan);
//			plansAtThisDepth.put(new Integer(constraintIdList.hashCode()), newPlan);
			lastPlan = newPlan;
			lastConstraintSet = constraintSet;
		    }
		    else
		    {
			OutputWriter.print("skip(worse) [");
			for (Iterator i = newPlan.knownPoints.iterator(); i.hasNext();)
			{
			    OutputWriter.print( (String)i.next()+" ");
			}
			OutputWriter.println("]");//, new cost="+newPlan.estimateCost(universe) +", old cost="+competingPlan.estimateCost(universe));

			OutputWriter.println("Discarded plan:");
			newPlan.printQueryPlan();
                        OutputWriter.print("Discarded Plan Cost = ");
                        OutputWriter.println(" = " + newPlan.estimateCost(universe));

			OutputWriter.println("Kept plan:");
			competingPlan.printQueryPlan();
                        OutputWriter.print("Kept Plan Cost = ");
                        OutputWriter.println(" = " + competingPlan.estimateCost(universe));
                        
                        //OutputWriter.println("  +++ not adding plan because we already have a better one that matches these constraints");
		    }
		}
	    }
	}

	// check if we have too many plans in chart -- this slows down
	// performance terribly, so once it gets big, we need to avoid
	// plans that don't look good at the start (even though they
	// could be better in the end).
	//
	// note that once we start being greedy, we continue to be
	// greedy for each additional round -- e.g. we don't bother
	// trying to optimize the 4th level if we looked only at the
	// best solution on the 3rd level
	if (useGreedy || plansAtThisDepth.size()>MAX_PLAN_CHART_SIZE)
	{
	    if (useGreedy)
	    {
		OutputWriter.println("  +++ greedily truncating list size to 1 (keeping the best only)");
	    }
	    else
	    {
		OutputWriter.println("  +++ TOO MANY PLANS in chart ("
				     +plansAtThisDepth.size()+"), truncating");
                System.out.println("Too many plans in chart ("
				     +plansAtThisDepth.size()+"), truncating");
	    }

	    // find best plan, keep that one, discard the rest

	    // start out assuming first plan is best (just so we have
	    // somewhere to start)
	    ShapeQueryPlan bestPlan = lastPlan;
	    HashSet bestConstraintSet = lastConstraintSet;

	    Iterator iter = plansAtThisDepth.keySet().iterator();
	    while (iter.hasNext())
	    {
		HashSet currentConstraintSet = (HashSet) iter.next();
		ShapeQueryPlan currentPlan =
		    (ShapeQueryPlan) plansAtThisDepth.get(
			currentConstraintSet);

		if (currentPlan.estimateCost(universe)<
		    bestPlan.estimateCost(universe))
		{
		    bestPlan = currentPlan;
		    bestConstraintSet = currentConstraintSet;
		}
	    }

	    // now that we know best plan, remove all previous plans
	    plansAtThisDepth.clear();

	    // insert just our best plan
	    plansAtThisDepth.put(bestConstraintSet, bestPlan);

	    useGreedy = true;
	}

	// save this plan list
	planList.addElement(plansAtThisDepth);

	OutputWriter.print("+++++++++++++++++++++++++++++++++++++++++++");
	OutputWriter.println("+++ done building chart at depth "+depth
			     +", plan count at this depth = "+
			     plansAtThisDepth.size());
        //System.out.println("Depth:" + depth + " Plan Count:" + plansAtThisDepth.size());

	if (depth==constraints.size()-1)
	{
	    // we're done -- we have no more constraints to process

	    OutputWriter.println("  +++ plan search complete, returning ["
				 +plansAtThisDepth.size()+" plans at this depth (should be 1)]");

	    OutputWriter.println("      lastplan rule count = "
				 +lastPlan.rules.size());

	    // return best constraint of last row -- there should be
	    // just 1 plan in that row
	    return lastPlan;
	}
	
//	OutputWriter.println("  +++ plan search not yet complete, recursing");

	// we need to recurse on next page in chart
	return bestDynamicPlan(planList,
			       depth+1,
			       universe,
			       useGreedy);
    }

    private ShapeQueryPlan dynamicProgrammingSolution(PointUniverse universe)
    {
	ShapeQueryPlan bestPlan = null;

	bestPlan =  bestDynamicPlan(new Vector<HashMap<HashSet,ShapeQueryPlan>>(), 0, universe, false);

	return bestPlan;
    }


    // -------------------------------------------------------------
    // 'execute' function
    public int execute(PointUniverse universe,
		       long timer[])
    {
        return execute (universe, removeDupes, timer);
    }

    public int execute(PointUniverse universe,
		       boolean eliminateDuplicates)
    {
        return execute (universe, eliminateDuplicates, new long[2]);
    }
    
    // execute this query plan over all point sets in the universe.
    // return the number of matches found.
    public int execute(PointUniverse universe,
		       boolean eliminateDuplicates, long timer[])
    {
        removedDuplicates = eliminateDuplicates;
	int resultCount = 0;

	OutputWriter.println("=================================");

	Calendar overallStartTime;
	Calendar overallFinishTime;
	
	Calendar startTime;
        Calendar finishTime;

	startTime = Calendar.getInstance();
	overallStartTime = startTime;

	// build query plan
	OutputWriter.print("* building query plan choices [from "
			     +constraints.size()
			     +" constraints]\n");
	buildQueryPlanList(universe);
	OutputWriter.print("* query plan choice build complete.\n");

	finishTime = Calendar.getInstance();
        timer[0] = (finishTime.getTimeInMillis() - startTime.getTimeInMillis());

        OutputWriter.print("===> query plan search time: " + timer[0]
			     +" milliseconds\n");
	OutputWriter.println("=================================");

	if (possibleQueryPlans.size()==0)
	{
	    OutputWriter.println("No query plan to execute.");
	    return -1;
	}

	Iterator iter = possibleQueryPlans.iterator();
        

	ShapeQueryPlan bestPlan = (ShapeQueryPlan) possibleQueryPlans.getFirst();
	float bestCost = bestPlan.estimateCost(universe,false);
	int count = 0;
	int bestCount = 1;
	while (iter.hasNext())
	{
	    count++;

	    ShapeQueryPlan currentPlan = (ShapeQueryPlan)(iter.next());
	    float currentCost = currentPlan.estimateCost(universe,false);
	    //OutputWriter.println("plan #"+count+" has cost "+currentCost);
	    if (currentCost < bestCost)
	    {
		bestCost = currentCost;
		bestPlan = currentPlan;
		bestCount = count;
	    }
	}

	if (bestPlan!=null)
	{
	    OutputWriter.println("best plan:");
	    bestPlan.printQueryPlan();
            OutputWriter.print("Plan Cost = ");
            OutputWriter.println(" = " + bestPlan.estimateCost(universe));
	}

// 	if (true)
// 	{
// 	OutputWriter.println("SKIPPING EXECUTE STEP");
// 	return null;
// 	}

	if (bestPlan!=null)
	{
	    //OutputWriter.println("chose plan #"+bestCount+" with cost: "+bestCost);
	    //bestPlan.printQueryPlan();

	    Iterator setIter = universe.pointSets.iterator();
	    
            //track how long to execute the query on all loaded datasets
            startTime = Calendar.getInstance();
	    while (setIter.hasNext())
	    {
		PointSet set = (PointSet) setIter.next();
		OutputWriter.println("Executing on set: "+set.visibleName);

		// execute.  results are stored with the set for later
		// retrieval.  get back result count to add to running
		// total.
		resultCount+=bestPlan.execute(set, eliminateDuplicates, this.getEdgesToDraw());
	    }
            finishTime = Calendar.getInstance();
            timer[1] = (finishTime.getTimeInMillis() - startTime.getTimeInMillis());
	}
	else
	{
	    OutputWriter.println("no suitable plan found");
	    resultCount = -1;
	}

	overallFinishTime = Calendar.getInstance();
        OutputWriter.printstat("=> Total Search Time: "
			     +(overallFinishTime.getTimeInMillis()-
			       overallStartTime.getTimeInMillis())
			     +" milliseconds");

        OutputWriter.printstat("=> Result Count: " +(resultCount));
        
	return resultCount;
    }

    // -------------------------------------------------------------
    // static helper function

    // given measure 'angle', an angle in degrees, return the
    // equivalent measure in the range -180 < n <= 180
    public static float angleAdjust(float measure)
    {
	float startMeasure = measure;
	measure = measure % 360;
	if (measure > 180.0f)
	{
	    measure-=360.0f;
	}
	else if (measure <=-180.0f)
	{
	    measure+=360.0f;
	}

// 	if (startMeasure != measure)
// 	{
// 	    OutputWriter.println("converting ["+startMeasure+"] to ["
// 			       +measure+"]");
// 	}

	return measure;
    }

    LinkedList getConstraints()
    {
	return constraints;
    }

    // return the names of the points involved in this query
    String[] getPointList()
    {
	Iterator iter = constraints.iterator();

	// list of distinct points
	HashSet<String> pointListSet = new HashSet<String>();

	while (iter.hasNext())
	{
	    // add relevant points to distinct list
	    Constraint currentConstraint = (Constraint) iter.next();
	    pointListSet.add(currentConstraint.myRequiredEdgeStartPointName);
	    pointListSet.add(currentConstraint.myRequiredEdgeEndPointName);
	    pointListSet.add(currentConstraint.myTargetEdgeStartPointName);
	    pointListSet.add(currentConstraint.myTargetEdgeEndPointName);
	}

	// turn distinct list into String[]
	iter = pointListSet.iterator();

	// make an array the size of the distinct point set
	String pointList[] = new String[pointListSet.size()];

	// copy from set to array
	int i = 0; // index of current position in array
	while (iter.hasNext())
	{
	    String currentPoint = (String) iter.next();
	    pointList[i] = currentPoint;
	    i++;
	}
	return pointList;
    }
}
