package ShapeQuery;
// ShapeQueryExec -- program to perform shape queries on pointsets
// 
// Notes about writing queries:
// 1) fields in the file must be separated by a single tab.  blank
//   lines and lines beginning with # are ignored.
//
// 2) When specifying angles, pay attention to the direction of the
//    lines -- the angle is the DIFFERENCE between the first line and
//    the second line, i.e., the amount of counterclockwise rotation
//    required on the target edge to make it parallel (in the same
//    direction) to the relative edge

// some symbols not allowed in point names:
//   tab, space, dash, colon


import java.util.*;  // for vector
import java.io.*;    // for reading files

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;
import java.applet.*;
import java.net.*;
import java.awt.Toolkit;



// class to handle applet stuff and kick off the shqpe query process
public class ShapeQueryExec extends Applet
{
    ViewWindow window;
    ViewPanel viewPanel;

    public ShapeQueryExec()
    {
	if (runStandalone)
	{
	    execute();
	}
    }

    private void execute()
    {
	// ------------------------------------------------
	// output what environment we're working with

	if (runStandalone)
	{
	    OutputWriter.println("running as standalone application\n\n");
	}
	else
	{
	    OutputWriter.println("running as applet\n\n");
	}

	// ------------------------------------------------
	// figure out where to look for file inputs

	URL codeBase = null;
	if (!runStandalone)
	{
	    try
	    {
		codeBase = this.getCodeBase();
	    }
	    catch (java.lang.NullPointerException e)
	    {
		OutputWriter.println("cannot get doc base due to exception: "+e.getMessage());
		return;
	    }
	}


	// ------------------------------------------------
	// set up display

	// initialize viewpanel
	viewPanel = new ViewPanel(this, 
				  runStandalone,
				  codeBase);

	//	OutputWriter.println("codebase = ["+codeBase+"]");

	// decide whether to run in a window or as an applet
	window = null;
	if (runStandalone)
	{
	    window = new ViewWindow();
	    window.add(viewPanel);
	}
	else
	{
	    add(viewPanel);
	    repaint();
	}
    }

//    public static void main(String args[])
//    {
//	runStandalone = true;
//	ShapeQueryExec s = new ShapeQueryExec();
//    }

    public void init()
    {
	execute();
    }

    public void start()
    {
    }

    public void stop()
    {
    }

    public void destroy()
    {
    }

    public void paint(Graphics g)
    {
	if (viewPanel!=null)
	{
	    viewPanel.paint(g);
	}
    }

    public void repaint(Graphics g)
    {
	viewPanel.repaint();
    }

    int count;
    static boolean runStandalone;  // false by default
}

// if running standalone, need a Frame to put panels in
class ViewWindow extends Frame 
{
    int width;
    int height;

    ViewWindow()
    {
	// set up our window
	super("results");
	this.setTitle("Query Results");
	this.addWindowListener(new WindowAdapter()
	    {
		public void windowClosing(WindowEvent e)
		    {
			OutputWriter.println("window closed, exiting...");
			System.exit(1);
		    }
	    });

	width = 1000;
	height = 1000;
        setSize(width, height);
    }
}


// handle user input and do layout
class ViewPanel extends Panel
		implements DocumentListener
{
    // size of viewable area
    int width;
    int height;

    // -----------------------------------
    // basic classes

    // area we'll draw into
    DrawArea drawArea;

    // the universe of points we'll query
    PointUniverse universe;

    // the object we'll use to execute queries
    ShapeQuery query;

    // -----------------------------------
    // information about the current result set

    // what solution we're currently displaying
    int currentPointSet;

    int currentSolution;
    int currentSolutionOfTotal;
    int totalSolutionCount;

    // ----------------------------------------
    private Configger con;
    // things that should be read from a config file

    // names of shapes we might search for are read in from the config file
    private static final String USER_DEFINED = "User-Defined (Below)";


    // map query names to the files that back them
    HashMap<String, String> queryFileNames;


    // ----------------------------------------
    // user selecctions

    // parameters of the current search

    // the shape we're currently interested in searching for
    String currentSearch;

    // tolerance to use for executing queries
    float tolerance;
    TextField toleranceField; // where user enters tolerance

    // whether or not we should eliminate duplicate point tuples in
    // results
    boolean eliminateDuplicates;


    // the query this user has hand-entered; disabled because it wasn't working right
    String userDefinedQuery;

    // text area for user to enter his hand-made query
    TextArea queryInput;


    // ----------------------------------------
    // information about the available data sets 
    HashMap<String, DataSet> allowedDataSets;      // which data sets the user can choose from
    HashSet activeDataSets;       // which data sets we should use in this query
    java.awt.List dataSetChoices;  // list of choices (some will be selected)


    // ----------------------------------------

    // the points that are in this query (which the user might want to
    // display in the CSV output)
    String pointList[];         // names of points involved in query
    LinkedList constraintList;  // constraints (from which we figure out edges) in query


    // ----------------------------------------
    // applet stuff

    boolean runStandalone;
    URL codeBase;

    // make a new view area
    public ViewPanel(Applet newApplet, // need to pass applet for getImage()
		     boolean newRunStandalone,
		     URL newCodeBase)
    {
	width= 800;
	height = 800;
	setSize(width, height);
	
	runStandalone = newRunStandalone;
	codeBase = newCodeBase;

	// general information about results
	currentPointSet = 0;
	currentSolution = -1;
	totalSolutionCount = -1;

	// remember whether or not to eliminate duplicate results
	eliminateDuplicates = true;

	// initial tolerance
	tolerance = .05f;

        //Get the configger object that reads in the config.txt file
        try{
            con = new Configger("config_test_setnumber.txt", newCodeBase);
        }
        catch (IOException ioe)
        {
            System.out.println("Reading config file failed: " + ioe.getMessage());
            //bug out here
        }
        
	queryFileNames = con.getQueryMap();
	allowedDataSets = con.getDataSetMap();
        


	activeDataSets = new HashSet();

	// ==================================================================
	// tons of drawing stuff

	// -------------------------------------------
	// panel for drawing output

	drawArea = new DrawArea(newApplet,
				newRunStandalone,
				newCodeBase);

	// -------------------------------------------
	// panel for user input

	// ------------------------------------------------------------
	// make buttons

	Button nextButton = new Button("Next Result");
	nextButton.addActionListener(new ActionListener()
	      {
		public void actionPerformed(ActionEvent e)
		{
		    nextShape();
		}
	      }
	    );

	Button executeButton = new Button("Execute Query");
	executeButton.addActionListener(new ActionListener()
	      {
		public void actionPerformed(ActionEvent e)
		{
		    executeSelectedQuery();
		}
	      }
	    );


	dataSetChoices = new java.awt.List(4, true);

	String dataSetNames[] = (String[]) allowedDataSets.keySet().toArray( new String[0]);
	Arrays.sort(dataSetNames);
	for (int i = 0; i < dataSetNames.length; i++)
	{
	    dataSetChoices.add( dataSetNames[i]);
	}

	Button loadDataSetButton = new Button("Load Datasets");
	loadDataSetButton.addActionListener(new ActionListener()
	      {
		public void actionPerformed(ActionEvent e)
		{
		    loadDatasets();
		}
	      }
	    );

	Button chooseCSVOutput = new Button("CSV Output...");
	chooseCSVOutput.addActionListener(new ActionListener()
	    {
		public void actionPerformed(ActionEvent e)
		{
		    chooseCSVOutputWindow();
		}
	      }
	    );


	// a jcombobox in awt
	Choice desiredQuery = new Choice();
	desiredQuery.setSize(100, 25);
	desiredQuery.addItem("(Select A Query...)");
        String querynames[] = (String[])queryFileNames.keySet().toArray(new String[0]);
        Arrays.sort(querynames);
        for(int i = 0; i < querynames.length; i++){
            desiredQuery.addItem(querynames[i]);
        }


	desiredQuery.addItemListener(new ItemListener()
	    {
		public void itemStateChanged(ItemEvent e)
		{
		    currentSearch = (String)(e.getItem());
		}
	    }
				     );

	Checkbox eliminateDuplicatesCheckbox = new Checkbox("Discard Duplicate Results", null, true);

	eliminateDuplicatesCheckbox.addItemListener(new ItemListener()
	    {
		public void itemStateChanged(ItemEvent e)
		{
		    eliminateDuplicates = !eliminateDuplicates;
		}
	    }
				     );

	// tolerance input area
	Panel tolerancePanel = new Panel();
	tolerancePanel.setLayout(new BorderLayout());
	Label toleranceLabel = new Label("Tolerance:");
	tolerancePanel.add(toleranceLabel, BorderLayout.WEST);
	toleranceField = new TextField(""+tolerance);
	tolerancePanel.add(toleranceField, BorderLayout.CENTER);

	// get updates to this box
	toleranceField.addKeyListener(new KeyListener()
	    {
		public void keyPressed(KeyEvent e)
		{
		    tolerance = new Float(toleranceField.getText()).floatValue();
		}
		public void keyReleased(KeyEvent e)
		{
		    tolerance = new Float(toleranceField.getText()).floatValue();
		}
		public void keyTyped(KeyEvent e)
		{
		    tolerance = new Float(toleranceField.getText()).floatValue();
		}
	    } );


	Panel executeBar = new Panel(new BorderLayout());
	executeBar.add(desiredQuery, BorderLayout.CENTER);
	executeBar.add(executeButton, BorderLayout.EAST);
	executeBar.add(tolerancePanel, BorderLayout.SOUTH);

	Panel dataSetBar = new Panel(new BorderLayout());
	dataSetBar.add(loadDataSetButton, BorderLayout.EAST);
	dataSetBar.add(dataSetChoices, BorderLayout.CENTER);

	Panel buttonBar = new Panel(new GridLayout(0,1));
	buttonBar.add(dataSetBar);
	buttonBar.add(executeBar);

	Panel bottomBar = new Panel(new BorderLayout());
	bottomBar.add(eliminateDuplicatesCheckbox, BorderLayout.NORTH);
	bottomBar.add(nextButton, BorderLayout.WEST);
	bottomBar.add(chooseCSVOutput, BorderLayout.EAST);
	

	buttonBar.add(bottomBar, BorderLayout.SOUTH);

	// ------------------------------------------------------------
	// set up "console"

	TextArea textOutput = new TextArea("Console output:\n\n");
	textOutput.setColumns(60);
	textOutput.setRows(8);
    //OutputWriter.setOutput(textOutput);

	// set up area for user-defined query
	queryInput = new TextArea("#relative-edge |target-edge    |Dist/  |factor |differ-|toler- |absol-\n#start  end    |start  end     |Angle  |       |ence   |ance   |ute?\n");
	queryInput.setColumns(60);
	queryInput.setRows(4);
	queryInput.setFont(new Font("Monospaced",
				    Font.PLAIN, 12));
	queryInput.addTextListener(new TextListener()
	    {
		public void textValueChanged(TextEvent e)
		{
		    updateUserDefinedQuery();
		}
	      }
	    );


	Panel textBoxes = new Panel(new BorderLayout());
	textBoxes.add(textOutput, BorderLayout.CENTER);

	Panel selectionBar = new Panel(new BorderLayout());
	selectionBar.add(buttonBar, BorderLayout.WEST);
	selectionBar.add(textBoxes, BorderLayout.EAST);

	// -------------------------------------------
	// can't get scrolling to work
 	ScrollPane scrollPane = new ScrollPane(
				       ScrollPane.SCROLLBARS_ALWAYS);
	scrollPane.add(drawArea);
 	scrollPane.setSize(1000,1000);

	// -------------------------------------------
	setLayout(new BorderLayout());
	add(scrollPane, BorderLayout.CENTER);
	add(selectionBar, BorderLayout.NORTH);
	//	add(drawArea, BorderLayout.SOUTH);
        setVisible(true);
	repaint();

	// end drawing stuff
	// ---------------------------------------------------

	// -------------------------------------------
	// initialize point universe and query executor

	universe = new PointUniverse(newRunStandalone, newCodeBase);
	query = new ShapeQuery(newRunStandalone, newCodeBase);
    }

    void updateUserDefinedQuery()
    {
	userDefinedQuery = queryInput.getText();
// 	OutputWriter.println("got query: ["
// 			     +userDefinedQuery
// 			     +"]");
    }

    void chooseCSVOutputWindow()
    {
	Frame dialogFrame = new Frame("Select CSV Output Fields");
	CSVChooser chooser = new CSVChooser(this, dialogFrame,
					    pointList, constraintList,
					    universe,
					    query.edgesToDraw,
					    currentSearch);
    }

    // given a list of data set names to load, load the corresponding sets
    private void loadPointUniverse(String dataSets[])
    {
	// make a collection of pointsets.  this includes reading in
	// and storing the data, calculating relevant derived
	// information, and building appropriate indicies

        //universe.removeAllPointSets();  
	universe = new PointUniverse();
        String err = "";

	for (int i = 0; i < dataSets.length; i++)
	{
	    DataSet dataSet = (DataSet) allowedDataSets.get(dataSets[i]);
	    
	    if (dataSet==null)
	    {
		OutputWriter.print("Error: data set '"+dataSets[i]+"' not found.\n");
                err = " (with errors)";
	    }
	    else if (universe.addPointSet(dataSet.fileName, dataSet.visibleName) == null)
	    {
		OutputWriter.print("Error loading point set: '"+dataSet.visibleName+"'\n");
                err = " (with errors)";
	    }
	}
	OutputWriter.print("Load complete" + err + ".\n");
    }

    // run 'execute' with whatever query has been selected
    private void executeSelectedQuery()
    {
	if (currentSearch==USER_DEFINED)
	{
		executeQuery(false, "");
	}
	else 
	{
	    String searchFileName = (String)(queryFileNames.get(currentSearch));
	    if (searchFileName==null)
	    {
		OutputWriter.println("no file associated with search of type '"
				     +currentSearch+"'");
	    }
	    else
	    {
		executeQuery(true, searchFileName);
	    }
	}
    }

    // load the selected datasets
    private void loadDatasets()
    {
	String selectedSets[] = dataSetChoices.getSelectedItems();

	// forget old state
	drawArea.setPointSet(null);
	totalSolutionCount = 0;

	loadPointUniverse(selectedSets);
	
	garbageCollect();
    }

    void garbageCollect()
    {
	// force garbage collection to throw away old data structures.
	// i don't think this solved the problem.
	OutputWriter.print("requesting garbage collection...  ");
	System.gc();
	OutputWriter.print("done.");
    }


    // actually invoke the query execution
    private void executeQuery(boolean queryIsFileName,
			      String queryString)
    {
	query = new ShapeQuery(runStandalone, codeBase);
	garbageCollect();

	// ------------------------------------
	// first, figure out this query

	if (queryIsFileName)
	{
	    // read query from file
	    if (!query.readQueryFile(queryString, tolerance))
	    {
		OutputWriter.println("unable to read query file");
		return;
	    }
	}
	else // passed query directly
	{
	    if (!query.readQuery(queryString, tolerance))
	    {
		OutputWriter.println("unable to understand query");
		return;
	    }
	}

	// get back information so it can be given to user in CSV output
	pointList = query.getPointList();
	constraintList = query.getConstraints();

	// ------------------------------------
	// now proceed to try to run query

	if (universe.pointSets.size()<=0)
	{
	    OutputWriter.println("no data set to run query against");
	}
	else 
	{
	    // try running query
	    int resultCount = query.execute(universe, eliminateDuplicates);

	    if (resultCount==0)
	    {
		OutputWriter.print("no solutions\n");

		// remember how many results we got for later
		totalSolutionCount = resultCount;
	    }
	    else
	    {
		OutputWriter.print("\nSolution count: "+resultCount + "\n");
		currentPointSet = 0;
		currentSolution = -1;
		currentSolutionOfTotal = -1;
		
		setPointSet(  (PointSet)(universe.pointSets.elementAt(currentPointSet)));


		// remember how many results we got for later
		totalSolutionCount = resultCount;

		nextShape();  // go to the first shape
	    }
	}
    }


    // go to the next shape
    // 
    void nextShape()
    {
	if (totalSolutionCount<1)
	{
	    OutputWriter.println("no solutions to display");
	    return;
	}

	// this math should all be simplified.

	// try the next solution in this set
	currentSolution++;
	currentSolutionOfTotal++;

	if (currentSolution >= ( (PointSet)(universe.pointSets.elementAt(currentPointSet)) ).getSolutionSet().size())
	{
	    // we're done with solutions in the current point set
	    currentSolution=0;

	    // look for a new point set with an available solution
	    boolean foundNewPointSet = false;
	    for (int i = currentPointSet+1;
		 i < universe.pointSets.size();
		 i++)
	    {
		if ( ( (PointSet)(universe.pointSets.elementAt(i) )).getSolutionSet().size()>0)
		{
		    // we found a new pointset with some solutions
		    currentPointSet = i;
		    foundNewPointSet = true;
		    break;
		}
	    }

	    if (!foundNewPointSet)
	    {
		// try wrapping around to beginning and finding an
		// available solution

		currentSolutionOfTotal = 0;

		for (int i = 0;
		     i <= currentPointSet && i < universe.pointSets.size();
		     i++)
		{
		    if ( ( (PointSet)(universe.pointSets.elementAt(i) )).getSolutionSet().size()>0)
		    {
			// we found a new pointset with some solutions
			currentPointSet = i;
			foundNewPointSet = true;
			break;
		    }
		}
	    }

	    if (!foundNewPointSet)
	    {
		OutputWriter.println("Unable to locate additional solutions.");
	    }
	    else
	    {
		setPointSet(  (PointSet)(universe.pointSets.elementAt(currentPointSet)));
	    }
	}

	
	// now that we've figured out the right shape, go ahead and draw it
 	drawArea.setShape(
			  // thhe next QuerySolution
			  ((PointSet) (universe.pointSets.elementAt(currentPointSet))).getNextSolution(),

			  // the edges to display
			  query.getEdgesToDraw(),

			  currentSolution, // index of this current solution in this set

			  // solutions in this pointset
			  ((PointSet) (universe.pointSets.elementAt(currentPointSet))).getSolutionSet().size(),

			  currentSolutionOfTotal, // index of this solution among all solutions
			  
			  totalSolutionCount  // total solutions in all sets
			  );
    }

    void setPointSet(PointSet set)
    {
	drawArea.setPointSet(set);
    }



    // --------------------------------------------------------------------
    // for implementing documentlistener

    // Gives notification that an attribute or set of attributes changed.
    public void changedUpdate(DocumentEvent e)
    {
    }

    // Gives notification that there was an insert into the document.
    public void insertUpdate(DocumentEvent e)
    {
        try
	{
	    userDefinedQuery = e.getDocument().getText(
		e.getDocument().getStartPosition().getOffset(),
		e.getDocument().getEndPosition().getOffset());
	    //          System.out.println("tolerance="+tolerance);
	}
        catch (Exception e2)
	{
	    OutputWriter.println("input error");
	}
    }

    // Gives notification that a portion of the document has been removed.
    public void removeUpdate(DocumentEvent e)
    {
	try
	{
	    userDefinedQuery = e.getDocument().getText(
		e.getDocument().getStartPosition().getOffset(),
		e.getDocument().getEndPosition().getOffset());
	}
	catch (Exception e2)
	{
	    OutputWriter.println("input error");
	}
    }
}


class DrawArea extends Panel
    implements Scrollable // scrollable part not actually complete
{
    // size info
    int width, height;
    int imageHeight, imageWidth;

    // --------------------------------
    // current shape information
    String message;
    Image currentImage;
    PointPair currentShape[];    // edges comprising current shape
    PointSet pointSet;            // point space we're drawing in

    // the solution we're currently on
    QuerySolution currentSolution;

    // what shape to draw
    HashMap currentEdgesToDraw;


    // numbers the user wants to see
    int currentIndex;
    int solutionsInThisSet;
    int indexAmongTotal;
    int totalSolutions;

    // --------------------------------
    // cache information so we don't have to look it up a lot

    HashMap<String, Image> images;
    HashMap<Image, ImageSize> imageSizes;


    // --------------------------------
    // for running as applet
    boolean runStandalone;
    URL codeBase;
    Applet applet;

    public DrawArea(Applet newApplet,
		    boolean newRunStandalone,
		    URL newCodeBase)
    {
	width = 1000;
	height = 1000;

	imageWidth = 0;
	imageHeight = 0;

	setSize(width, height);
	setBackground(Color.white);
	setVisible(true);

	currentImage = null;
	message = "nothing to display";

	images = new HashMap<String, Image>();
	imageSizes = new HashMap<Image, ImageSize>();
	
	currentShape = null;
	pointSet = null;

        runStandalone = newRunStandalone;
        codeBase = newCodeBase;
	applet = newApplet;
    }

    void setMessage(String s)
    {
	message = s;
	repaint();
    }

    public void setShape(QuerySolution solution,
			 HashMap edgesToDraw,
			 int _currentIndex,
			 int _solutionsInThisSet,
			 int _indexAmongTotal,
			 int _totalSolutions)
    {
	// record drawing info
	currentSolution = solution;
	currentEdgesToDraw = edgesToDraw;

	// record solution # info for printing
	currentIndex = _currentIndex;	    
	indexAmongTotal = _indexAmongTotal;
	solutionsInThisSet = _solutionsInThisSet;
	totalSolutions = _totalSolutions;    

	OutputWriter.println("next shape...");

	repaint();
    }

    // reconfigure the current pointset.  if none, remove what's on the screen.
    public void setPointSet(PointSet set)
    {
	pointSet = set;
	if (set==null)
	{
	    currentImage = null;
	    currentShape = null;
	    currentSolution = null;
	}
	else
	{
	    currentImage = null;//getImage(set.imageFileName);
	}
	repaint();
    }

    // find image in our hash of already-opened images; if not found, load it
    private Image getImage(String imageFileName)
    {
	
	Image thisImage = (Image) (images.get(imageFileName));

	OutputWriter.println("looking up image '"+imageFileName
			     +"'...");

	if (thisImage==null) 
	{
	    // image not found -- open the file that contains the image	    
	    OutputWriter.println("image '"+imageFileName
				 +"' not found in cache, loading...");

	    
	    if (runStandalone)
	    {
		OutputWriter.println("  reading image directly from filesystem...");
		try
		{
		    thisImage = ImageIO.read(new File(imageFileName));
		    //	    width = image.getWidth();
		    //		    height = image.getHeight()+200;
		}
		catch(Exception e)
		{
		thisImage = null;
		OutputWriter.println("file read failed: "
				   +e.getMessage());
		//	    width = 600;
		//		    height = 200;
		}
	    }
	    else // running as applet
	    {
		String path = codeBase+imageFileName;

		OutputWriter.println("  reading image from url ["
			       +codeBase+imageFileName+"]");

		BufferedInputStream imgStream = new BufferedInputStream(
		    applet.getClass().getResourceAsStream(path));
		if (imgStream!=null) // check for error
		{
		    int availableBytes;
		    try 
		    {
			availableBytes = imgStream.available();
		    }
		    catch (java.io.IOException e)
		    {
			OutputWriter.println("stream failed: " + e.getMessage());
			availableBytes=-1;
		    }
		    if (availableBytes>0)
		    {
			byte buf[] = new byte[availableBytes];
			int bytesRead = -1;
			try
			{
			    bytesRead = imgStream.read(buf, 0, availableBytes);
			    imgStream.close();
			}
			catch (java.io.IOException ioe)
			{
			    OutputWriter.println("Couldn't read stream from file: " + path);
			    bytesRead = -1;
			}
			
			if (bytesRead<=0)
			{
			    OutputWriter.println("no image data");
			}
			else
			{
			    thisImage = Toolkit.getDefaultToolkit().createImage(buf);
			}
		    }
		    else
		    {
			OutputWriter.println("no data available");
		    }
		}
	    }

	    images.put(imageFileName, thisImage);
	    
	} // end of image read

 	thisImage.getWidth(this);
 	thisImage.getHeight(this);

	return thisImage;
    }	
	
    private void drawPoints(Graphics2D g2d)
    {
	// --------------------------------------------------------------
	// draw points in pointset

	if (pointSet==null)
	    return;

	g2d.setColor(Color.cyan);

	Iterator iter = pointSet.points.iterator();

	while (iter.hasNext())
	{
	    Point p = (Point)iter.next();
 	    int x = (new Float(p.getX())).intValue();
 	    int y = (new Float(p.getY())).intValue();
	    g2d.fillOval(x-3, 
			 imageHeight-y-3,
			 6, 6);
	}
    }

    // redraw our environment.  !!! this takes a long time -- should
    // probably break this into separate functions, and when the shape
    // changes, redraw only the part that changed.
    public void paint(Graphics g)
    {
        super.paint(g);

        Graphics2D g2d=(Graphics2D) g;

	// ----------------------------------------------------------------
	// print messages first
	if (currentImage!=null)
	{
	    g.drawImage(currentImage, 0,0, this);

	    // !!! there's a bug here -- if we don't know the size of
	    // the image, we should do something smarter than nothing.
	    //
	    // (fortunately all the images are the same size right
	    // now, so it just works)
	    if (imageSizes.keySet().contains(currentImage))
	    {
		imageWidth = ((ImageSize)(imageSizes.get(currentImage))).width;
		imageHeight = ((ImageSize)(imageSizes.get(currentImage))).height;
	    }

	    g.drawString("Solution: "+(indexAmongTotal+1)+" of "+totalSolutions,
			 100, 50);

	    g.drawString("Solution in this Set: "+(currentIndex+1)+" of "+solutionsInThisSet,
			 250, 50);

 	    g.drawString("Set Name: "+pointSet.getVisibleName(),
			 500, 50);
	}
	else
	{
 	    g.drawString(message, width/2, height/2);
	}


	// ----------------------------------------------------------------
	// next, draw all labelled points
	drawPoints(g2d);

	// ----------------------------------------------------------------
	// last, if we have a shape to draw, draw it
	if (currentSolution!=null)
	{
	    g2d.setStroke(new BasicStroke(3));  // thick lines
	    
	    // iterate over all edges, drawing them
	    Iterator edgeIter = currentEdgesToDraw.keySet().iterator();

	    while (edgeIter.hasNext())
	    {
		// ------------------------------
		// get the current edge (this is a set containing two
		// point names), e.g. a set {"p1", "p2"}
		HashSet currentEdge = (HashSet)edgeIter.next();

		// ------------------------------
		// for the current edge, get the two points
		String firstPointName = null;
		String secondPointName = null;

		if (currentEdge.size()!=2)
		{
		    OutputWriter.println("yikes!  wrong number of points in current edge ("
					 +currentEdge.size()+", should be 2)");
		}

		Iterator pointIter = currentEdge.iterator();
		
		while (pointIter.hasNext())
		{
		    if (firstPointName == null)
			firstPointName=(String)pointIter.next();
		    else
			secondPointName=(String)pointIter.next();
		}

		// ------------------------------
		// get the color too
		Color currentColor = (Color)currentEdgesToDraw.get(currentEdge);

		// ------------------------------
		// now look up the point locations corresponding to these point names

		Point firstPoint = (Point)currentSolution.getPointNames().get(firstPointName);
		Point secondPoint = (Point)currentSolution.getPointNames().get(secondPointName);

		try
		{
		    int x1 = (new Float(firstPoint.getX())).intValue();
		    int y1 = (new Float(firstPoint.getY())).intValue();
		    int x2 = (new Float(secondPoint.getX())).intValue();
		    int y2 = (new Float(secondPoint.getY())).intValue();

		    g2d.setColor(currentColor);
		    g2d.drawLine(x1, imageHeight-y1, x2, imageHeight-y2);
		}
		catch (java.lang.NullPointerException e)
		{
		    // this indicates the solution doesn't include all
		    // the lines that we expected
		    OutputWriter.println("Warning: solution is missing a line (line segment is null)");
		}

	    }
	}
    }


    // scrolling stuff -- not implemented
    public Dimension getPreferredScrollableViewportSize()
    {
	return new Dimension(300,700);
    }

    public int getScrollableBlockIncrement(Rectangle visibleRect,
				    int orientation,
				    int direction) 
    {
	return 50;
    }

    public boolean getScrollableTracksViewportHeight() 
    {
	return true;
    }

    public boolean getScrollableTracksViewportWidth() 
    {
	return true;
    }

    public int getScrollableUnitIncrement(Rectangle visibleRect,
				   int orientation,
				   int direction)
    {
	return 50;
    }


    // to get image sizes
    public boolean imageUpdate(Image _img, int _infoflags,
			int _x, int _y, int _width, int _height) 
    {
	// we want size of image.

	// parent needs to be invoked to grab actual image
	boolean ret =
	    super.imageUpdate(_img, _infoflags, _x, _y, _width, _height);

//  	OutputWriter.println("imageUpdate called (width="
//  			     +_width+", height="
//  			     +_height+")");
	
	imageWidth = _width;
	imageHeight = _height;
	
	ImageSize imageSize = new ImageSize();
	imageSize.width = _width;
	imageSize.height = _height;
	
	imageSizes.put(_img, imageSize);

	return ret;
    }
}

class ImageSize
{
    int width;
    int height;
}

// ask user what fields he wants to see in CSV output
class CSVChooser extends Dialog implements ItemListener
{
    // what items have been selected
    HashSet<String> fieldsToDisplay;

    private ViewPanel parent;
    private PointUniverse universe;
    private String currentSearch;

    CSVChooser(ViewPanel _parent, Frame frame,
	       String points[],
	       LinkedList constraints,
	       PointUniverse _universe,
	       HashMap edgesToDraw,
	       String _currentSearch)
    {
	super(frame, "Select CSV Outputs...", true);
	fieldsToDisplay = new HashSet<String>();
	parent = _parent;
	universe = _universe;
	currentSearch = _currentSearch;

	// what choices we'll offer
	Vector<String> lengths = new Vector<String>();
	Vector<String> angles = new Vector<String>();
	Vector<String> ratios = new Vector<String>();

	Panel pointPanel = null;
	Panel lengthPanel = null;
	Panel anglePanel = null;
	Panel ratioPanel = null;
	Panel colorPanel = null;

	if (points == null)
	{
	    Label label = new Label("Must execute a query first");
	    add(label, BorderLayout.CENTER);
	}
	else // valid query has been executes, so we can give choices
	{
	    // ------------------------------------------------
	    // figure out what constraints to offer as csv choices
	    for (Iterator iter = constraints.iterator();
		 iter.hasNext();
		)
	    {
		Constraint currentConstraint = (Constraint) iter.next();

		if (currentConstraint.myRuleType==Constraint.RULE_TYPE_DISTANCE)
		{
		    // make a choice for the required edge of this constraint
		    lengths.add("Length "
				+currentConstraint.myRequiredEdgeStartPointName
				+" "
				+currentConstraint.myRequiredEdgeEndPointName);

		    // next make a choice for the target edge of this constraint
		    lengths.add("Length "
				+currentConstraint.myTargetEdgeStartPointName
				+" "
				+currentConstraint.myTargetEdgeEndPointName);

		    // last, make a choice for the ratio of these two
		    // edges (allow inverse ratio as well)
		    ratios.add("Ratio "
			       +currentConstraint.myTargetEdgeStartPointName
			       +"-"
			       +currentConstraint.myTargetEdgeEndPointName
			       +":"
			       +currentConstraint.myRequiredEdgeStartPointName
			       +"-"
			       +currentConstraint.myRequiredEdgeEndPointName);

		    ratios.add("Ratio "
			       +currentConstraint.myRequiredEdgeStartPointName
			       +"-"
			       +currentConstraint.myRequiredEdgeEndPointName
			       +":"
			       +currentConstraint.myTargetEdgeStartPointName
			       +"-"
			       +currentConstraint.myTargetEdgeEndPointName);

		}
		else if (currentConstraint.myRuleType==Constraint.RULE_TYPE_ANGLE)
		{
		    // make a choice for the required edge of this constraint
		    angles.add("Angle "
			       +currentConstraint.myTargetEdgeStartPointName
			       +"-"
			       +currentConstraint.myTargetEdgeEndPointName
			       +" "
			       +currentConstraint.myRequiredEdgeStartPointName
			       +"-"
			       +currentConstraint.myRequiredEdgeEndPointName);


		    angles.add("Angle "
			       +currentConstraint.myRequiredEdgeStartPointName
			       +"-"
			       +currentConstraint.myRequiredEdgeEndPointName
			       +" "
			       +currentConstraint.myTargetEdgeStartPointName
			       +"-"
			       +currentConstraint.myTargetEdgeEndPointName);
		}
	    }

	    // give point choices
	    pointPanel = new Panel();  // overall panel
	    Panel pointSelectPanel = new Panel();  // sub-panel with choices
	    pointPanel.setLayout(new BorderLayout());
	    pointSelectPanel.setLayout(new FlowLayout());
	    Label label = new Label("Point Coordinates:");
	    pointPanel.add(label, BorderLayout.NORTH);
	    
	    Arrays.sort(points);
	    for (int i = 0; i < points.length; i++)
	    {
		Checkbox b = new Checkbox("Point "+points[i]);
		pointSelectPanel.add(b);
		b.addItemListener(this);
	    }
	    pointPanel.add(pointSelectPanel, BorderLayout.CENTER);


	    // give length choices
	    lengthPanel = new Panel();  // overall panel
	    Panel lengthSelectPanel = new Panel();  // sub-panel with choices
	    lengthPanel.setLayout(new BorderLayout());
	    lengthSelectPanel.setLayout(new FlowLayout());
	    label = new Label("Edge Lengths:");
	    lengthPanel.add(label, BorderLayout.NORTH);

	    String lengthsArray[] = (String[]) lengths.toArray(new String[0]);
	    Arrays.sort(lengthsArray);
	    for (int i = 0; i < lengthsArray.length; i++)
	    {
		Checkbox b = new Checkbox(lengthsArray[i]);
		lengthSelectPanel.add(b);
		b.addItemListener(this);
	    }
	    lengthPanel.add(lengthSelectPanel, BorderLayout.CENTER);


	    // give angle choices
	    anglePanel = new Panel();  // overall panel
	    Panel angleSelectPanel = new Panel();  // sub-panel with choices
	    anglePanel.setLayout(new BorderLayout());
	    angleSelectPanel.setLayout(new FlowLayout());
	    label = new Label("Angles:");
	    anglePanel.add(label, BorderLayout.NORTH);

	    String anglesArray[] = (String[]) angles.toArray(new String[0]);
	    Arrays.sort(anglesArray);
	    for (int i = 0; i < anglesArray.length; i++)
	    {
		Checkbox b = new Checkbox(anglesArray[i]);
		angleSelectPanel.add(b);
		b.addItemListener(this);
	    }
	    anglePanel.add(angleSelectPanel, BorderLayout.CENTER);


	    // give ratio choices
	    ratioPanel = new Panel();  // overall panel
	    Panel ratioSelectPanel = new Panel();  // sub-panel with choices
	    ratioPanel.setLayout(new BorderLayout());
	    ratioSelectPanel.setLayout(new FlowLayout());
	    label = new Label("Edge Ratios:");
	    ratioPanel.add(label, BorderLayout.NORTH);

	    String ratiosArray[] = (String[]) ratios.toArray(new String[0]);
	    Arrays.sort(ratiosArray);
	    for (int i = 0; i < ratiosArray.length; i++)
	    {
		Checkbox b = new Checkbox(ratiosArray[i]);
		ratioSelectPanel.add(b);
		b.addItemListener(this);
	    }
	    ratioPanel.add(ratioSelectPanel, BorderLayout.CENTER);

	    // ----------------------------------------
	    // show what colors correspond to what edges

	    colorPanel = new Panel();
	    colorPanel.setLayout(new BorderLayout());
	    colorPanel.add(new Label("Colors of Each Edge:"), BorderLayout.NORTH);

	    Panel colorLabelPanel = new Panel();
	    colorLabelPanel.setLayout(new FlowLayout());

	    // we'll put edge labels into map so we can sort them
	    // before we print them
	    HashMap<String, Label> edgeLabels = new HashMap<String, Label>();

	    // --------------------
	    // first figure out what labels to print (and set their
	    // colors appropriately)
	    
	    for (Iterator i = edgesToDraw.keySet().iterator();
		 i.hasNext();
		 )
	    {
		// this is the list of 2 points in an edge that we drew
		HashSet edge = (HashSet) i.next();

		// get the point names from this hashset
		String firstPointName = null;
		String secondPointName = null;

		Iterator pointIter = edge.iterator();
		while (pointIter.hasNext())
		{
		    if (firstPointName == null)
			firstPointName=(String)pointIter.next();
		    else
			secondPointName=(String)pointIter.next();
		}

		// figure out the name of this label -- it's the two
		// points together, but we want to print them in
		// order, e.g. "p1-p2" not "p2-p1".
		String name = firstPointName+"-"+secondPointName;
		if (firstPointName.compareTo(secondPointName)>0)
		{
		    name = secondPointName+"-"+firstPointName;
		}

		// make a label out of the names of the two points in
		// this edge; set color to the color that this edge is
		// in the drawing.
		label = new Label(name);
		label.setForeground((Color)edgesToDraw.get(edge));

		// remember this label so we can print it after we
		// sort
		edgeLabels.put(name, label);
	    }

	    // --------------------
	    // now sort the labels by name
	    
	    String sortedNames[] = (String[]) edgeLabels.keySet().toArray(new String[] {});
	    Arrays.sort(sortedNames);

	    for (int i = 0; i < sortedNames.length; i++)
	    {
		colorLabelPanel.add((Label) edgeLabels.get(sortedNames[i]));
	    }

	    colorPanel.add(colorLabelPanel, BorderLayout.CENTER);

	} // end valid query has been executed so we can give choices
	setLayout(new BorderLayout());

	Panel selectPanel = new Panel();
	selectPanel.setLayout(new GridLayout(0,1));
	if (pointPanel!=null)
	{
	    selectPanel.add(pointPanel);
	    selectPanel.add(lengthPanel);
	    selectPanel.add(anglePanel);
	    selectPanel.add(ratioPanel);
	    selectPanel.add(colorPanel);
	}
	add(selectPanel, BorderLayout.CENTER); 


	Button cancelButton = new Button("Cancel");
	cancelButton.addActionListener(new ActionListener()
	      {
		public void actionPerformed(ActionEvent e)
		{
		    setVisible(false);
		}
	      }
	    );

	Button makeCSVButton = new Button("Display Output");
	makeCSVButton.addActionListener(new ActionListener()
	      {
		public void actionPerformed(ActionEvent e)
		{
//		    parent.setCSVOutput(fieldsToDisplay);
		    setVisible(false);

		    Frame dialogFrame = new Frame("CSV Output");
		    CSVOutput output = new CSVOutput(parent,
						     dialogFrame,
						     universe,
						     currentSearch,
						     fieldsToDisplay);
		    

		}
	      }
	    );

	//	Dimension dim = getToolkit().getScreenSize();
	//	d.setLocation(dim.width/4,dim.height/3);	

	Panel buttonPanel = new Panel(new FlowLayout());
	buttonPanel.add(cancelButton);
	buttonPanel.add(makeCSVButton);
	add(buttonPanel, BorderLayout.SOUTH);
	setSize(700,500);
	setVisible(true);
    }

    public void itemStateChanged(ItemEvent e)
    {
	boolean selected = (e.getStateChange()==ItemEvent.SELECTED);

	String name = (String) e.getItem();

	if (selected)
	{
	    fieldsToDisplay.add(name);
	}
	else
	{
	    fieldsToDisplay.remove(name);
	}
    }
}


// information about a dataset to load
class DataSet
{
    String visibleName;     // name to show user
    String fileName;        // file of point coordinates
    String imageFileName;   // file of image to display as backdrop

    DataSet(String _fileName,
	    String _imageFileName,
	    String _visibleName)
    {
	visibleName = _visibleName;
	fileName = _fileName;
	imageFileName = _imageFileName;
    }
}


// display the output that the user requested in csv format
class CSVOutput extends Dialog
{
    private ViewPanel parent;

    CSVOutput(ViewPanel _parent, 
	      Frame frame,
	      PointUniverse universe,
	      String currentSearch,
	      HashSet<String> desiredCSVOutput)
    {
	super(frame, "CSV Output", true);
	parent = _parent;
	setLayout(new BorderLayout());

	Iterator setIter = universe.pointSets.iterator();
	
	int csvRowCount = 0;

	TextArea CSVTextOutput = new TextArea();

	// print header
	CSVTextOutput.append("Query: "+currentSearch+",Set Name,");
	String[] csvOutputFields = (String[]) desiredCSVOutput.toArray( new String[0]);

	Arrays.sort(csvOutputFields);

	for (int i = 0; i < csvOutputFields.length; i++)
	{
	    String currentField = csvOutputFields[i];
	    if (currentField.matches("^Angle.*") ||
		currentField.matches("^Length.*") ||
		currentField.matches("^Ratio.*"))
	    {
		CSVTextOutput.append(currentField);
	    }
	    else
	    {
		CSVTextOutput.append(currentField+"-x,"
				   +currentField+"-y");
	    }

	    // only print comma if this isn't the last field
	    if (i < csvOutputFields.length - 1)
		CSVTextOutput.append(",");
	}
	CSVTextOutput.append("\n");

	// look through all pointsets in universe
	while (setIter.hasNext())
	{
	    PointSet set = (PointSet) setIter.next();
	    
	    Iterator solIter = set.getSolutionSet().iterator();

	    // look through all solutions for this pointset
	    while (solIter.hasNext())
	    {
		QuerySolution currentSolution = (QuerySolution) solIter.next();


		csvRowCount++;
		CSVTextOutput.append("Solution #"+csvRowCount+","
				   +set.getVisibleName()+",");

		for (int i = 0; i < csvOutputFields.length; i++)
		{
		    String currentField = csvOutputFields[i];
		    String items[] = currentField.split(" ");

		    if (items[0].equals("Angle"))
		    {
			// angles are specified as pairs of points
			// separated by dashes

			// grab first set of points
			String subItems[] = items[1].split("-");
			String edge1point1 = subItems[0];
			String edge1point2 = subItems[1];

			// grab second set of points
			subItems = items[2].split("-");
			String edge2point1 = subItems[0];
			String edge2point2 = subItems[1];

			PointPair firstPair = new PointPair(
			    (Point)currentSolution.getPointNames().get(edge1point1)
			    ,
			    (Point)currentSolution.getPointNames().get(edge1point2)
			    );

			PointPair secondPair = new PointPair(
			    (Point)currentSolution.getPointNames().get(edge2point1)
			    ,
			    (Point)currentSolution.getPointNames().get(edge2point2)
			    );

			CSVTextOutput.append(""+
				ShapeQuery.angleAdjust(firstPair.angle-secondPair.angle));

		    }
		    else if (items[0].equals("Length"))
		    {
			PointPair pair = new PointPair(
			    (Point)currentSolution.getPointNames().get(items[1])
			    ,
			    (Point)currentSolution.getPointNames().get(items[2])
			    );
			
			CSVTextOutput.append(""+pair.distance);
		    }
		    else if (items[0].equals("Ratio"))
		    {
			// ratios are specified as pairs of points
			// separated by dashes; the pairs are separated by a colon

			String subItems[] = new String[4];

			// precedes dash and colon
			subItems[0] = (items[1].split(":"))[0].split("-")[0];
			subItems[1] = (items[1].split(":"))[0].split("-")[1];
			subItems[2] = (items[1].split(":"))[1].split("-")[0];
			subItems[3] = (items[1].split(":"))[1].split("-")[1];

			PointPair firstPair = new PointPair(
			    (Point)currentSolution.getPointNames().get(subItems[0])
			    ,
			    (Point)currentSolution.getPointNames().get(subItems[1])
			    );

			PointPair secondPair = new PointPair(
			    (Point)currentSolution.getPointNames().get(subItems[2])
			    ,
			    (Point)currentSolution.getPointNames().get(subItems[3])
			    );
			
			float ratio = firstPair.distance/secondPair.distance;
			
			CSVTextOutput.append(""+ratio);
		    }
		    else // this is a point
		    {
			Point currentPoint = 
			    (Point)currentSolution.getPointNames().
			    get(items[1]);
			CSVTextOutput.append(currentPoint.getX()+","+currentPoint.getY());
		    }

		    // only print comma if this isn't the last field
		    if (i < csvOutputFields.length - 1)
			CSVTextOutput.append(",");
		}

		CSVTextOutput.append("\n");
	    }
	}


	Button doneButton = new Button("Done");
	doneButton.addActionListener(new ActionListener()
	      {
		public void actionPerformed(ActionEvent e)
		{
		    setVisible(false);
		}
	      }
	    );

	add(CSVTextOutput, BorderLayout.CENTER);
	add(doneButton, BorderLayout.SOUTH);
	setSize(600,400);
	setVisible(true);
    }
}
