/*
 * ShapeQueryExecuter.java
 *
 * Created on February 21, 2005, 9:21 PM
 * Provides functionality for simple execution of queries.
 */
package ShapeQuery;


import java.io.Writer;
import java.io.IOException;
import java.io.File;
import java.io.FileWriter;
import java.util.Iterator;
import java.util.HashMap;
import java.util.LinkedList;

/**
 * Provides functionality for simple execution of queries.
 * @author David Hessing
 */

/** Class to simplify making queries */
public class ShapeQueryExecuter {

    //public static final String DTD_FILE = "results.dtd";
    private PointUniverse universe;
    private ShapeQuery query;
    private int solutionCount; //number of solutions to the query last time we ran it.
    private long[] timer; //times for plan time and execution time for last time we ran this query.
    
    /** Creates a new instance of ShapeQueryExecuter */
    public ShapeQueryExecuter() {
        solutionCount = -1;
        timer = new long[2];
    }
    

    /**
     * Set the point universe. Must be done before executing a query.
     * 
     * @param u A pointuniverse that contains loaded datasets.
     */
    public void setUniverse(PointUniverse u)
    {
        universe = u;
    }
    
    /**
     * Set the query. Must be done before executing a query.
     * 
     * @param q A query object that has already loaded a query from some source.
     */
    public void setQuery (ShapeQuery q)
    {
        query = q;
    }
    
    /**
     * Execute the query on the universe using the query object. Eliminates
     * duplicates according to what was set in the query XML (default is true).
     * 
     * @return The number of results, or -1 on an error.
     */
    public int executeQuery()
    {
        //check that the universe and the query objects are not null.
        if (!checkUQ()) {return -1;}
        solutionCount = query.execute(universe, timer);
        return solutionCount;
    }
    

    /**
     * Execute the query on the universe using the query object.
     * 
     * @param eliminateDuplicates Specifies whether duplicate results should be removed from the resultset.
     * @return The number of results, or -1 on an error.
     */
    public int executeQuery(boolean eliminateDuplicates)
    {
        //check that the universe and the query objects are not null.
        if (!checkUQ()) {return -1;}
        solutionCount = query.execute(universe, eliminateDuplicates, timer);
        return solutionCount;
    }

    /**
     * Checks that the query and universe objects are not null and that
     * the universe object has pointsets loaded.
     *
     * @return true if the two objects exist and the universe contains pointsets. False otherwise.
     */
    private boolean checkUQ() {
        if (universe == null || query == null)
        {
            OutputWriter.printerror("ShapeQueryExecuter: Either the universe or query is null. Set these before calling executeQuery().\n");
            return false;
        }
	if (universe.pointSets.size()<=0)
	{
	    OutputWriter.printerror("ShapeQueryExecuter: The point universe is empty. No data set to run query against.\n");
            return false;
	}
       return true;
    }
    
    
    /**
     * Writes the results of the most recently executed query out as SVG to files 
     * written to the specified path. Additionally creates a "results.html" file
     * that contains all the generated SVG files in <object> tags for viewing.
     *
     * @return True on successful write. False otherwise.
     */
    public boolean writeSVGresults(String path) {
        if (solutionCount == -1) {
            OutputWriter.printerror("ShapeQueryExecuter: No queries have been run yet.\n");
            return false;
        }
        //if (!path.endsWith("\\")) 
        //    path += "\\";
        
        //delete all old results sitting here
        File f = new File(path);
        String old[] = f.list();
        if (old != null) {
            for (int i = 0; i < old.length; i++) {
                if (old[i].endsWith(".svg")) {
                    f = new File(path + old[i]);
                    f.delete();
                }
            }
        }
        
        LinkedList<String> filenames = new LinkedList<String>();
        try {
            for (int i = 0; i < universe.getPointSetCount(); i++) {
                //get the pointset and queue up the name of it's result file
                PointSet ps = universe.getPointSetAt(i);
                String filename = ps.getVisibleName() + "_results.svg";
                filenames.add(filename);

                //get a description of these results for display in the SVG
                String desc = "Query \"" + query.getQueryName() + "\" executed on Pointset \"" + ps.getVisibleName() + "\". ";
                desc += (query.getDuplicateStatus()) ? "Duplicates eliminated." : "Duplicates NOT eliminated.";
                
                //get the image file behind the church
                File gif = null; String gifName = "";
                if(QueryBean.SERVER) {
                    gif = new File(path + "pics/" + ps.getVisibleName() + ".gif");
                    gifName = new String("pics/" + ps.getVisibleName() + ".gif");
                }
                else {
                    gif = new File(path + "\\pics\\" + ps.getVisibleName() + ".gif");
                    gifName = gif.toString();
                }
                if(!gif.exists()) {
                   gif=new File("");
                   gifName = "";
                }
                                
                //write out this set's svg
                FileWriter fw = new FileWriter(path + filename, false);
                QuerySolutionSet.writeSVGSolutions(desc, path, gifName, ps.getSolutionSet(), ps, fw, 1.0f, 50.0f);
            }
            //now write out an html file to tie them all together
            FileWriter fw = new FileWriter(path + "results.html", false);
            fw.write("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n");
            fw.write("<html>\n<head><title>Results of Query \"" + query.getQueryName() + "\"</title></head>\n<body>");
            fw.write("<h2>Results of Query \"" + query.getQueryName() + "\"</h2>");
            //fw.write("<br>Click in the blue box in the upper left to scroll through results.<br>");
            fw.write("A viewer is required to see SVG files. If you do not have one, click <a href=\"http://www.adobe.com/svg\">here</a>.");
            //fw.write("<br><br>Additionally, you can view the log files and see the SVG generated at each step of query processing <a href=\"log/\">here</a>. Note that currently the server does not correctly deliver the SVG files, and you may see an XML error. If this occurs download them locally.");
            fw.write("<br><br>");
            
            Iterator iter = filenames.iterator();
            int i=0;
            while (iter.hasNext()) {
                String filename = (String)iter.next();
                fw.write("<table width=\"1000\" border=\"1\">");
                fw.write("<tr>");
                fw.write("<td width=\"850\" valign=\"left\" nowrap>");               
                    
                    //fw.write("<li><a href=\"" + filename + "\">" + filename + "</a></li>");
                    //fw.write("<object data=\"" + filename + "\" width=\"800\" height=\"600\" type=\"image/svg+xml\"></object>\n");
                    fw.write("<embed src=\"" + filename + "\" width=\"850\" height=\"650\" type=\"image/svg+xml\">\n");
                    
                fw.write("</td>");

                if(QueryBean.STATISTICAL_ANALAYSIS) {
                    PointSet ps = universe.getPointSetAt(i++);
                    int resultCount = (ps.getSolutionSet()).size();
                    
                    fw.write("<form name=\"form"+i+"\" method=\"post\" action=\"submitResults.jsp\">");                   
                    fw.write("<input type=\"hidden\" name=\"churchName\" value=\"" + ps.getVisibleName() + "\">");
                    fw.write("<input type=\"hidden\" name=\"queryName\" value=\"" + query.getQueryName() + "\">");
                    fw.write("<input type=\"hidden\" name=\"resultCount\" value=\"" + resultCount + "\">");

                    fw.write("<td width=\"150\" valign=\"left\" nowrap>");
                    fw.write("<font size=\"2\" face=\"Arial\">");
                
                    for(int j=1; j <=resultCount; j++) {
                        fw.write("<input type=\"checkbox\" value=1 name=\"Result"+j+"\" checked/>  Result "+j+"<br>");
                        if(j%30==0)
                            fw.write("</td><td width=\"150\" valign=\"left\" nowrap>");
                            fw.write("<font size=\"2\" face=\"Arial\">");
                    }
                
                    fw.write("<br><br><input type=\"submit\" value=\"Submit Results\"/>");
                    fw.write("</td>");

                    fw.write("</form>");    
                }
                fw.write("</tr");
                fw.write("</table");
            }
                        
            fw.write("</body></html>");
            fw.flush();
            fw.close();
        }
        catch(IOException ioe) {
            OutputWriter.printerror("ShapeQueryExecuter: Bad stuff happened while writing SVG results - " + ioe.getMessage());
            return false;
        }
        //write the results
        return true;
    }
    
    
    /**
     * Writes the results of the most recently executed query out as XML to the given Writer.
     *
     * @param writer Where to write the results to. 
     * @returns True on successful write. False otherwise.
     */
    public boolean writeXMLresults(Writer writer) {
        if (solutionCount == -1) {
            OutputWriter.printerror("ShapeQueryExecuter: No queries have been run yet.\n");
            return false;
        }
        int countcheck = 0;
        int numDatasets = universe.getPointSetCount();
        
        try {
            writer.write("<?xml version=\"1.0\"?>\n");
            //writer.write("<!DOCTYPE results SYSTEM \"" + DTD_FILE +"\">\n");
            writer.write("<results>\n");
            
            writeXMLQueryInfo(writer);
            for(int i = 0; i < numDatasets; i++) {
                writeXMLResultSet(writer, universe.getPointSetAt(i));
            }
            writer.write("</results>\n");
            writer.flush();
            writer.close();
        }
        catch (IOException ioe) {
            OutputWriter.printerror("ShapeQueryExecuter: Error writing to Writer: " + ioe.getMessage() + ".\n");
            return false;
        }
  
        return true;
    }

    
    /**
     * Writes out the <query></query> section of the results XML.
     * @param writer Where to write the to.
     */
    private void writeXMLQueryInfo(Writer writer) throws IOException {
        writer.write("  <query>\n");
        writer.write("    <queryname>" + query.getQueryName() + "</queryname>\n");
        writer.write("    <removeduplicates>" + query.getDuplicateStatus() + "</removeduplicates>\n");      
        writer.write("    <indextime>" + universe.getTotalIndexTime() + "</indextime>\n");
        writer.write("    <csstreetime>" + universe.getTotalCssTreeBuildTime() + "</csstreetime>\n");
        writer.write("    <plantime>" + timer[0] + "</plantime>\n");
        writer.write("    <executiontime>" + timer[1] + "</executiontime>\n");
        writer.write("    <resultcount>" + solutionCount + "</resultcount>\n");        
        writer.write("  </query>\n");
    }
     
    /**
     * Writes out the <resultset></resultset> section of the results XML.
     * @param writer Where to write the to.
     * @param ps A PointSet with an existing QuerySolutionSet.
     */
    private void writeXMLResultSet(Writer writer, PointSet ps) throws IOException {
        writer.write("  <resultset pointset=\""+ ps.getVisibleName() +"\" usedCached=\"" + ps.fromCache() + "\">\n");
        QuerySolutionSet qse = ps.getSolutionSet();
        //HashMap edgesToDraw = qse.getEdgesToDraw();
        Iterator iter = qse.listIterator();
        
        QuerySolution qs = null;
        Iterator piter = null;
        Point p = null;
        String pname = "";

        //Upload the Results into the Database, if the Statistical Analysis is being performed
        int resultCounter=0;
        //System.out.println(QueryBean.STATISTICAL_ANALAYSIS);
        if(QueryBean.STATISTICAL_ANALAYSIS)
            QueryBean.db.deleteResultSet(ps.getVisibleName(), query.getQueryName());
        
        while (iter.hasNext()) {
            qs = (QuerySolution)iter.next();
            resultCounter++;
            //HashMap pointNames = qs.getPointNames();
            piter = (qs.getPointNames()).keySet().iterator();

            writer.write("    <shape>\n");
            while (piter.hasNext()) {
                pname = (String)piter.next();
                p = (Point) qs.getPoint(pname);
                writer.write("      <point handle=\"" + pname + "\"");
                if (p.hasExternalID()) {
                  writer.write(" id=\"" + p.getExternalID() + "\"/>\n"); }
                else { 
                    writer.write(">\n"); 
                    writer.write("        <x>" + p.getX() + "</x>\n");
                    writer.write("        <y>" + p.getY() + "</y>\n");
                    writer.write("      </point>\n");
                }

                //Upload the Results into the Database, if the Statistical Analysis is being performed
                if(QueryBean.STATISTICAL_ANALAYSIS)
                    QueryBean.db.loadResultSet(ps.getVisibleName(), query.getQueryName(), resultCounter, pname, p.getExternalID());

            }
            writer.write("    </shape>\n");
        }
        writer.write("  </resultset>\n");
        
        
    }
    
}




