package ShapeQuery;
// ShapeQueryPlan class -- given some ordered constraints, provide a
//                         plan for performing that query while
//                         respecting that order


import java.util.*;  // for vector

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;


class ShapeQueryPlan implements Cloneable
{
    // Rules, in order, to execute to perform query.  
    Vector rules;             

    // -----------------------------------
    // points and constraints
    
    // while we're building query, we keep track of which points we've
    // already found ("known").  the general principal is that
    //    {some known points} + {a constraint} -> find {an unknown point}
    HashSet knownPoints;      // points already found by our
			      // current set of rules

    // the constraints used in this plan.  these are used so we can
    // pick the best plan of all those that implement the same set of
    // constraints.
    private HashSet constraintSet;    // constraints that have been
				      // added to this

    // -----------------------------------
    // caching variables for performance

    // computing the cost is a little bit expensive -- cache the value
    // after it's computed once so we don't need to recompute
    float costCache;
    PointUniverse universeCache; // pointer to the universe for which
				 // we calculated costCache; if the
				 // universe changes, need to
				 // recompute.


    // -----------------------------------
    // should we force use of band join?  if false, band join is never
    // used.
    private static final boolean USE_BAND_JOIN = false;



    // build new plan
    ShapeQueryPlan()
    {
 	OutputWriter.println("\n* making new shapequeryplan");

	constraintSet = new HashSet();

	rules = new Vector();
	knownPoints = new HashSet();  

	costCache = -1f;
	universeCache = null;
    }

    // print
    void printQueryPlan()
    {
	if (rules==null)
	{
	    OutputWriter.println("No query plan to display");
	    return;
	}

	Iterator iter = rules.iterator();

	if (!iter.hasNext())
	{
	    OutputWriter.println("Query plan is empty");
	}
	else
	{
	    while (iter.hasNext())
	    {
		ShapeQueryRule rule = (ShapeQueryRule)iter.next();

		rule.print();
	    }
	}
    }

    // return the constraints in this plan.  needed to give user
    // choice of which constraints he wants to see in csv output.
    public HashSet getConstraints()
    {
	return constraintSet;
    }

    public float estimateCost(PointUniverse universe) {
        return estimateCost(universe, true);
    }

    // Give a general idea of how expensive this query is, based on
    // rules in plan.
    //
    // The goal is, for each rule, to figure out:
    //   1) how expensive is it for each partial solution it is
    //      applied to?
    //   2) how many rows will it give back?
    //
    // The individual rule classes return this information.
    //
    // By multiplying the costs by the rows, we get back a total cost
    // for the query
    public float estimateCost(PointUniverse universe, boolean flag)
    {

	if (costCache!=-1f && universeCache == universe && OutputWriter.DEBUG == false)
	{
	    // don't waste time computing estimated cost if we've
	    // already done so for this universe

	    return costCache;
	}

	float cost = -1f;
	float rowCount = -1f;

	Iterator iter = rules.iterator();
	while (iter.hasNext())
	{
	    ShapeQueryRule rule = (ShapeQueryRule) iter.next();
	    if (rowCount==-1f)
	    {
                float tempCost = rule.estimateCost(universe, rowCount);
		cost = tempCost;
		rowCount = rule.estimateReturnedRowMultiplier(universe);
                if(flag==true)
                    OutputWriter.print(String.valueOf(tempCost)+","+rowCount);

	    }
	    else
	    {
                float tempCost = rule.estimateCost(universe, rowCount);
		cost += tempCost;
		rowCount = rowCount * rule.estimateReturnedRowMultiplier(universe);
                if(flag==true)
                    OutputWriter.print(" + "+String.valueOf(tempCost)+","+rowCount);

		// += cost * (  (ShapeQueryRule) (iter.next()) ).estimateCost(universe);
	    }

	    //	    System.out.println("cost="+cost);
	}

	costCache = cost;
	universeCache = universe;
	return cost;
    }

    public Object clone()
    {
	// make a deep copy of this plan

	// don't need to clone shape form
	ShapeQueryPlan newPlan = new ShapeQueryPlan();
	newPlan.rules = (Vector) rules.clone();
	newPlan.knownPoints = (HashSet) knownPoints.clone();
	newPlan.constraintSet = (HashSet) constraintSet.clone();
	return newPlan;
    }

    public boolean containsConstraint(Constraint constraint)
    {
	if (constraintSet.contains(constraint))
	{
//	    OutputWriter.println("  +++++found constraint ("+constraintSet.size()+" total constraints)");
	    return true;
	}
	return false;
    }

    // add this constraint to our rule list.  return true on success,
    // false if we cannot add it.
    public boolean addConstraintAsNextRule(Constraint constraint)
    {
	costCache = -1f;  // forget any cached cost because we're
			  // changing rules

	ShapeQueryRule ruleToAdd = null; // the rule we'll add, made
					 // out of this constraint

	if (USE_BAND_JOIN && rules.size()==0)  // we have no rules yet
	{
	    ruleToAdd = new BandJoinRule(constraint);
	    knownPoints.add(constraint.myRequiredEdgeStartPointName);
	    knownPoints.add(constraint.myRequiredEdgeEndPointName);
	    knownPoints.add(constraint.myTargetEdgeStartPointName);
	    knownPoints.add(constraint.myTargetEdgeEndPointName);
	}
	else  // not starting a band join
	{

	if (rules.size()==0)  // we have no rules yet
	{
	    // if we don't have any rules yet, make this a first-edge
	    // rule.

	    ruleToAdd = new FirstEdgeRule(constraint);

	    // we now know these points
	    knownPoints.add(constraint.myRequiredEdgeStartPointName);
	    knownPoints.add(constraint.myRequiredEdgeEndPointName);
	    //	    knownPoints.add(constraint.myTargetEdgeStartPointName);
	    //	    knownPoints.add(constraint.myTargetEdgeEndPointName);

	    rules.addElement(ruleToAdd);

            OutputWriter.print("made FirstEdgeRule rule from constraint: ");
	    constraint.print();

            
	    // forget that we just added a rule so we add another one
	    ruleToAdd = null;

	} // end we have no rules yet

	// we already have some rules, either because we had them at
	// the beginning, or from the firstedge rule above
	{

	    // -----------------------------------------------------
	    // FIRST, try to make a verify rule out of this rule --
	    // this is cheaper than an index lookup

	    // if we already know about all 4 points involved, we can
	    // add this rule as a verify
	    if (knownPoints.contains(
		    constraint.myRequiredEdgeStartPointName) &&
		knownPoints.contains(
		    constraint.myRequiredEdgeEndPointName) &&
		knownPoints.contains(
		    constraint.myTargetEdgeStartPointName) &&
		knownPoints.contains(
		    constraint.myTargetEdgeEndPointName))
	    {
		// use verify rule because we already know these
		// points
		ruleToAdd = new VerifyRule(constraint);
                
		OutputWriter.print("made verify rule from constraint: ");
		constraint.print();
	    }
            
            //if exactly two points composing one edge are known
	    else if ((knownPoints.contains(
		    constraint.myRequiredEdgeStartPointName) &&
		knownPoints.contains(
		    constraint.myRequiredEdgeEndPointName) &&
		!knownPoints.contains(
		    constraint.myTargetEdgeStartPointName) &&
		!knownPoints.contains(
		    constraint.myTargetEdgeEndPointName)) || 
                (!knownPoints.contains(
		    constraint.myRequiredEdgeStartPointName) &&
		!knownPoints.contains(
		    constraint.myRequiredEdgeEndPointName) &&
		knownPoints.contains(
		    constraint.myTargetEdgeStartPointName) &&
		knownPoints.contains(
		    constraint.myTargetEdgeEndPointName))) // we can't make verify rule out of this constraint;
		 // try index lookup rule
	    {
                // HERE WE ADD THE USE OF THE <ANGLE, DISTANCE> AND <DISTANCE, ANGLE> 
                // INDICES. -DJH
                
                // We want to use the <ANGLE, DISTANCE> or <DISTANCE, ANGLE> 
                // indices only if exactly two out of the four points are known.
                // if three out of the four points are known, the other index lookups 
                // are cheaper.
                
                
                //reverse the constraint if it is the target edge that is known
                if (!knownPoints.contains(
		    constraint.myRequiredEdgeStartPointName) &&
		!knownPoints.contains(
		    constraint.myRequiredEdgeEndPointName) &&
		knownPoints.contains(
		    constraint.myTargetEdgeStartPointName) &&
		knownPoints.contains(
		    constraint.myTargetEdgeEndPointName)) {
                    
		    //OutputWriter.print("  reversing constraint:\n");
		    //constraint.print();

                    //constraint = constraint.copy();
                    constraint.reverse();

		    //OutputWriter.print("  to:\n");
		    //constraint.print();
                }
                
                
                //now add an EdgeIndexLookupRule
                ruleToAdd = new EdgeIndexLookupRule(constraint);

                // we now know two new points
                knownPoints.add(constraint.myTargetEdgeStartPointName);
                knownPoints.add(constraint.myTargetEdgeEndPointName);

                OutputWriter.print("made EdgeIndexlookup rule from constraint: ");
                constraint.print();
                
                
            }
            //at this point the possibility is that either 3 points are known or 1 point is known
            else 
            {
                

		// ---------------------------------------------------------
		// BEFORE WE MAKE AN <ID, ANGLE> OR <ID, DISTANCE> INDEX LOOKUP 
                // RULE, check if by
		// swapping the required and target points, we can apply
		// this rule.
		// 
		// To test, we see if we know both target points and
		// ONE required point, for a total of 3/4 points
		// involved in the constriant we're working with, with
		// the missing one being part of the relative edge.
		//
		// If that's the case, we'll swap the relative and
		// target edges.
		//
		// (If we knew all 4, we handled this in the 'verify'
		// section above, so we're not here.)  (If we know 3/4
		// and the missing one is part of the target edge,
		// we'll handle it below.)
		if (// check for first target point
		    knownPoints.contains(
			constraint.myTargetEdgeStartPointName) &&
			
		    // check for second target point
		    knownPoints.contains(
			constraint.myTargetEdgeEndPointName) &&
		    (
			// check for first optional point
			knownPoints.contains(
			    constraint.myRequiredEdgeStartPointName) ||

			// check for second optional point
			knownPoints.contains(
			    constraint.myRequiredEdgeEndPointName))
		    ) // end of test for 3/4 points, including both target
		    // ones
		{
		    // we invert this rule (i.e. swap 'required' and
		    // 'target' and adjust measures appropriately), and
		    // then we can add rule for this constraint because we
		    // already know about the edge it requires, as well as
		    // one point in the target edge
		    
                    //OutputWriter.print("  reversing constraint:\n");
		    //constraint.print();

                    //constraint = constraint.copy();
                    constraint.reverse();

		    //OutputWriter.print("  to:\n");
		    //constraint.print();

		}  // end of deciding whether to reverse the current
		   // constraint

		// ---------------------------------------------------------
		// NEXT, we can check to see if we can apply this rule
		// (knowing that reversal was done above if
		// appropriate)

		// See if we know both required points and ONE target
		// point. (If we knew all 4, we handled this in the
		// 'verify' section above, so we're not here; if we
		// knew 3/4 and the missing one was a required point,
		// we reversed the constraint above.)
		if (
		    // check for first required point
		    knownPoints.contains(
			constraint.myRequiredEdgeStartPointName) &&

		    // check for second required point
		    knownPoints.contains(
			constraint.myRequiredEdgeEndPointName) &&

		    // check for optional points
		    (
			// check for first optional point
			knownPoints.contains(
			    constraint.myTargetEdgeStartPointName) ||
			
			// check for second optional point
			knownPoints.contains(
			    constraint.myTargetEdgeEndPointName)
			)
		    ) // end of check for 3/4 points, including both
		      // required ones
		{
		    // we can add rule for this constraint because we
		    // already know about the edge it requires, as
		    // well as one point in the target edge

		    // use indexlookup rule because we don't know
		    // one of the needed points

		    // figure out which point is actually new
		    String newPointName =
			constraint.myTargetEdgeStartPointName;
		    String relativePointName =
			constraint.myTargetEdgeEndPointName;
		    if (knownPoints.contains(
			    constraint.myTargetEdgeStartPointName))
		    {
			// we already know target start point
			newPointName =
			    constraint.myTargetEdgeEndPointName;
			relativePointName =
			    constraint.myTargetEdgeStartPointName;
		    }

		    // -----------------------------------
		    // NOW, we know we can add this rule.  It's
		    // already been reversed if appropriate.  We know
		    // which point we need to find and how we're going
		    // to find it.  So finally, add the rule.
		    ruleToAdd = new PointIndexLookupRule(newPointName,
			relativePointName,
			constraint);

		    // we now know the new point
		    knownPoints.add(newPointName);

		    OutputWriter.print("made PointIndexlookup rule from constraint: ");
		    constraint.print();
		} 
	    }
	} // end we already have some rules
	} // end not starting a band join

	// ruleToAdd is null if we didn't make a rule from this
	// constraint; otherwise it's the rule we made.
	if (ruleToAdd!=null)
	{
	    // add the rule we just created
	    rules.addElement(ruleToAdd);

	    // remember that this constraint is used in this query plan
	    constraintSet.add(constraint);

	    // since we added this constraint, return true
	    return true;
	}

	return false; // we didn't add this constraint -- return false
    }

    // adds all of the constraints supplied to the current plan.
    // return true if we found a valid plan.
    public boolean addConstraintListAsNextRules(LinkedList constraintsToAdd)
    {
	costCache = -1f;  // forget any cached cost because we're
			  // changing rules

	OutputWriter.println("* adding "+
			     constraintsToAdd.size()
			     +" new constraints...");

	int iterationCount = 0;
	// iterate until all constraints have been turned into rules

	while (!constraintsToAdd.isEmpty())
	{
	    boolean addedThisRound = false; // have we added anything
					    // on this iteration?

	    // -------------------------------------------------------------
	    // before we look to add this constraint, some sanity checking
	    iterationCount++;
	    if (iterationCount>1000)
	    {
		OutputWriter.println("*** No query plan found (after 1000 iterations)");
		return false;
	    }

	    Iterator iter = constraintsToAdd.iterator();

	    // print constraints
// 	    while (iter.hasNext())
// 	    {
// 		Constraint currentConstraint = (Constraint)iter.next();
// 		currentConstraint.print();
// 	    }
// 	    iter = constraintsToAdd.iterator();

 	    // iterate through constraints to find one we can turn into a rule

	    // loop until either we add something or we run out of
	    // things to add
	    while (!addedThisRound && iter.hasNext())
	    {
		Constraint currentConstraint = (Constraint)iter.next();

		if (addConstraintAsNextRule(currentConstraint))
		{
		    iter.remove();
		    addedThisRound = true;
		}
	    }

	    if (!addedThisRound)
	    {
		// we went through all our constraints and couldn't add any
		OutputWriter.println("*** No query plan found (impossible to generate)");
		rules = null;
		return false;
	    }
	}

	return true;
    }

    // execute this query.  'set' is the plane of points to examine.
    int execute(PointSet set, boolean eliminateDuplicates, HashMap edgesToDraw)
    {
	if (rules==null)
	{
	    OutputWriter.println("No query plan to execute.");
	    return -1;
	}
	
	// to time execution
	Calendar startTime;
	Calendar finishTime;

	Iterator iter = rules.iterator();
	QuerySolutionSet solutions = new QuerySolutionSet(edgesToDraw);

	OutputWriter.print("* performing query...");

	// time query
	startTime = Calendar.getInstance();

        OutputWriter.clearOldSVG(set.visibleName);

        /*
        int i = 1;

	while (iter.hasNext())
	{
 	    ShapeQueryRule currentRule = (ShapeQueryRule)(iter.next());
 	    solutions = currentRule.apply(solutions, set);
	    OutputWriter.print("["+solutions.size()+" solutions]\n");

            //this is for testing - DJH
            OutputWriter.printSVGResults(set.visibleName + "_" + i, currentRule.getDescription(), solutions, set);
            i++;

            //the special case of an absolute distance requires one edge in a 
            //constraint to be an edge from a point to itself. This causes a HUGE 
            //number of duplicates, generated all by the first edge rule. So we 
            //remove them here. Obviously, something smarter could be done. -DJH
            if (currentRule instanceof FirstEdgeRule && ((FirstEdgeRule)currentRule).edgeToSelf()) {
        	solutions = QuerySolutionSet.eliminateDuplicates(solutions);
                OutputWriter.printSVGResults(set.visibleName + "_" + i, "FirstEdge Duplicates Eliminated", solutions, set);
                i++;
            }
        }
        */

        solutions = recurseRules(set, solutions, rules, 1);
	OutputWriter.print("query returned " + solutions.size()+" solutions.\n");

	finishTime = Calendar.getInstance();
	OutputWriter.print("===> search time: "
			   +(finishTime.getTimeInMillis()-
			     startTime.getTimeInMillis())
			   +" milliseconds\n");

	if (eliminateDuplicates) {
	    solutions = QuerySolutionSet.eliminateDuplicates(solutions);
            //OutputWriter.printSVGResults(set.visibleName + "_" + i, "Duplicates Eliminated", solutions, set);
        }
	// store the full solutions (not just as displayable vector)
	// with each set
	set.setSolutionSet(solutions);

	return solutions.size();
    }

    
    QuerySolutionSet recurseRules(PointSet set, QuerySolutionSet solutions, Vector rules, int k) {
                
        if(rules.size()==0) {
            //OutputWriter.print("\t\t["+solutions.size()+" final solutions]\n");
            return (QuerySolutionSet)solutions.clone();
        }
        
        if(rules.elementAt(0) instanceof FirstEdgeRule) {
            ShapeQueryRule currentRule = (ShapeQueryRule)(rules.elementAt(0));
            rules.remove(0);
            solutions = currentRule.apply(solutions, set);
            if (currentRule instanceof FirstEdgeRule && ((FirstEdgeRule)currentRule).edgeToSelf())
                solutions = QuerySolutionSet.eliminateDuplicates(solutions);
        }

        QuerySolutionSet<QuerySolution> finalSolutions = new QuerySolutionSet<QuerySolution>(solutions.getEdgesToDraw());
        QuerySolutionSet<QuerySolution> buffer = new QuerySolutionSet<QuerySolution>(solutions.getEdgesToDraw());
        Iterator<QuerySolution> iter = solutions.iterator();
        int count=0;

        while(iter.hasNext()) {
            QuerySolution tempSol = iter.next();
            buffer.add(tempSol);
            //solutions.remove(tempSol);
            count++;
            
            if(count%k==0 || !iter.hasNext()) {
                //OutputWriter.print("["+buffer.size()+"/"+count+"/"+solutions.size()+" solutions]\n");   
                ShapeQueryRule currentRule = (ShapeQueryRule)(rules.elementAt(0));
                buffer = currentRule.apply(buffer, set);
            
                Vector leftOverRules = (Vector)rules.clone();
                leftOverRules.remove(0);
                if(buffer.size()>0) {
                    buffer = recurseRules(set, buffer, leftOverRules, k);
                    
                    for(Iterator iterBuffer = buffer.iterator();iterBuffer.hasNext();)
                        finalSolutions.add((QuerySolution)iterBuffer.next());
                }
                buffer.clear();
            }
       }

        //OutputWriter.print("\t\t["+finalSolutions.size()+" final solutions]\n");     
        return finalSolutions;
    }
        
    
    // given measure 'angle', an angle in degrees, return the
    // equivalent measure in the range -180 < n <= 180
    public static float angleAdjust(float measure)
    {
	float startMeasure = measure;
	measure = measure % 360;
	if (measure > 180.0f)
	{
	    measure-=360.0f;
	}
	else if (measure <=-180.0f)
	{
	    measure+=360.0f;
	}

// 	if (startMeasure != measure)
// 	{
// 	    OutputWriter.println("converting ["+startMeasure+"] to ["
// 			       +measure+"]");
// 	}

	return measure;
    }
}
