package ShapeQuery;

import java.util.HashSet;
import java.util.Iterator;

// ShapeQueryRule.java: abstract class defining operations that must
//                      exist across all types of query rules

abstract class ShapeQueryRule
{
    // apply this rule to the given solution set; return a new
    // solution set with all the solutions from the supplied solution
    // set that can match this rule (extended as appropriate).
    abstract QuerySolutionSet apply(QuerySolutionSet<QuerySolution> solutions,
				    PointSet points);

    // return multiplier of how many rows will be added for each row
    // in a solution set that is passed in.  some rules will usually
    // increase that number, while others will decrease it; a decrease
    // is represented by a number between 0 and 1
    abstract float estimateReturnedRowMultiplier(PointUniverse universe);

    // given a number of rows, estimate how expensive this query will
    // be to execute
    abstract float estimateCost(PointUniverse universe, float rows);

    // print out what this rule looks like
    abstract void print();
    
    // get a detailed description of what this rule is going to do. For debugging purposes. DJH
    abstract String getDescription();
    
    
    /**
     * Computes an average reductive factor for a set of labels. A reductive factor
     * for a given label is the number of times the label appears divided by the number
     * of points.
     * @param somelabels A set of all labels that appeared in a constraint.
     * @param universe A point universe with the loaded PointSets.
     * @return Average reductive factor of all the labels in the set.
     */
    float computeLabelReduction(HashSet<String> somelabels, PointUniverse universe) {
        float factor = 0;
        String l = "";

        Iterator<String> iter = somelabels.iterator();
        while (iter.hasNext()) {
            l = iter.next();
            factor += universe.getLabelReduction(l);
        }

        //if there were no labels return 1, else return the averaged reductive factor
        return (factor == 0) ? 1 : (factor / (float)somelabels.size());
    }

}
