package ShapeQuery;
// VerifyRule.java: A rule that runs through a solution set and
//                  verifies that two edges have a certain
//                  relationship (i.e. distance or angle).  Any points
//                  that do not have this relationship are removed
//                  from the returned set.
//
//                  A VerifyRule is made from a Constraint.

import java.util.*;  // for vector
import java.io.*;    // for reading files
import java.net.*;

// for drawing
import javax.swing.*;
import javax.swing.event.*;
import java.awt.geom.Line2D;
import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.awt.event.*;

class VerifyRule extends ShapeQueryRule
{
    // variables of this rule.  these might be better stored as a
    // "constraint" rather than as a bunch of separate variables.
    String myTargetEdgeStartPointName;
    String myTargetEdgeEndPointName;
    String myRelativeEdgeStartPointName;
    String myRelativeEdgeEndPointName;
    int myRuleType;
    float myFactor;
    float myDifference;
    float myTolerance;
    boolean myToleranceIsAbsolute;

    Constraint con;
    
    VerifyRule(Constraint constraint)
    {
        myTargetEdgeStartPointName = constraint.myTargetEdgeStartPointName;
	myTargetEdgeEndPointName = constraint.myTargetEdgeEndPointName;
	myRelativeEdgeStartPointName = constraint.myRequiredEdgeStartPointName;
	myRelativeEdgeEndPointName = constraint.myRequiredEdgeEndPointName;
	myRuleType = constraint.myRuleType;
	myFactor = constraint.myFactor;
	myDifference = constraint.myDifference;
	myTolerance = constraint.myTolerance;
	myToleranceIsAbsolute = constraint.myToleranceIsAbsolute;
        con = constraint.copy();
    }

    // print out this rule
    void print()
    {
	OutputWriter.println("> verify rule on points: "
			   +myTargetEdgeStartPointName+", "
			   +myTargetEdgeEndPointName+" and "
			   +myRelativeEdgeStartPointName+", "
			   +myRelativeEdgeEndPointName + "       " + getDescription());
    }


    String getDescription() {
        String s = "VerifyRule - ";
        switch(myRuleType)
	{
	    case Constraint.RULE_TYPE_DISTANCE:
                s += "Distance Verify: "; break;
            case Constraint.RULE_TYPE_ANGLE:
                s += "Angle Verify: "; break;
            default:
                s += "Unknown Ruletype! "; break;
        }
        s += "RelEdge:(" + myRelativeEdgeStartPointName + "," + myRelativeEdgeEndPointName + ") ";
        s += "TarEdge:(" + myTargetEdgeStartPointName + "," + myTargetEdgeEndPointName + ") ";
        s += "Factor=" + myFactor + " Diff=" + myDifference + " Tol=" + myTolerance + " Abs=" + myToleranceIsAbsolute;
        return s;
        
    }
    // cost/rows estimation:
    //
    // the idea here is that the amount of work we need to do is
    // proportional to the number of rows.  ideally there would be a
    // factor here to make this number proportional to the cost you
    // get back from the other rules.
    //
    // note that in the case of the verify rule, it's always cheaper
    // to execute all your verify rules before doing anything else.
    // so i may be misestimating the number of rows returned, but
    // since it should be substantially smaller than you'd get from
    // another rule, verify rules should always be preferred.  the
    // time this would matter is where you have 2 candidate verify
    // rules and you want to execute the more restrictive one first;
    // this function won't tell you which one is better.  but since
    // verify rules are so cheap to execute anyway, it doesn't make a
    // big difference which order you do them in.
    float estimateCost(PointUniverse universe, float rows)
    {
	return rows;
    }

    float estimateReturnedRowMultiplier(PointUniverse universe)
    {
	// verify rules ususally reduce rowcount by a lot
        //reduce the rowcount by the label reduction factor - DJH
        return .05f * computeLabelReduction(con.getAllLabels(), universe);
	//return .01f;
    }


    // apply this rule to the current set to reducing the number of
    // rows to only those that match this constraint
    QuerySolutionSet apply(QuerySolutionSet solutions,
			   PointSet pointSet)

    {
	OutputWriter.printDebugDetail("  applying verify rule");
	
	// rather than copy the solutions list, we'll edit it in place
	Iterator iter = solutions.iterator(); // initial solutions
	
	int count = 0;
	while (iter.hasNext())
	{
	    // look at next solution
	    QuerySolution currentSolution = (QuerySolution)iter.next();

            
            
            //CHECK IF TWO EDGES HAVE REQUIRED POINT LABELS - DJH
            if(!pointsHaveCorrectLabels(con, currentSolution)) {
                iter.remove();
                continue;
            }
            
            
            
            
	    // -------------------------------------------------
	    // figure out what angle/distance we're looking for

	    // the measure we're looking for
	    float targetMeasure = 0.0f;

	    if (myRelativeEdgeStartPointName!="")
	    {
		// this is a relative constraint; construct the relative edge
		PointPair relativeEdge = new PointPair(
		    currentSolution.getPoint(myRelativeEdgeStartPointName),
		    currentSolution.getPoint(myRelativeEdgeEndPointName));

		switch(myRuleType)
		{
		    case Constraint.RULE_TYPE_DISTANCE:
		    {
			targetMeasure = relativeEdge.distance;
			break;
		    }
		    case Constraint.RULE_TYPE_ANGLE:
		    {
			targetMeasure = relativeEdge.angle;
			break;
		    }
		}

		// calculate the target based on the multiplier and difference
		targetMeasure = myFactor * targetMeasure + myDifference;

	    }
	    else //(NOTE BY DJH) - BASED ON HOW VERIFY RULES ARE CONSTRUCTED IN SHAPEQUERYPLAN, I DON'T THINK THIS CONDITION WILL EVER BE MET
	    {
		// an absolute constraint; 'difference' is our target
		// measure
		targetMeasure = myDifference;
	    }

	    // -------------------------------------------------
	    // apply tolerance to find max and min
	    float lowMeasure, highMeasure;
	    if (myToleranceIsAbsolute)
	    {
		lowMeasure = targetMeasure - myTolerance;
		highMeasure = targetMeasure + myTolerance;
	    }
	    else
	    {
		lowMeasure = targetMeasure*(1-myTolerance/100.0f);
		highMeasure = targetMeasure*(1+myTolerance/100.0f);
	    }

	    // -------------------------------------------------
	    // compare with target edge

	    Point firstPoint = 
		currentSolution.getPoint(myTargetEdgeStartPointName);
	    Point secondPoint = 
		currentSolution.getPoint(myTargetEdgeEndPointName);

	    if (firstPoint==null)
	    {
		OutputWriter.println("point "
				     +myTargetEdgeStartPointName
				     +" not found in current solution");
		return null;
	    }
	    if (secondPoint==null)
	    {
		OutputWriter.println("point "
				     +myTargetEdgeEndPointName
				     +" not found in current solution");
		return null;
	    }

	    PointPair targetEdge = new PointPair(firstPoint,
						 secondPoint);

	    switch(myRuleType)
	    {
	        case Constraint.RULE_TYPE_DISTANCE:
		{
                    //OutputWriter.print(" D ");
		    if (targetEdge.distance > highMeasure ||
			targetEdge.distance < lowMeasure)
		    {
			// target edge doesn't meet requirement
			iter.remove();
		    }
		    break;
		}
		case Constraint.RULE_TYPE_ANGLE:
		{
                    //OutputWriter.print(" A ");
		    // adjust measures to be in range -180 < n <= 180
		    highMeasure = ShapeQuery.angleAdjust(highMeasure);
		    lowMeasure = ShapeQuery.angleAdjust(lowMeasure);

		    // check order of angles; if highMeasure <
		    // lowMeasure, that's an angle wrapping situation
		    // that we need to resolve
		    if (highMeasure > lowMeasure)
		    {
			if (targetEdge.angle > highMeasure ||
			    targetEdge.angle < lowMeasure)
			{
			    // target edge doesn't meet requirement 
			    iter.remove();
			}
		    }
		    else
		    {
			// we have an angle wrapping situation -- e.g.:
			//
			//     low \  |
			//   measure\ |
			//           \|
			// ------------------------
			//           /|
			//   high   / |
			// measure /  |
			//
			// (lowmeasure = 135, highmeasure = -135
			// (adjusted from 225)

			// test for angle around x axis
			if ((targetEdge.angle < highMeasure &&
			     targetEdge.angle > -180.0f) ||
			    (targetEdge.angle > lowMeasure &&
			     targetEdge.angle <= 180.0f))
			{
			    // keep it
			}
			else
			{
			    // target edge doesn't meet requirement 
			    iter.remove();
			}
		    }
		    break;
		}
	    }

	    // warning: we may have just removed the current solution

	    count++;
            
	    if (count%10000==0)
		OutputWriter.printDebugDetail(".");
	}
	return solutions;
    }
    
    
    
    
    
    /**
     * Checks if all four points to verify also have the correct labels.
     *
     * @param constr The constraint that has lists of allowable labels for each point.
     * @param currentSolution The current solution to that has some points to check.
     * @return True if all points have one or more matching labels (if no labels are required for 
     * a given point the test of the point returns true).
     */
    private boolean pointsHaveCorrectLabels(Constraint constr, QuerySolution currentSolution) {
        Point relStart = currentSolution.getPoint(myRelativeEdgeStartPointName);
        Point relEnd = currentSolution.getPoint(myRelativeEdgeEndPointName);
        Point tarStart = currentSolution.getPoint(myTargetEdgeStartPointName);
        Point tarEnd = currentSolution.getPoint(myTargetEdgeEndPointName);
        
        return (relStart.labelMatches(constr.myRequiredEdgeStartPointLabels) &&
                relEnd.labelMatches(constr.myRequiredEdgeEndPointLabels) &&
                tarStart.labelMatches(constr.myTargetEdgeStartPointLabels) && 
                tarEnd.labelMatches(constr.myTargetEdgeEndPointLabels));
    }
    
    
    
    
}
